/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the main TGNG class.

*/
#include <utility>
#include <vector>

#include <TMath.h>
#include <TH1D.h>
#include <TH2D.h>
#include <TObjString.h>

#include "TGNGP.h"



TGNGP::TGNGP() :
	TNamed("",""),
	fUnits(),
	fAttributes(),
	fEpsilon_b(0.05),
	fEpsilon_be(0.05),
	fEpsilon_n(0.0006), 
	fEpsilon_ne(0.0006), 
	fRampTime(500000),
	fLambda(300),    
	fT_max(100),     
	fAlpha(0.5),     
	fBeta(0.0005),
	fMaxUnitCnt(100),
	fLambdaDelay(0),
	fGamma(0.02),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fExtensions(),
	fLishWeight(kFALSE),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fRawInput(0),
	fUseLearningInhibition(kFALSE),
	fUseEpsilonRamp(kFALSE),
	fGrowth(kTRUE),
	fTopLayerSize(-1)
{
	// default constructor setting the parameters of the GNG to
	// the values proposed in
	//
	// Fritzke, B. A Growing Neural Gas Network Learns Topologies 
	// Advances in Neural Information Processing Systems 7, 
	// MIT Press, 1995, 625-632
	//
	// the gng is not initialized

	fUnits.SetOwner(kTRUE);
	fExtensions.SetOwner(kTRUE);
	fAttributes.SetOwner(kTRUE);	
}


TGNGP::TGNGP(const char* name, const char* title) :
	TNamed(name,title),
	fUnits(),
	fAttributes(),
	fEpsilon_b(0.05),
	fEpsilon_be(0.05),
	fEpsilon_n(0.0006), 
	fEpsilon_ne(0.0006), 
	fRampTime(500000),
	fLambda(300),    
	fT_max(100),     
	fAlpha(0.5),     
	fBeta(0.0005),
	fMaxUnitCnt(100),
	fLambdaDelay(0),
	fGamma(0.02),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fExtensions(),
	fLishWeight(kFALSE),
	fLishWeightCenter(0.25),
	fLishWeightSteepness(25.0),
	fRawInput(0),
	fUseLearningInhibition(kFALSE),
	fUseEpsilonRamp(kFALSE),
	fGrowth(kTRUE),
	fTopLayerSize(-1)
{
	// constructor setting the parameters of the GNG to
	// the values proposed in
	//
	// Fritzke, B. A Growing Neural Gas Network Learns Topologies 
	// Advances in Neural Information Processing Systems 7, 
	// MIT Press, 1995, 625-632
	//
	// the gng is not initialized

	fUnits.SetOwner(kTRUE);
	fExtensions.SetOwner(kTRUE);
	fAttributes.SetOwner(kTRUE);
}

TGNGP::~TGNGP()
{
	// the GNG is reset and the extensions are deleted

	ResetGNG();
	fExtensions.Delete();
	fAttributes.DeleteAll();
}

void TGNGP::Init(Int_t topLayerSize, TGNGDataElement *prototype, TGNGDistanceMeasure *distMeasure)
{
	// initializes the gng with two units that have the given prototypes
	// and the supplied distance measure.

	if (prototype == 0) {
		RaiseError(TGNGP::ecParamNull,"Prototypes measure can not be 0!");
		return;
	}

	if (distMeasure == 0) {
		RaiseError(TGNGP::ecParamNull,"Distance measure can not be 0!");
		return;
	}

	if (distMeasure->IsCompatible(prototype,0) == kFALSE) {
		RaiseError(TGNGP::ecIncomp,"Distance measure not compatible with data elements!");
		return;		
	}

	if (topLayerSize < 1) {
		RaiseError(TGNGP::ecFatal,"Top layer size can not be smaller 1!");
		return;				
	}

	if (fInitialized == kTRUE)
		ResetGNG();

	fDistMeasure = distMeasure;

	fTopLayerSize = topLayerSize;

	for (Int_t i = 0; i < fTopLayerSize; ++i) {

		TObjArray *topLayerUnit = new TObjArray();
		topLayerUnit->SetOwner(kTRUE);

		fUnits.Add(topLayerUnit);

		TGNGDataElement *newElem = static_cast<TGNGDataElement*>(prototype->Clone());
		newElem->Randomize();

		topLayerUnit->Add(new TGNGUnit(newElem,(0*fTopLayerSize)+i));

		newElem = static_cast<TGNGDataElement*>(prototype->Clone());
		newElem->Randomize();

		topLayerUnit->Add(new TGNGUnit(newElem,(1*fTopLayerSize)+i));

	}

	delete prototype;

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));

			for (Int_t j = 0; j < fTopLayerSize; ++j) {
				TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(j));
				Int_t nrOfNodes = topLayerUnit->GetEntriesFast();
				for (Int_t k = 0; k < nrOfNodes; ++k) {
					TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(k));
					curExt->OnNewUnit(0,curUnit);
				}
			}
			
		}
	}

	fInitialized = kTRUE;

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnGNGInit(0);
		}
	}
}

void TGNGP::ManualInit(TGNGDistanceMeasure *distMeasure)
{
	// starts manual initialization of the gng with no units and the supplied 
	// distance measure. New units and edges have to be added manually after 
	// a call to this method. The end of the manual initialization must be
	// signalled by a call to ManualInitDone().

	if (distMeasure == 0) {
		RaiseError(TGNGP::ecParamNull,"Distance measure can not be 0!");
		return;
	}

	if (fInitialized == kTRUE)
		ResetGNG();

	fDistMeasure = distMeasure;
}

void TGNGP::ManualInitDone()
{
	// finishes the manual initialization 

	if (fInitialized == kTRUE)
		return;

	Int_t nrOfTLUnits = fUnits.GetEntriesFast();

	if (nrOfTLUnits == 0) {
		RaiseError(TGNGP::ecFatal,"No top layer units defined during manual init!");
		return;		
	}

	for (Int_t i = 0; i < nrOfTLUnits; ++i) {
		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		if (nrOfNodes == 0) {
			RaiseError(TGNGP::ecFatal,"No bottom layer nodes defined for top layer unit during manual init!");
			return;					
		}

		for (Int_t j = 0; j < nrOfNodes; ++j) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));
			
			if (fDistMeasure->IsCompatible(curUnit->fPrototype,0) == kFALSE) {
				RaiseError(TGNGP::ecIncomp,"Distance measure not compatible with a data element defined during manual init!");
				return;		
			}
		}		
	}

	// report units and edges to extensions
	{ 
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));

			for (Int_t j = 0; j < fTopLayerSize; ++j) {
				TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(j));
				Int_t nrOfNodes = topLayerUnit->GetEntriesFast();
				for (Int_t k = 0; k < nrOfNodes; ++k) {
					TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(k));
					curExt->OnNewUnit(0,curUnit);
				}
			}

		}
	}

	{ 
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));

			for (Int_t j = 0; j < fTopLayerSize; ++j) {
				TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(j));
				Int_t nrOfNodes = topLayerUnit->GetEntriesFast();
				for (Int_t k = 0; k < nrOfNodes; ++k) {
					TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(k));

					Int_t nrOfEdges = curUnit->fEdges.GetEntriesFast();
					for (Int_t l = 0; l < nrOfEdges; ++l) {
						TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(l));
						curExt->OnNewEdge(0,curEdge);
					}

				}
			}
		}
	}

	fInitialized = kTRUE;

	{ // report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			curExt->OnGNGInit(0);
		}
	}
}

void TGNGP::ResetGNG()
{
	// resets the GNG

	if (fInitialized == kFALSE)
		return;

	// delete edges manually as they are shared between units
	Int_t unitCnt = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < unitCnt; ++i) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		for (Int_t j = 0; j < nrOfNodes; ++j) {

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));

			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t k = 0; k < edgeCnt; ++k) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(k));

				// report to extensions
				if ((curEdge->fUnitA != 0) && (curEdge->fUnitB != 0)) { 
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t l = 0; l < nrOfExt; ++l) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(l));
						curExt->OnEdgeRemoval(0,curEdge);
					}
				}	

				if (curEdge->fUnitA != 0) {
					curEdge->fUnitA = 0;
				} else if (curEdge->fUnitB != 0) {
					curEdge->fUnitB = 0;
				}

				if ((curEdge->fUnitA == 0) && (curEdge->fUnitB == 0)) {
					delete curEdge;
					curUnit->fEdges.RemoveAt(k);
				}
			}

			curUnit->fEdges.Clear();
		}
	}

	for (Int_t i = 0; i < unitCnt; ++i) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		for (Int_t j = 0; j < nrOfNodes; ++j) {

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t k = 0; k < nrOfExt; ++k) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
					curExt->OnUnitRemoval(0,curUnit);
				}
			}	

		}		
	}

	fUnits.Delete();

	fInitialized = kFALSE;

	fAttributes.DeleteAll();

	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		curExt->OnGNGReset(0);
	}	
}


TGNGUnit* TGNGP::Feed(const TGNGDataElement *input, Double_t *minDist, TObject *rawInput, Double_t *ratio)
{
	// Inputs a vector into the GNG and returns the BMU.
	// The GNG is modified.

	if (fRawInput) {
		delete fRawInput;
		fRawInput = 0;
	}

	if (rawInput) {
		fRawInput = rawInput->Clone();
	}

	//printf("check inputs\n");
	if (fInitialized == kFALSE) {
		RaiseError(TGNGP::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TGNGP::ecParamNull,"Input is null!");
		return 0;				
	}

	fInpCnt++;

	// find first and second BMU within and across all top layer units
	TGNGUnit *atlS1 = 0;
	TGNGUnit *atlS2 = 0;

	Double_t atlS1ApproxDist = std::numeric_limits<Double_t>::max();
	Double_t atlS2ApproxDist = std::numeric_limits<Double_t>::max();

	std::vector<TGNGUnit*> bmus(fTopLayerSize*2,0);
	std::vector<Double_t> bmuDists(fTopLayerSize*2,0);
	std::vector<TGNGEdge*> bmuEdge(fTopLayerSize,0);

	Int_t unitCnt = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < unitCnt; ++i) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		Double_t s1ApproxDist = std::numeric_limits<Double_t>::max();
		Double_t s2ApproxDist = std::numeric_limits<Double_t>::max();

		Int_t idxS1 = i * 2;
		Int_t idxS2 = idxS1 + 1;

		for (Int_t j = 0; j < nrOfNodes; ++j) {

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));
			Double_t curDist = fDistMeasure->GetApproxDistance(curUnit->fPrototype,input,0);

			// within
			if (curDist < s1ApproxDist) {
				bmus[idxS2]  = bmus[idxS1];
				s2ApproxDist = s1ApproxDist;
				bmus[idxS1]  = curUnit;
				s1ApproxDist = curDist;
			} else if (curDist < s2ApproxDist) {
				bmus[idxS2]  = curUnit;
				s2ApproxDist = curDist;
			}

			// across
			if (curDist < atlS1ApproxDist) {
				atlS2           = atlS1;
				atlS2ApproxDist = atlS1ApproxDist;
				atlS1           = curUnit;
				atlS1ApproxDist = curDist;
			} else if (curDist < atlS2ApproxDist) {
				atlS2           = curUnit;
				atlS2ApproxDist = curDist;
			}
		}

		if (bmus[idxS1] != 0) {
			bmuDists[idxS1] = fDistMeasure->GetDistance(bmus[idxS1]->fPrototype,input,0);
		} else {
			bmuDists[idxS1] = TMath::QuietNaN();
		}

		if (bmus[idxS2] != 0) {
			bmuDists[idxS2] = fDistMeasure->GetDistance(bmus[idxS2]->fPrototype,input,0);
		} else {
			bmuDists[idxS2] = TMath::QuietNaN();
		}

	}

	if ((atlS1 == 0) || (atlS2 == 0)) {
		RaiseError(TGNGP::ecNoBMU,"Could not find first and second BMU across TL units! Something is very wrong...");
		return 0;				
	}

	for (Int_t i = 0; i < unitCnt; ++i) {
		Int_t idxS1 = i * 2;
		Int_t idxS2 = idxS1 + 1;
		if ((bmus[idxS1] == 0) || (bmus[idxS2] == 0)) {
			RaiseError(TGNGP::ecNoBMU,"Could not find first and second BMU within TL units! Something is very wrong...");
			return 0;				
		}				
	}

	// get real distances 

	Double_t s1Dist = fDistMeasure->GetDistance(atlS1->fPrototype,input,0);
	Double_t s2Dist = fDistMeasure->GetDistance(atlS2->fPrototype,input,0);

	if (minDist != 0) {
		*minDist = s1Dist;
	}

	Double_t dRatio = 1.0;
	Double_t relDist = fDistMeasure->GetDistance(atlS1->fPrototype,atlS2->fPrototype,0);

	if ((TMath::IsNaN(relDist) == kFALSE) && (relDist != 0)) {
		dRatio = ((s1Dist - s2Dist) / relDist) + 1.0;
	}

	if (dRatio > 1.0) dRatio = 1.0;
	if (dRatio < 0.0) dRatio = 0.0;		

	if (ratio) {
		*ratio = dRatio;
	}

	if (TMath::IsNaN(s1Dist)) {
		RaiseError(TGNGP::ecNaN,"Distance to BMU is NaN! Aborting...");
		return 0;				
	}

	if (TMath::IsNaN(s2Dist)) {
		RaiseError(TGNGP::ecNaN,"Distance to second BMU is NaN! Aborting...");
		return 0;				
	}

	//printf("found BMUs\n");

	// increase edge age of all s1 neighbours
	// try to find edge to s2
	//printf("increase edge age of all s1 neighbours / try to find edge to s2\n");

	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		TGNGUnit *s2 = bmus[(i*2)+1];

		TGNGEdge *s2Edge = 0;	
		Int_t edgeCnt = s1->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(j));	
			if (fUseLearningInhibition) {	
				curEdge->fAge += (1.0 - s1->fLearnInh);
			} else {
				curEdge->fAge += 1;			
			}
			if ((curEdge->fUnitA == s2) || (curEdge->fUnitB == s2))
				s2Edge = curEdge;
		}

		// reset edge age to s2, create edge if necessary
		//printf("reset edge age to s2, create edge if necessary\n");
		if (s2Edge != 0) {
			s2Edge->fAge = 0;
		} else {
			s2Edge = new TGNGEdge(s1,s2);
			s1->fEdges.Add(s2Edge);
			s2->fEdges.Add(s2Edge);

			{ // report to extensions
				Int_t nrOfExt = fExtensions.GetEntriesFast();
				for (Int_t j = 0; j < nrOfExt; ++j) {
					TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(j));
					curExt->OnNewEdge(0,s2Edge);
				}
			}	

		}

		bmuEdge[i] = s2Edge;
	}

	//printf("increased edge ages\n");

	// search top layer edge 
	TGNGEdge *atlS2Edge = 0;
	Int_t edgeCnt = atlS1->fEdges.GetEntriesFast();
	for (Int_t i = 0; i < edgeCnt; ++i) {
		TGNGEdge *curEdge = static_cast<TGNGEdge*>(atlS1->fEdges.At(i));	
		if ((curEdge->fUnitA == atlS2) || (curEdge->fUnitB == atlS2)) {
			atlS2Edge = curEdge;
			break;
		}
	}

	if (atlS2Edge != 0) {
		atlS2Edge->fAge = 0;
	} else {
		atlS2Edge = new TGNGEdge(atlS1,atlS2);
		atlS1->fEdges.Add(atlS2Edge);
		atlS2->fEdges.Add(atlS2Edge);

		{ // report to extensions
			Int_t nrOfExt = fExtensions.GetEntriesFast();
			for (Int_t j = 0; j < nrOfExt; ++j) {
				TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(j));
				curExt->OnNewEdge(0,atlS2Edge);
			}
		}	
	}

	//printf("found top layer edge\n");


	// delete old edges
	//printf("delete old edges\n");
	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		TGNGUnit *s2 = bmus[(i*2)+1];

		Int_t edgeCnt = s1->fEdges.GetEntriesFast();
		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(j));
			if (curEdge->fAge > (Double_t)fT_max) {

				TGNGUnit *otherUnit = (curEdge->fUnitA != s1) ? curEdge->fUnitA : curEdge->fUnitB;

				{ // report to extensions
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t k = 0; k < nrOfExt; ++k) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
						curExt->OnEdgeRemoval(0,curEdge);
					}
				}	

				s1->fEdges.RemoveAt(j);
				otherUnit->fEdges.Remove(curEdge);
				otherUnit->fEdges.Compress();
				delete curEdge;
			}		
		}
		s1->fEdges.Compress();

	}

	//printf("deleted old edges\n");

	// delete isolated units	
	//printf("delete isolated units\n");
	if (fGrowth) {

		for (Int_t i = 0; i < unitCnt; ++i) {

			TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
			Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

			for (Int_t j = 0; j < nrOfNodes; ++j) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));
				if (curUnit->fEdges.GetEntriesFast() == 0) {

					{ // report to extensions
						Int_t nrOfExt = fExtensions.GetEntriesFast();
						for (Int_t k = 0; k < nrOfExt; ++k) {
							TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
							curExt->OnUnitRemoval(0,curUnit);
						}
					}	
					topLayerUnit->RemoveAt(j);
					delete curUnit;
				}

			}
			topLayerUnit->Compress();
			nrOfNodes = topLayerUnit->GetEntriesFast();
			// update unit indices
			for (Int_t j = 0; j < nrOfNodes; ++j) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));
				curUnit->fCurIdx = (j*fTopLayerSize)+i;
			}
		}
		
	}

	//printf("deleted isolated units\n");

	// update accumulated error of s1
	//printf("update accumulated error of s1\n");

	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		Double_t s1D = bmuDists[i*2];

		s1->fAccErr += s1D*s1D;

	}

	//printf("updated errors\n");


	Double_t eps_b = fEpsilon_b;
	Double_t eps_n = fEpsilon_n;

	if (fUseEpsilonRamp) {

		eps_b = fEpsilon_be;
		eps_n = fEpsilon_ne;

		if ((Double_t)fInpCnt < fRampTime) {
			eps_b = fEpsilon_b * TMath::Power(fEpsilon_be / fEpsilon_b, (Double_t)fInpCnt / fRampTime);
			eps_n = fEpsilon_n * TMath::Power(fEpsilon_ne / fEpsilon_n, (Double_t)fInpCnt / fRampTime);
		}

	}

	// adapt s1 prototype
	//printf("adapt s1 prototype\n");
	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		if (fUseLearningInhibition) {
			s1->fPrototype->Adapt(input,eps_b * (1.0 - s1->fLearnInh));
		} else {
			s1->fPrototype->Adapt(input,eps_b);
		}
	}

	// adapt s1 neighbor prototypes
	//printf("adapt s1 neighbor prototypes\n");

	TObjString lishkey("LISH");

	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		Double_t s1D = bmuDists[i*2];
		TGNGUnit *s2 = bmus[(i*2)+1];
		Double_t s2D = bmuDists[(i*2)+1];

		Int_t edgeCnt = s1->fEdges.GetEntriesFast();

		for (Int_t j = 0; j < edgeCnt; ++j) {
			TGNGEdge *curEdge = static_cast<TGNGEdge*>(s1->fEdges.At(j));
			
			Double_t lw = 1.0;
			if (fLishWeight) {
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
				if (lish != 0) {
					lw = GetEdgeLishStrength(lish);
				}			
			}
			
			if (curEdge->fUnitA != s1) {
				if (fUseLearningInhibition) {
					curEdge->fUnitA->fPrototype->Adapt(input,eps_n*lw* (1.0 - s1->fLearnInh));
				} else {
					curEdge->fUnitA->fPrototype->Adapt(input,eps_n*lw);
				}
			} else {
				if (fUseLearningInhibition) {
					curEdge->fUnitB->fPrototype->Adapt(input,eps_n*lw* (1.0 - s1->fLearnInh));
				} else {
					curEdge->fUnitB->fPrototype->Adapt(input,eps_n*lw);
				}
			}
		}

		// set learning inhibition of s1
		s1->fLearnInh = 1.0;

		TGNGEdge *s2Edge = bmuEdge[i];

		// report to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t k = 0; k < nrOfExt; ++k) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(k));
			curExt->OnUnitUpdate(0,s1,s1D,s2,s2D,s2Edge,input);
		}
		

	}

	//printf("adapted prototypes\n");


	// check if we should add a new unit
	//printf("check if we should add a new unit\n");
	if ((fInpCnt > fLambdaDelay) && ((fInpCnt % fLambda) == 0) && (fGrowth == kTRUE)) {

		{ // report to extensions
			Int_t nrOfExt = fExtensions.GetEntriesFast();
			for (Int_t i = 0; i < nrOfExt; ++i) {
				TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
				curExt->OnLambda(0,fInpCnt);
			}
		}	

		for (Int_t u = 0; u < unitCnt; ++u) {

			TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(u));
			Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

			if ((UInt_t)nrOfNodes < fMaxUnitCnt) {

				// find unit with greatest error
				//printf("find unit with greatest error\n");
				TGNGUnit *mxErr    =  0;
				Double_t  mxErrVal = -1;
				for (Int_t i = 0; i < nrOfNodes; ++i) {
					TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(i));
					if (curUnit->fAccErr > mxErrVal) {
						mxErr    = curUnit;
						mxErrVal = curUnit->fAccErr;
					}
				}

				if (mxErr == 0) {
					RaiseError(TGNGP::ecNoMaxErr,"Could not find unit with greatest error! Something is very wrong...");
					return 0;				
				}

				// find neighbor with greatest error
				//printf("find neighbor with greatest error\n");
				TGNGUnit *mxErrNb    =  0;
				Double_t  mxErrNbVal = -1;
				TGNGEdge *mxErrEdge  =  0;
				edgeCnt = mxErr->fEdges.GetEntriesFast();
				for (Int_t i = 0; i < edgeCnt; ++i) {
					TGNGEdge *curEdge   = static_cast<TGNGEdge*>(mxErr->fEdges.At(i));
					TGNGUnit *otherUnit = (curEdge->fUnitA != mxErr) ? curEdge->fUnitA : curEdge->fUnitB;
					if ((otherUnit->fCurIdx % fTopLayerSize) != u)
						continue;
					if (otherUnit->fAccErr > mxErrNbVal) {
						mxErrNb    = otherUnit;
						mxErrNbVal = otherUnit->fAccErr;
						mxErrEdge  = curEdge;
					}
				}

				if (mxErrNb == 0) {
					continue;
					//RaiseError(TGNGP::ecNoMaxErr,"Could not find neighbor unit with greatest error! Something is very wrong...");
					//return 0;				
				}

				// create new unit between mxErr and mxErrNb
				//printf("create new unit between mxErr and mxErrNb\n");
				TGNGUnit *newUnit = new TGNGUnit(mxErr->fPrototype->Interpolate(mxErrNb->fPrototype,0.5),(nrOfNodes*fTopLayerSize)+u);
				// delete old edge
				//printf("delete old edge\n");

				{ // report to extensions
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t i = 0; i < nrOfExt; ++i) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
						curExt->OnEdgeRemoval(0,mxErrEdge);
					}
				}	
				mxErr->fEdges.Remove(mxErrEdge);
				mxErr->fEdges.Compress();
				mxErrNb->fEdges.Remove(mxErrEdge);
				mxErrNb->fEdges.Compress();
				delete mxErrEdge;

				// create two new edges
				//printf("create two new edges\n");
				TGNGEdge *newEdge1 = new TGNGEdge(mxErr,  newUnit);
				TGNGEdge *newEdge2 = new TGNGEdge(mxErrNb,newUnit);

				mxErr->fEdges.Add(newEdge1);
				mxErrNb->fEdges.Add(newEdge2);
				newUnit->fEdges.Add(newEdge1);
				newUnit->fEdges.Add(newEdge2);

				// reduce error of mxErr and mxErrNb
				//printf("reduce error of mxErr and mxErrNb\n");
				mxErr->fAccErr -= fAlpha * mxErr->fAccErr;
				if (mxErr->fAccErr < 0.0)
					mxErr->fAccErr = 0.0;
				mxErrNb->fAccErr -= fAlpha * mxErrNb->fAccErr;
				if (mxErrNb->fAccErr < 0.0)
					mxErrNb->fAccErr = 0.0;

				// set the error of the new unit to the reduced error of mxErr
				//printf("set the error of the new unit to the reduced error of mxErr\n");
				newUnit->fAccErr = mxErr->fAccErr;

				topLayerUnit->Add(newUnit);
				

				{ // report to extensions
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t i = 0; i < nrOfExt; ++i) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
						curExt->OnNewUnit(0,newUnit);
					}
				}

				{ // report to extensions
					Int_t nrOfExt = fExtensions.GetEntriesFast();
					for (Int_t i = 0; i < nrOfExt; ++i) {
						TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
						curExt->OnNewEdge(0,newEdge1);
						curExt->OnNewEdge(0,newEdge2);
					}
				}			
			}
		}
	}

	//printf("added units\n");


	// reduce the error of all units by beta and
	// reduce the learning inhibition of all units by gamma
	//printf("reduce the error of all units by beta\n");

	for (Int_t u = 0; u < unitCnt; ++u) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(u));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		for (Int_t i = 0; i < nrOfNodes; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(i));

			curUnit->fAccErr -= fBeta * curUnit->fAccErr;
			if (curUnit->fAccErr < 0.0)
				curUnit->fAccErr = 0.0;

			curUnit->fLearnInh -= fGamma * curUnit->fLearnInh;
			if (curUnit->fLearnInh < 0.0)
				curUnit->fLearnInh = 0.0;
		}
	}

	//printf("reduced error and learning inhibition\n");

	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		Double_t s1D = bmuDists[i*2];
		TGNGUnit *s2 = bmus[(i*2)+1];
		Double_t s2D = bmuDists[(i*2)+1];

		// report result to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t j = 0; j < nrOfExt; ++j) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(j));
			curExt->OnFeedResult(0,s1,s1D,s2,s2D,input);
		}

	}

	//printf("done\n");

	//printf("return\n");
	return atlS1;
}



TGNGUnit* TGNGP::Peek(const TGNGDataElement *input, Double_t *minDist, TObject *rawInput, Double_t *ratio)
{
	// Inputs a vector into the GNG and returns the BMU.
	// The GNG is not modified.

	if (fRawInput) {
		delete fRawInput;
		fRawInput = 0;
	}

	if (rawInput) {
		fRawInput = rawInput->Clone();
	}

	// finds BMU
	if (fInitialized == kFALSE) {
		RaiseError(TGNGP::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TGNGP::ecParamNull,"Input is null!");
		return 0;				
	}

	fInpCnt++;


	// find first and second BMU within and across all top layer units
	TGNGUnit *atlS1 = 0;
	TGNGUnit *atlS2 = 0;

	Double_t atlS1ApproxDist = std::numeric_limits<Double_t>::max();
	Double_t atlS2ApproxDist = std::numeric_limits<Double_t>::max();

	std::vector<TGNGUnit*> bmus(fTopLayerSize*2,0);
	std::vector<Double_t> bmuDists(fTopLayerSize*2,0);

	Int_t unitCnt = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < unitCnt; ++i) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(i));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		Double_t s1ApproxDist = std::numeric_limits<Double_t>::max();
		Double_t s2ApproxDist = std::numeric_limits<Double_t>::max();

		Int_t idxS1 = i * 2;
		Int_t idxS2 = idxS1 + 1;

		for (Int_t j = 0; j < nrOfNodes; ++j) {

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(j));
			Double_t curDist = fDistMeasure->GetApproxDistance(curUnit->fPrototype,input,0);

			// within
			if (curDist < s1ApproxDist) {
				bmus[idxS2]  = bmus[idxS1];
				s2ApproxDist = s1ApproxDist;
				bmus[idxS1]  = curUnit;
				s1ApproxDist = curDist;
			} else if (curDist < s2ApproxDist) {
				bmus[idxS2]  = curUnit;
				s2ApproxDist = curDist;
			}

			// across
			if (curDist < atlS1ApproxDist) {
				atlS2           = atlS1;
				atlS2ApproxDist = atlS1ApproxDist;
				atlS1           = curUnit;
				atlS1ApproxDist = curDist;
			} else if (curDist < atlS2ApproxDist) {
				atlS2           = curUnit;
				atlS2ApproxDist = curDist;
			}
		}

		if (bmus[idxS1] != 0) {
			bmuDists[idxS1] = fDistMeasure->GetDistance(bmus[idxS1]->fPrototype,input,0);
		} else {
			bmuDists[idxS1] = TMath::QuietNaN();
		}

		if (bmus[idxS2] != 0) {
			bmuDists[idxS2] = fDistMeasure->GetDistance(bmus[idxS2]->fPrototype,input,0);
		} else {
			bmuDists[idxS2] = TMath::QuietNaN();
		}

	}

	if ((atlS1 == 0) || (atlS2 == 0)) {
		RaiseError(TGNGP::ecNoBMU,"Could not find first and second BMU across TL units! Something is very wrong...");
		return 0;				
	}

	for (Int_t i = 0; i < unitCnt; ++i) {
		Int_t idxS1 = i * 2;
		Int_t idxS2 = idxS1 + 1;
		if ((bmus[idxS1] == 0) || (bmus[idxS2] == 0)) {
			RaiseError(TGNGP::ecNoBMU,"Could not find first and second BMU within TL units! Something is very wrong...");
			return 0;				
		}				
	}

	// get real distances 

	Double_t s1Dist = fDistMeasure->GetDistance(atlS1->fPrototype,input,0);
	Double_t s2Dist = fDistMeasure->GetDistance(atlS2->fPrototype,input,0);

	if (minDist != 0) {
		*minDist = s1Dist;
	}

	Double_t dRatio = 1.0;
	Double_t relDist = fDistMeasure->GetDistance(atlS1->fPrototype,atlS2->fPrototype,0);

	if ((TMath::IsNaN(relDist) == kFALSE) && (relDist != 0)) {
		dRatio = ((s1Dist - s2Dist) / relDist) + 1.0;
	}

	if (dRatio > 1.0) dRatio = 1.0;
	if (dRatio < 0.0) dRatio = 0.0;		

	if (ratio) {
		*ratio = dRatio;
	}

	if (TMath::IsNaN(s1Dist)) {
		RaiseError(TGNGP::ecNaN,"Distance to BMU is NaN! Aborting...");
		return 0;				
	}

	if (TMath::IsNaN(s2Dist)) {
		RaiseError(TGNGP::ecNaN,"Distance to second BMU is NaN! Aborting...");
		return 0;				
	}


	for (Int_t i = 0; i < unitCnt; ++i) {

		TGNGUnit *s1 = bmus[i*2];
		Double_t s1D = bmuDists[i*2];
		TGNGUnit *s2 = bmus[(i*2)+1];
		Double_t s2D = bmuDists[(i*2)+1];

		// report result to extensions
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		for (Int_t j = 0; j < nrOfExt; ++j) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(j));
			curExt->OnPeekResult(0,s1,s1D,s2,s2D,input);
		}

	}

	return atlS1;
}

void TGNGP::SetEpsilon_b(Double_t val)
{
	fEpsilon_b = val;
}

void TGNGP::SetEpsilon_be(Double_t val)
{
	fEpsilon_be = val;
}

void TGNGP::SetEpsilon_n(Double_t val)
{
	fEpsilon_n = val;
}

void TGNGP::SetEpsilon_ne(Double_t val)
{
	fEpsilon_ne = val;
}

void TGNGP::SetRampTime(Double_t val)
{
	fRampTime = val;
}

void TGNGP::SetLambda(UInt_t val)
{
	fLambda = val;
}

void TGNGP::SetT_max(UInt_t val)
{
	fT_max = val;
}

void TGNGP::SetAlpha(Double_t val)
{
	fAlpha = val;
}

void TGNGP::SetBeta(Double_t val)
{
	fBeta = val;
}

void TGNGP::SetMaxUnitCnt(UInt_t val)
{
	fMaxUnitCnt = val;
}

void TGNGP::SetLambdaDelay(UInt_t val)
{
	fLambdaDelay = val;
}

void TGNGP::SetParams(Double_t eps_b, Double_t eps_n, UInt_t lambda, UInt_t tmax, Double_t alpha, Double_t beta, UInt_t maxCnt)
{
	fEpsilon_b  = eps_b;
	fEpsilon_n  = eps_n;
	fLambda     = lambda;
	fT_max      = tmax;
	fAlpha      = alpha;
	fBeta       = beta;
	fMaxUnitCnt = maxCnt;
}

Double_t TGNGP::GetEpsilon_b()
{
	return fEpsilon_b;
}

Double_t TGNGP::GetEpsilon_be()
{
	return fEpsilon_be;
}

Double_t TGNGP::GetEpsilon_n()
{
	return fEpsilon_n;
}

Double_t TGNGP::GetEpsilon_ne()
{
	return fEpsilon_ne;
}

Double_t TGNGP::GetRampTime()
{
	return fRampTime;
}

UInt_t TGNGP::GetLambda()
{
	return fLambda;
}

UInt_t TGNGP::GetT_max()
{
	return fT_max;
}

Double_t TGNGP::GetAlpha()
{
	return fAlpha;
}

Double_t TGNGP::GetBeta()
{
	return fBeta;
}

TGNGDistanceMeasure* TGNGP::GetDistanceMeasure() 
{
	// Returns a reference to the currently used distance measure

	return fDistMeasure; 
} 

UInt_t TGNGP::GetMaxUnitCnt()
{
	return fMaxUnitCnt;
}

UInt_t TGNGP::GetLambdaDelay()
{
	return fLambdaDelay;
}

ULong64_t TGNGP::GetInputCount()
{
	// Returns the number of inputs seen by the GNG so far.
	// Inputs provided to Peek() are not counted.

	return fInpCnt;
}

void TGNGP::AddExtension(TGNGExtension *ext)
{
	// Add the extension "ext" to the GNG.
	// The GNG will "own" the extension, i.e., it will manage the memory

	fExtensions.Add(ext);
}

void TGNGP::RemoveExtension(TGNGExtension *ext)
{
	// Remove the extension from the GNG.
	// The corresponding memory will no longer be managed by the GNG.

	fExtensions.Remove(ext);
	fExtensions.Compress();
}

void TGNGP::RaiseError(UInt_t errorCode, const char* errMsg)
{
	// Method that should be used by extensions to inform about errors.

	printf("ERROR: %s\n", errMsg);

	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		curExt->OnError(0,errorCode,errMsg);
	}		
}

TMatrixD* TGNGP::GetAdjacencyMatrix()
{
	// Returns an adjacency matrix of the GNG units.
	// The returned matrix is owned by the caller.

	if (fTopLayerSize <= 0)
		return 0;

	Int_t total = fTopLayerSize * fMaxUnitCnt;

	TMatrixD *result = new TMatrixD(total,total);
	result->Zero();

	for (Int_t u = 0; u < fTopLayerSize; ++u) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(u));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		for (Int_t i = 0; i < nrOfNodes; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(i));
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			// index of curUnit
			Int_t x = (u * fTopLayerSize) + i;
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
				// index of otherUnit
				Int_t y = ((otherUnit->fCurIdx % fTopLayerSize) * fTopLayerSize) + (otherUnit->fCurIdx / fTopLayerSize);
				(*result)(x,y) = 1.0;
			}
		}

	}

	return result;
}

TMatrixD* TGNGP::GetDistanceMatrix()
{
	// Returns a distance matrix describing the relative distances between the 
	// units prototypes.
	// The returned matrix is owned by the caller.

	if (fTopLayerSize <= 0)
		return 0;

	Int_t total = fTopLayerSize * fMaxUnitCnt;

	TMatrixD *result = new TMatrixD(total,total);
	result->Zero();

	for (Int_t u = 0; u < fTopLayerSize; ++u) {

		TObjArray *topLayerUnit = static_cast<TObjArray*>(fUnits.At(u));
		Int_t nrOfNodes = topLayerUnit->GetEntriesFast();

		for (Int_t i = 0; i < nrOfNodes; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(topLayerUnit->At(i));
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			// index of curUnit
			Int_t x = (u * fTopLayerSize) + i;
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;
				// index of otherUnit
				Int_t y = ((otherUnit->fCurIdx % fTopLayerSize) * fTopLayerSize) + (otherUnit->fCurIdx / fTopLayerSize);
				(*result)(x,y) = fDistMeasure->GetDistance(curUnit->fPrototype,otherUnit->fPrototype,0);
			}
		}

	}

	return result;
}

TH1D* TGNGP::GetDistanceDistribution()
{
	// Returns a histogram of all relative distances between units in the GNG
	// The returned histogram is owned by the caller.

	TMatrixD *dm = GetDistanceMatrix();

	if (dm == 0) return 0;

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = dm->GetNrows();

	TH1D *result = new TH1D("GNGPDistHist",Form("GNGP distance distribution for %s",fDistMeasure->ClassName()),nrOfUnits,0,maxVal+(maxVal / (Double_t)nrOfUnits));
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			result->Fill((*dm)(i,j));

		}

	delete dm;

	return result;
}

TH1D* TGNGP::GetNbDistanceDistribution()
{
	// Returns a histogram of all relative distances between units that are
	// connected by an edge in the GNG.
	// The returned histogram is owned by the caller.

	TMatrixD *dm = GetDistanceMatrix();
	TMatrixD *am = GetAdjacencyMatrix();

	if ((dm == 0) || (am == 0)) {
		if (dm) delete dm;
		if (am) delete am;
		return 0;
	}

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = dm->GetNrows();

	TH1D *result = new TH1D("GNGNBDistHist",Form("GNG neighbor distance distribution for %s",fDistMeasure->ClassName()),nrOfUnits,0,maxVal+(maxVal / (Double_t)nrOfUnits));
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			if ((*am)(i,j) < 0.5) continue;

			result->Fill((*dm)(i,j));

		}

	delete dm;
	delete am;

	return result;
}

void TGNGP::FillDistanceDistribution(TH2D *hist, Double_t xVal)
{
	// Fills in the distance distribution of all relative distances between 
	// units in the GNG at position xVal in the provided 2D histogram.

	TMatrixD *dm = GetDistanceMatrix();

	if (dm == 0) return;

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = dm->GetNrows();

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			hist->Fill(xVal,(*dm)(i,j));

		}

	delete dm;
}

void TGNGP::FillNbDistanceDistribution(TH2D *hist, Double_t xVal)
{
	// Fills in the distance distribution of all relative distances between 
	// units that are connected by an edge in the GNG at position xVal in the 
	// provided 2D histogram.

	TMatrixD *dm = GetDistanceMatrix();
	TMatrixD *am = GetAdjacencyMatrix();

	if ((dm == 0) || (am == 0)) {
		if (dm) delete dm;
		if (am) delete am;
		return;
	}

	Double_t maxVal = dm->Max();

	Int_t nrOfUnits = dm->GetNrows();

	for (Int_t i = 0; i < nrOfUnits; ++i)
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if (i == j) continue;

			if ((*am)(i,j) < 0.5) continue;

			hist->Fill(xVal,(*dm)(i,j));

		}

	delete dm;
	delete am;
}


TH1D* TGNGP::GetGradDistribution()
{
	// Returns a histogram of all unit grades (number of edges incident to a 
	// unit) in the GNG.
	// The returned histogram is owned by the caller.

	TMatrixD *am = GetAdjacencyMatrix();

	if (am == 0) return 0;

	Int_t nrOfUnits = am->GetNrows();

	TH1D *result = new TH1D("GNGGradHist",Form("GNG grad distribution for %s",fDistMeasure->ClassName()),nrOfUnits+1,-0.5,(Double_t)nrOfUnits+0.5);
	result->SetDirectory(0);

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		Double_t grad = 0;
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if ((*am)(i,j) > 0.5)
				grad += 1.0;

		}
		result->Fill(grad);
	}

	delete am;

	return result;	
}

void TGNGP::FillGradDistribution(TH2D *hist, Double_t xVal)
{
	// Fills the distribution of units grades (number of edges incident to a 
	// unit) in the GNG at position xVal in the provided 2D histogram.

	TMatrixD *am = GetAdjacencyMatrix();

	if (am == 0) return;

	Int_t nrOfUnits = am->GetNrows();

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		Double_t grad = 0;
		for (Int_t j = 0; j < nrOfUnits; ++j) {

			if ((*am)(i,j) > 0.5)
				grad += 1.0;

		}
		hist->Fill(xVal,grad);
	}

	delete am;	
}


Int_t TGNGP::GetNrOfExtensions()
{
	// returns the number of registered extensions

	return fExtensions.GetEntriesFast();
}

TGNGExtension* TGNGP::GetExtension(Int_t idx)
{
	// returns the extension with index idx

	return static_cast<TGNGExtension*>(fExtensions.At(idx));
}

TGNGExtension* TGNGP::GetExtension(const char *classname)
{
	// returns the first extension with a matching "classname".

	TString sTerm(classname);
	Int_t nrOfExt = fExtensions.GetEntriesFast();
	for (Int_t i = 0; i < nrOfExt; ++i) {
		TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
		TString curName(curExt->ClassName());
		if (curName == sTerm)
			return curExt;
	}	
	return 0;
}

void TGNGP::Print(Option_t* option) const
{
	// print information about the gng
	// option "g" prints gng information (default)
	// option "e" prints extension information
	// option "p" prints all gng parameters

	TString opt = option;
	opt.ToLower();

	printf("\n");
	
	if ((opt.Length() == 0) || (opt.Contains("g"))) {
		// print general gng information
		printf("Growing Neural Gas Population\n\n");
		printf("Status:   %s\n", fInitialized ? "initialized" : "not initialized");
		printf("# Top Level Units:  %d\n",fUnits.GetEntriesFast());
		printf("# Inputs: %lld\n\n",fInpCnt);
		printf("Distance Measure: %s\n\n", (fDistMeasure != 0) ? fDistMeasure->ClassName() : "null");
	}

	if (opt.Contains("e")) {		
		Int_t nrOfExt = fExtensions.GetEntriesFast();
		printf("Number of extensions: %d\n", nrOfExt);
		for (Int_t i = 0; i < nrOfExt; ++i) {
			TGNGExtension *curExt = static_cast<TGNGExtension*>(fExtensions.At(i));
			printf("Extension %d: %s\n", i, curExt->ClassName());
		}	
		printf("\n");			
	}

	if (opt.Contains("p")) {
		printf("Epsilon b:  %f (adaptation of BMU)\n", fEpsilon_b);
		printf("Epsilon n:  %f (adaptation of neighbors)\n", fEpsilon_n);
		printf("Lambda:     %d (new unit interval)\n", fLambda);
		printf("Alpha:      %f (decrease of accumulated error by insertion)\n", fAlpha);
		printf("Beta:       %f (general decrease of accumulated error)\n", fBeta);
		printf("MaxUnitCnt: %d (max. number of units)\n\n", fMaxUnitCnt);
	}


}

void TGNGP::SetLishWeighting(Bool_t enable)
{
	fLishWeight = enable;
}

void TGNGP::SetLishWeightCenter(Double_t center)
{
	fLishWeightCenter = center;
}

void TGNGP::SetLishWeightSteepness(Double_t steepness)
{
	fLishWeightSteepness = steepness;
}

Bool_t TGNGP::GetLishWeighting()
{
	return fLishWeight;
}

Double_t TGNGP::GetLishWeightCenter()
{
	return fLishWeightCenter;
}

Double_t TGNGP::GetLishWeightSteepness()
{
	return fLishWeightSteepness;
}

Double_t TGNGP::GetEdgeLishStrength(TH1D *hist)
{
	Int_t lishSize = hist->GetNbinsX();
	Double_t cnt = 0;
	Double_t avg = 0;
	for (Int_t l = 1; l <= lishSize; ++l) {
		Double_t bc = 0;
		if ((bc = hist->GetBinContent(l)) > 0) {
			avg += hist->GetBinError(l) / bc;
			cnt += 1;
		} else {
			avg += 1;
			cnt += 1;
		}
	}
	if (cnt > 0) {
		avg /= cnt;
	} else {
		avg = 1;
	}

	return 1.0 - 1.0 / TMath::Power(1.0 + 0.5 * TMath::Exp(-fLishWeightSteepness * (avg - fLishWeightCenter)) ,2) ;

}

TObject* TGNGP::GetLastRawInput()
{
	return fRawInput;
}

void TGNGP::SetLearningInhibition(Bool_t enable)
{
	fUseLearningInhibition = enable;
}

Bool_t TGNGP::GetLearningInhibition()
{
	return fUseLearningInhibition;
}

void TGNGP::SetGamma(Double_t val)
{
	fGamma = val;
}

Double_t TGNGP::GetGamma()
{
	return fGamma;
}

void TGNGP::SetEpsilonRamp(Bool_t enable)
{
	fUseEpsilonRamp = enable;
}

Bool_t TGNGP::GetEpsilonRamp()
{
	return fUseEpsilonRamp;
}

void TGNGP::SetGrowth(Bool_t enable)
{
	fGrowth = enable;
}

Bool_t TGNGP::GetGrowth()
{
	return fGrowth;
}

Int_t TGNGP::GetTopLayerSize()
{
	return fTopLayerSize;
}