#ifndef TGNGP_H
#define TGNGP_H

/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the declaration of the main TGNGP class. See .cpp File for
more information.

*/

#include <TNamed.h>
#include <TObjArray.h>
#include <TMatrixD.h>
#include <TH1D.h>
#include <TH2D.h>
#include <TMap.h>

#include "TGNGDataElement.h"
#include "TGNGDistanceMeasure.h"
#include "TGNGUnit.h"
#include "TGNGEdge.h"
#include "TGNGExtension.h"

class TGNGP : public TNamed 
{
	ClassDef(TGNGP,1);

/*
	Change log:

*/

public:
	enum ErrorCodes { ecNoInit, ecParamNull, ecNoBMU, ecNaN, ecNoMaxErr, ecIncomp, ecExtension, ecFatal };

	TGNGP();
	TGNGP(const char* name, const char* title);
	~TGNGP();

	void Init(Int_t topLayerSize, TGNGDataElement *prototype, TGNGDistanceMeasure *distMeasure);

	void ManualInit(TGNGDistanceMeasure *distMeasure);
	void ManualInitDone();

	TGNGUnit* Feed(const TGNGDataElement *input, Double_t *minDist = 0, TObject *rawInput = 0, Double_t *ratio = 0); // learn input (element will not be owned)

	TGNGUnit* Peek(const TGNGDataElement *input, Double_t *minDist = 0, TObject *rawInput = 0, Double_t *ratio = 0); // cluster input (element will not be owned)

	void SetEpsilon_b(Double_t val);
	void SetEpsilon_be(Double_t val);
	void SetEpsilon_n(Double_t val);
	void SetEpsilon_ne(Double_t val);
	void SetRampTime(Double_t val);
	void SetLambda(UInt_t val);
	void SetT_max(UInt_t val);
	void SetAlpha(Double_t val);
	void SetBeta(Double_t val);
	void SetMaxUnitCnt(UInt_t val);

	void SetLambdaDelay(UInt_t val);

	void SetGamma(Double_t val);

	void SetParams(Double_t eps_b, Double_t eps_n, UInt_t lambda, UInt_t tmax, Double_t alpha, Double_t beta, UInt_t maxCnt);

	void SetLishWeighting(Bool_t enable);
	void SetLishWeightCenter(Double_t center);
	void SetLishWeightSteepness(Double_t steepness);

	Double_t GetEpsilon_b();
	Double_t GetEpsilon_be();
	Double_t GetEpsilon_n();
	Double_t GetEpsilon_ne();
	Double_t GetRampTime();
	UInt_t   GetLambda();
	UInt_t   GetT_max();
	Double_t GetAlpha();
	Double_t GetBeta();
	UInt_t   GetMaxUnitCnt();

	UInt_t GetLambdaDelay();

	Double_t GetGamma();

	Bool_t GetLishWeighting();
	Double_t GetLishWeightCenter();
	Double_t GetLishWeightSteepness();

	TGNGDistanceMeasure* GetDistanceMeasure();

	TMatrixD* GetAdjacencyMatrix();
	TMatrixD* GetDistanceMatrix();

	TH1D* GetDistanceDistribution();
	TH1D* GetNbDistanceDistribution();

	void FillDistanceDistribution(TH2D *hist, Double_t xVal);
	void FillNbDistanceDistribution(TH2D *hist, Double_t xVal);

	TH1D* GetGradDistribution();
	void FillGradDistribution(TH2D *hist, Double_t xVal);

	TObjArray fUnits; //|| array of fTopLayerSize TObjArrays of units of the GNGP

	ULong64_t GetInputCount();

	void AddExtension(TGNGExtension *ext); // will be owned
	void RemoveExtension(TGNGExtension *ext);

	Int_t GetNrOfExtensions();
	TGNGExtension* GetExtension(Int_t idx);
	TGNGExtension* GetExtension(const char *classname);

	void RaiseError(UInt_t errorCode, const char* errMsg);

	void Print(Option_t* option = "") const;

	TMap fAttributes; // custom attributes to this GNG (map is owner)

	TObject* GetLastRawInput();

	void SetLearningInhibition(Bool_t enable);
	Bool_t GetLearningInhibition();

	void SetEpsilonRamp(Bool_t enable);
	Bool_t GetEpsilonRamp();

	void SetGrowth(Bool_t enable);
	Bool_t GetGrowth();

	Int_t GetTopLayerSize();

private:
	void ResetGNG();

	Double_t GetEdgeLishStrength(TH1D *hist);

	// parameters
	Double_t fEpsilon_b;   // adaptation of BMU
	Double_t fEpsilon_be;  // adaptation of BMU at the end of ramp
	Double_t fEpsilon_n;   // adaptation of BMU neighbors
	Double_t fEpsilon_ne;  // adaptation of BMU neighbors at the end of ramp
	Double_t fRampTime;    // time used for ramp
	UInt_t   fLambda;     // new unit interval
	UInt_t   fT_max;      // max. edge age
	Double_t fAlpha;      // decrease of accumulated error by insertion
	Double_t fBeta;       // general decrease of accumulated error 
	UInt_t   fMaxUnitCnt; // max. number of units (default: 100)

	UInt_t   fLambdaDelay; // number of inputs before growing starts

	Double_t fGamma; // decay of unit learning inhibition

	ULong64_t fInpCnt; // input count

	TGNGDistanceMeasure *fDistMeasure; //|| distance measure used

	Bool_t fInitialized; // indicates wether or not the gng was initialized

	TObjArray fExtensions; //|| extensions registered to the GNG

	Bool_t fLishWeight; // adaption of neighbors will be weighted by the average bin error in the edges LISH
	Double_t fLishWeightCenter;    // center of slope on generalized logistic function
	Double_t fLishWeightSteepness; // steepness of slope on generalized logistic function

	TObject *fRawInput; //! last raw input

	Bool_t fUseLearningInhibition; // indicates wether to apply learning inhibition
	Bool_t fUseEpsilonRamp; // indicates wether to change epsilons during training

	Bool_t fGrowth; // indicates wether the GNG will grow/shrink dynamically

	Int_t fTopLayerSize; // Number of fUnit arrays
};



#endif
