/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of the main TNG class.

*/
#include <TMath.h>
#include <TObjString.h>

#include "TNG.h"



TNG::TNG() :
	TNamed("",""),
	fUnits(),
	fEpsilon_start(1.0),
	fEpsilon_end(0.001),
	fLambda_start(50),
	fLambda_end(0.01),
	fRampTime(100000),
	fRampTime2(200000),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fTmp()
{
	// the ng is not initialized

	fUnits.SetOwner(kTRUE);
	fTmp.SetOwner(kFALSE);
}


TNG::TNG(const char* name, const char* title) :
	TNamed(name,title),
	fUnits(),
	fEpsilon_start(1.0),
	fEpsilon_end(0.001),
	fLambda_start(50),
	fLambda_end(0.01),
	fRampTime(100000),
	fRampTime2(200000),
	fInpCnt(0),
	fDistMeasure(0),
	fInitialized(kFALSE),
	fTmp()
{
	// the ng is not initialized

	fUnits.SetOwner(kTRUE);
	fTmp.SetOwner(kFALSE);
}

TNG::~TNG()
{
	// the NG is reset 

	ResetNG();
}

void TNG::Init(Int_t size, TGNGDataElement *prototype, TGNGDistanceMeasure *distMeasure)
{
	// initializes the ng with size units that have random prototypes based on
	// the supplied example prototype and the supplied distance measure.

	if (prototype == 0) {
		RaiseError(TNG::ecParamNull,"Prototype can not be 0!");
		return;
	}

	if (distMeasure == 0) {
		RaiseError(TNG::ecParamNull,"Distance measure can not be 0!");
		return;
	}

	if (fInitialized == kTRUE)
		ResetNG();

	fDistMeasure = distMeasure;

	fLambda_start = (Double_t)size / 2.0;

	TNGUnit *newUnit = new TNGUnit(prototype,0);
	fUnits.Add(newUnit);

	for (Int_t i = 1; i < size; ++i) {
		TGNGDataElement *other = static_cast<TGNGDataElement*>(prototype->Clone());
		other->Randomize();
		newUnit = new TNGUnit(other,i);
		fUnits.Add(newUnit);
	}

	fInitialized = kTRUE;

}


void TNG::ResetNG()
{
	// resets the NG

	if (fInitialized == kFALSE)
		return;

	Int_t unitCnt = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < unitCnt; ++i) {
		TNGUnit *curUnit = static_cast<TNGUnit*>(fUnits.At(i));
	}

	fUnits.Delete();

	fInitialized = kFALSE;
}


TNGUnit* TNG::Feed(const TGNGDataElement *input)
{
	// Inputs a vector into the NG and returns the BMU.
	// The NG is modified.

	//printf("check inputs\n");
	if (fInitialized == kFALSE) {
		RaiseError(TNG::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TNG::ecParamNull,"Input is null!");
		return 0;				
	}

	fTmp.Clear();

	Int_t nrOfUnits = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TNGUnit *curUnit = static_cast<TNGUnit*>(fUnits.At(i));
		curUnit->fCurrentDistance = fDistMeasure->GetDistance(curUnit->fPrototype,input,0);
		fTmp.Add(curUnit);
	}

	fTmp.Sort();

	fInpCnt += 1.0;

	Double_t epsilon = fEpsilon_end;
	Double_t lambda  = fLambda_end;


	if (fInpCnt < fRampTime) {
		epsilon = fEpsilon_start * TMath::Power(fEpsilon_end / fEpsilon_start, fInpCnt / fRampTime);
	}

	if (fInpCnt < fRampTime2) {
		lambda  = fLambda_start * TMath::Power(fLambda_end / fLambda_start, fInpCnt / fRampTime2);
	}


	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TNGUnit *curUnit = static_cast<TNGUnit*>(fTmp.At(i));
		curUnit->fPrototype->Adapt(input, epsilon * TMath::Exp(-(Double_t)i / lambda) );
	}

	//printf("return\n");
	return static_cast<TNGUnit*>(fTmp.At(0));
}

TNGUnit* TNG::Peek(const TGNGDataElement *input)
{
	// Inputs a vector into the NG and returns the BMU.
	// The NG is not modified.

	if (fInitialized == kFALSE) {
		RaiseError(TNG::ecNoInit,"GNG not initialized!");
		return 0;				
	}

	if (input == 0) {
		RaiseError(TNG::ecParamNull,"Input is null!");
		return 0;				
	}

	fInpCnt++;

	fTmp.Clear();

	Int_t nrOfUnits = fUnits.GetEntriesFast();
	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TNGUnit *curUnit = static_cast<TNGUnit*>(fUnits.At(i));
		curUnit->fCurrentDistance = fDistMeasure->GetDistance(curUnit->fPrototype,input,0);
		fTmp.Add(curUnit);
	}

	fTmp.Sort();

	return static_cast<TNGUnit*>(fTmp.At(0));
}


TGNGDistanceMeasure* TNG::GetDistanceMeasure() 
{
	// Returns a reference to the currently used distance measure

	return fDistMeasure; 
} 

ULong64_t TNG::GetInputCount()
{
	// Returns the number of inputs seen by the GNG so far.
	// Inputs provided to Peek() are not counted.

	return fInpCnt;
}

void TNG::RaiseError(UInt_t errorCode, const char* errMsg)
{
}

void TNG::Print(Option_t* option) const
{
	// print information about the ng

	TString opt = option;
	opt.ToLower();

	printf("\n");
	

}

void TNG::SetEpsilon_start(Double_t val)
{
	fEpsilon_start = val;
}

void TNG::SetEpsilon_end(Double_t val)
{
	fEpsilon_end = val;
}

void TNG::SetLambda_start(Double_t val)
{
	fLambda_start = val;
}

void TNG::SetLambda_end(Double_t val)
{
	fLambda_end = val;
}

void TNG::SetRampTime(Double_t val)
{
	fRampTime = val;
}

void TNG::SetRampTime2(Double_t val)
{
	fRampTime2 = val;
}


Double_t TNG::GetEpsilon_start()
{
	return fEpsilon_start;
}

Double_t TNG::GetEpsilon_end()
{
	return fEpsilon_end;
}

Double_t TNG::GetLambda_start()
{
	return fLambda_start;
}

Double_t TNG::GetLambda_end()
{
	return fLambda_end;
}

Double_t TNG::GetRampTime()
{
	return fRampTime;
}

Double_t TNG::GetRampTime2()
{
	return fRampTime2;
}

Double_t TNG::GetCurrentEpsilon()
{
	Double_t epsilon = fEpsilon_end;

	if (fInpCnt < fRampTime) {
		epsilon = fEpsilon_start * TMath::Power(fEpsilon_end / fEpsilon_start, fInpCnt / fRampTime);
	}

	return epsilon;
}

Double_t TNG::GetCurrentLambda()
{
	Double_t lambda  = fLambda_end;

	if (fInpCnt < fRampTime2) {
		lambda  = fLambda_start * TMath::Power(fLambda_end / fLambda_start, fInpCnt / fRampTime2);
	}

	return lambda;
}
