/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a histogram input data class that
is able to convert a TH1D histograms to TGNGVectorData. It can rebin the
histogram to a new number of bins using linear interpolation and it can apply
a logarithm to each bin value.
*/
#include <TRandom.h>
#include <TMath.h>
#include "TGNGHistData.h"

TGNGHistData::TGNGHistData() :
	TGNGVectorData()
{
}

TGNGHistData::TGNGHistData(Int_t bins, TObject *payload) :
	TGNGVectorData(bins,payload)
{
}

TGNGHistData::TGNGHistData(TH1D *hist, Bool_t log, Int_t rebinTo, TObject *payload) :
	TGNGVectorData()
{
	// constructor taking the TH1D hist and copying its values to the member 
	// fVector. If "log" is true, the logarithm of each bin is used. If 
	// "rebinTo" is greater 0 and not equal to the size of the histogram "hist"
	// the values of "hist" will be linearly interpolated to "rebinTo" bins.
	// The histogram "hist" will not be changed.

	fPayload = payload;

	Int_t histDim = hist->GetNbinsX();

	if ((rebinTo <= 0) || (rebinTo == histDim)) {
		fVector.ResizeTo(histDim);
		if (log == kTRUE) {
			for (Int_t i = 0; i < histDim; ++i)
				fVector[i] = TMath::Log(hist->GetBinContent(i+1)+1.0);
		} else {
			for (Int_t i = 0; i < histDim; ++i)
				fVector[i] = hist->GetBinContent(i+1);
		}
	} else { // rebin
		fVector.ResizeTo(rebinTo);

		Double_t step     = (Double_t)histDim / (Double_t)rebinTo;
		Double_t curStep  = 0.0;
		Double_t curMax   = step - 1.0;
		Double_t lastFrac = 1.0;
		if (log == kTRUE) {
			for (Int_t i = 0; i < rebinTo; ++i) {
				Double_t curVal = hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1) * lastFrac;
				curStep += 1.0;
				for (;curStep < curMax; curStep += 1.0) {
					curVal += hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1);
				}
				lastFrac = curStep - curMax;
				curVal += hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1) * (1.0 - lastFrac);
				curMax += step;
				fVector[i] = TMath::Log(curVal+1.0);
			}
		} else {
			for (Int_t i = 0; i < rebinTo; ++i) {
				Double_t curVal = hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1) * lastFrac;
				curStep += 1.0;
				for (;curStep < curMax; curStep += 1.0) {
					curVal += hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1);
				}
				lastFrac = curStep - curMax;
				curVal += hist->GetBinContent((Int_t)(TMath::Floor(curStep))+1) * (1.0 - lastFrac);
				curMax += step;
				fVector[i] = curVal;
			}
		}

	}

}

