/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a minkowski distance that can be used
by the gng to determine the distance between two data elements of type 
TGNGVectorData with equal dimensions. It is also able to handle 
TGNGRecursiveData and thus can be used in an RGNG.

*/
#include <limits>

#include <TMath.h>

#include "../datatypes/TGNGVectorData.h"
#include "../datatypes/TGNGRecursiveData.h"
#include "../core/TGNG.h"
#include "../core/TGNGUnit.h"
#include "TGNGMinkowskiDistance.h"


TGNGMinkowskiDistance::TGNGMinkowskiDistance() :
	TGNGDistanceMeasure(),
	fP(2),
	fLT(TGNGMinkowskiDistance::ltSingle),
	fRM(TGNGMinkowskiDistance::rmPeek)
{
	// default constructor setting the Minkowski parameter to 2
}

TGNGMinkowskiDistance::TGNGMinkowskiDistance(Double_t p, 
                                             TGNGMinkowskiDistance::LinkageType lt,
                                             TGNGMinkowskiDistance::RecursiveMode rm) :
	TGNGDistanceMeasure(),
	fP(p),
	fLT(lt),
	fRM(rm)
{
	// constructor setting the Minkowski parameter to "p" and the linkage type
}	

Double_t TGNGMinkowskiDistance::GetDistance(const TGNGDataElement *a, const TGNGDataElement *b, TGNG *gng)
{
	Double_t result = GetApproxDistance(a,b,gng);

	const TGNGVectorData *av = dynamic_cast<const TGNGVectorData*>(a);
	const TGNGVectorData *bv = dynamic_cast<const TGNGVectorData*>(b);

	if ((av == 0) || (bv == 0)) 
		return result;

	return TMath::Power(result,1.0/fP);
}

Double_t TGNGMinkowskiDistance::GetApproxDistance(const TGNGDataElement *a, const TGNGDataElement *b, TGNG *gng)
{
	Double_t result = 0;

	const TGNGVectorData *av = dynamic_cast<const TGNGVectorData*>(a);
	const TGNGVectorData *bv = dynamic_cast<const TGNGVectorData*>(b);

	if ((av == 0) || (bv == 0)) {

		// could be recursive
		TGNGRecursiveData *ar = dynamic_cast<TGNGRecursiveData*>(const_cast<TGNGDataElement*>(a));
		TGNGRecursiveData *br = dynamic_cast<TGNGRecursiveData*>(const_cast<TGNGDataElement*>(b));

		// fully recursive context
		if ((ar != 0) && (br != 0)) {

			Int_t ar_size = ar->fGNG.fUnits.GetEntriesFast();
			Int_t br_size = br->fGNG.fUnits.GetEntriesFast();

			switch (fLT) {

				case ltSingle : {
					result = numeric_limits<Double_t>::max();

					for (Int_t ai = 0; ai < ar_size; ++ai) {
						TGNGUnit *curUnitA = static_cast<TGNGUnit*>(ar->fGNG.fUnits.At(ai));
						for (Int_t bi = 0; bi < br_size; ++bi) {
							TGNGUnit *curUnitB = static_cast<TGNGUnit*>(br->fGNG.fUnits.At(bi));
							Double_t dist = GetDistance(curUnitA->fPrototype, curUnitB->fPrototype, gng);
							if (dist < result)
								result = dist;
						}
					}

				} break;

				case ltComplete : {
					result = numeric_limits<Double_t>::min();

					for (Int_t ai = 0; ai < ar_size; ++ai) {
						TGNGUnit *curUnitA = static_cast<TGNGUnit*>(ar->fGNG.fUnits.At(ai));
						for (Int_t bi = 0; bi < br_size; ++bi) {
							TGNGUnit *curUnitB = static_cast<TGNGUnit*>(br->fGNG.fUnits.At(bi));
							Double_t dist = GetDistance(curUnitA->fPrototype, curUnitB->fPrototype, gng);
							if (dist > result)
								result = dist;
						}
					}

				} break;

				case ltAverage : {
					Double_t rCnt = 0;
					result = 0;

					for (Int_t ai = 0; ai < ar_size; ++ai) {
						TGNGUnit *curUnitA = static_cast<TGNGUnit*>(ar->fGNG.fUnits.At(ai));
						for (Int_t bi = 0; bi < br_size; ++bi) {
							TGNGUnit *curUnitB = static_cast<TGNGUnit*>(br->fGNG.fUnits.At(bi));
							result += GetDistance(curUnitA->fPrototype, curUnitB->fPrototype, gng);
							rCnt += 1.0;
						}
					}

					if (rCnt > 0)
						result /= rCnt;

				} break;

				default : {
					return TMath::QuietNaN();
				}

			}

			return result;
		}

		// half recursive context a
		if ((av == 0) && (ar != 0) && (bv != 0)) {
			switch (fRM) {
				case rmPeek : {
					ar->fGNG.Peek(bv,&result);
				} break;
				case rmFeed : {
					ar->fGNG.Feed(bv,&result);
				} break;
			}
			
			return result;
		}

		// half recursive context b
		if ((av != 0) && (br != 0) && (bv == 0)) {
			switch (fRM) {
				case rmPeek : {
					br->fGNG.Peek(av,&result);
				} break;
				case rmFeed : {
					br->fGNG.Feed(av,&result);
				} break;
			}
			return result;
		}

		return TMath::QuietNaN();
	}

	if (av->fVector.GetNrows() != bv->fVector.GetNrows())
		return TMath::QuietNaN();

	Int_t nrOfRows = av->fVector.GetNrows();
	const double *avData = av->fVector.GetMatrixArray();
	const double *bvData = bv->fVector.GetMatrixArray();
	for (Int_t i = 0; i < nrOfRows; ++i) {
		// speedup-hack if fP == 2
		if (fP == 2.0) {
			Double_t delta = avData[i] - bvData[i];
			result += delta * delta;
		} else {
			result += TMath::Power(TMath::Abs(avData[i] - bvData[i]),fP);
		}
	}

	return result;
}

Bool_t TGNGMinkowskiDistance::IsCompatible(const TGNGDataElement *a, TGNG*)
{
	return (dynamic_cast<const TGNGVectorData*>(a)    != 0) ||
	       (dynamic_cast<const TGNGRecursiveData*>(a) != 0);
}

void TGNGMinkowskiDistance::SetP(Double_t p)
{
	// sets the Minkowski parameter to "p"
	fP = p;
}

Double_t TGNGMinkowskiDistance::GetP()
{
	// returns the currently used Minkowski parameter
	return fP;
}

void TGNGMinkowskiDistance::SetLinkageType(TGNGMinkowskiDistance::LinkageType lt)
{
	fLT = lt;
}

TGNGMinkowskiDistance::LinkageType TGNGMinkowskiDistance::GetLinkageType()
{
	return fLT;
}

void TGNGMinkowskiDistance::SetRecursiveMode(TGNGMinkowskiDistance::RecursiveMode rm)
{
	fRM = rm;
}

TGNGMinkowskiDistance::RecursiveMode TGNGMinkowskiDistance::GetRecursiveMode()
{
	return fRM;
}
