/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a hierarchical clustering extension
supporting single, complete, and average linkage as well as using the distance
measure provided by the gng or using the average bin error of local input space
histograms as distance.
*/
#include <limits>

#include <TMatrixD.h>

#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"
#include "../../core/TGNGDistanceMeasure.h"

#include "TGNGHierarchicalClusterExt.h"

#include "TGNGHcNode.h"

TGNGHierarchicalClusterExt::TGNGHierarchicalClusterExt() : 
	TGNGExtension(),
	fAttrKey("HCNODE"),
	fInterval(10000),
	fInpCnt(0),
	fTopLevel(),
	fDistanceType(TGNGHierarchicalClusterExt::dtDefault),
	fLinkageType(TGNGHierarchicalClusterExt::ltSingle)
{
	// default constructor using single linkage and the gng-provided distance
	// measure. Clustering is performed every 10000 inputs.

	fTopLevel.SetOwner(kFALSE);
}


void TGNGHierarchicalClusterExt::OnGNGReset(TGNG*)
{
	fInpCnt = 0;
	fTopLevel.Clear();
}


void TGNGHierarchicalClusterExt::OnFeedResult(TGNG *gng, TGNGUnit*, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	if ((++fInpCnt % fInterval) != 0) 
		return;

	// clear existing clustering
	fTopLevel.Clear();
	Int_t nrOfUnits = gng->fUnits.GetEntriesFast();

	if (nrOfUnits <= 0)
		return;

	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));

		TGNGHcNode *curNode = static_cast<TGNGHcNode*>(curUnit->fAttributes.GetValue(&fAttrKey));
		if (curNode == 0) {
			gng->RaiseError(TGNG::ecExtension,"Unit has no attribute \"HCNODE\"!");
			continue;
		}

		if (curNode->fParent != 0)
			delete curNode->fParent; // fParent will be set to 0 by fParent's destructor

		fTopLevel.Add(curNode);
	}

	// build a distance matrix
	TMatrixD distances(nrOfUnits,nrOfUnits);

	distances = std::numeric_limits<Double_t>::max();

	switch (fDistanceType) {

		case TGNGHierarchicalClusterExt::dtLISH : {

			TObjString lishkey("LISH");

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));
				Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
				for (Int_t j = 0; j < edgeCnt; ++j) {
					TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
					TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;

					// get lish
					TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&lishkey));
					if (lish == 0) {
						gng->RaiseError(TGNG::ecExtension,"Edge has no attribute \"LISH\"!");
						continue;
					}
					// get avg bin error
					Int_t lishSize = lish->GetNbinsX();
					Double_t cnt = 0;
					Double_t avg = 0;
					for (Int_t l = 1; l <= lishSize; ++l) {
						Double_t bc = 0;
						if ((bc = lish->GetBinContent(l)) > 0) {
							avg += lish->GetBinError(l) / bc;
							cnt += 1;
						} else {
							// test
							avg += 1;
							cnt += 1;
						}
					}
					if (cnt > 0) {
						avg /= cnt;
					} else {
						avg = 1;
					}

					// use average bin error as distance
					distances(i,otherUnit->fCurIdx) = avg;
				}
			}
			
		} break;

		case TGNGHierarchicalClusterExt::dtDefault : 
		default: {
			TGNGDistanceMeasure *dm = gng->GetDistanceMeasure();

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(gng->fUnits.At(i));
				Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
				for (Int_t j = 0; j < edgeCnt; ++j) {
					TGNGEdge *curEdge   = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
					TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;

					// calculate distance depending on choosen metric, for now use the normal distance measure
					distances(i,otherUnit->fCurIdx) = dm->GetDistance(curUnit->fPrototype,otherUnit->fPrototype,gng);
				}
			}

		} break;

	}

	// begin clustering
	while (fTopLevel.GetEntriesFast() > 1) {
		// find the two entries in fTopLevel with the shortest distance to each other
		Int_t entryA = 0;
		Int_t entryB = 1;

		Double_t minDist = std::numeric_limits<Double_t>::max();

		Int_t nrOfTLEs = fTopLevel.GetEntriesFast();
		for (Int_t i = 0; i < nrOfTLEs; ++i) {
			TGNGHcNode *nodeA = static_cast<TGNGHcNode*>(fTopLevel.At(i));
			for (Int_t j = i+1; j < nrOfTLEs; ++j) {
				TGNGHcNode *nodeB = static_cast<TGNGHcNode*>(fTopLevel.At(j));

				// get distance between clusters depending on metric
				Double_t curDist = std::numeric_limits<Double_t>::max();

				switch (fLinkageType) {

					case TGNGHierarchicalClusterExt::ltSingle : {

						Int_t nrOfUnitsA = nodeA->fLeafList.GetEntriesFast();
						Int_t nrOfUnitsB = nodeB->fLeafList.GetEntriesFast();
						for (Int_t k = 0; k < nrOfUnitsA; ++k) {
							TGNGUnit *curUnitA = static_cast<TGNGUnit*>(nodeA->fLeafList.At(k));
							for (Int_t l = 0; l < nrOfUnitsB; ++l) {
								TGNGUnit *curUnitB = static_cast<TGNGUnit*>(nodeB->fLeafList.At(l));

								Double_t dAB = distances(curUnitA->fCurIdx,curUnitB->fCurIdx);

								if (dAB < curDist)
									curDist = dAB;
							}
						}

					} break;

					case TGNGHierarchicalClusterExt::ltComplete : {

						curDist = 0;

						Int_t nrOfUnitsA = nodeA->fLeafList.GetEntriesFast();
						Int_t nrOfUnitsB = nodeB->fLeafList.GetEntriesFast();
						for (Int_t k = 0; k < nrOfUnitsA; ++k) {
							TGNGUnit *curUnitA = static_cast<TGNGUnit*>(nodeA->fLeafList.At(k));
							for (Int_t l = 0; l < nrOfUnitsB; ++l) {
								TGNGUnit *curUnitB = static_cast<TGNGUnit*>(nodeB->fLeafList.At(l));

								Double_t dAB = distances(curUnitA->fCurIdx,curUnitB->fCurIdx);

								if (dAB > curDist)
									curDist = dAB;
							}
						}
						
					} break;

					case TGNGHierarchicalClusterExt::ltAverage : {

						curDist = 0;
						Double_t nrOfDist = 0;

						Int_t nrOfUnitsA = nodeA->fLeafList.GetEntriesFast();
						Int_t nrOfUnitsB = nodeB->fLeafList.GetEntriesFast();
						for (Int_t k = 0; k < nrOfUnitsA; ++k) {
							TGNGUnit *curUnitA = static_cast<TGNGUnit*>(nodeA->fLeafList.At(k));
							for (Int_t l = 0; l < nrOfUnitsB; ++l) {
								TGNGUnit *curUnitB = static_cast<TGNGUnit*>(nodeB->fLeafList.At(l));

								Double_t dAB = distances(curUnitA->fCurIdx,curUnitB->fCurIdx);

								curDist  += dAB;
								nrOfDist += 1.0;
							}
						}

						if (nrOfDist > 0) {
							curDist /= nrOfDist;
						} else {
							curDist = std::numeric_limits<Double_t>::max();
						}
						
					} break;

					default : {
						gng->RaiseError(TGNG::ecExtension,"Unknown Linkage Type !");
					}

				}


				if (curDist < minDist) {
					minDist = curDist;
					entryA = i;
					entryB = j;
				}

			}
		}

		fTopLevel.Add(new TGNGHcNode(static_cast<TGNGHcNode*>(fTopLevel.At(entryA)),static_cast<TGNGHcNode*>(fTopLevel.At(entryB))));
		fTopLevel.RemoveAt(entryA);
		fTopLevel.RemoveAt(entryB);
		fTopLevel.Compress();
	}

}

void TGNGHierarchicalClusterExt::OnNewUnit(TGNG *gng, TGNGUnit *newUnit)
{
	if (newUnit->fAttributes.FindObject(&fAttrKey) != 0) {
		gng->RaiseError(TGNG::ecExtension,"Unit has already an attribute \"HCNODE\"!");
		return;
	}
	newUnit->fAttributes.Add(new TObjString(fAttrKey),new TGNGHcNode(newUnit));
}


void TGNGHierarchicalClusterExt::SetInterval(Int_t interval)
{
	fInterval = interval;
}

TGNGHcNode* TGNGHierarchicalClusterExt::GetRoot()
{
	if (fTopLevel.GetEntriesFast() == 0)
		return 0;
	return static_cast<TGNGHcNode*>(fTopLevel.At(0));
}

void TGNGHierarchicalClusterExt::SetDistanceType(Int_t dType)
{
	fDistanceType = dType;
}

void TGNGHierarchicalClusterExt::SetLinkageType(Int_t lType)
{
	fLinkageType = lType;
}

