/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a logging mechanism that allows to save
the whole gng network into a root file at regular intervals during learning.
The default interval is to save the gng after every input. For each gng reset
a new tree with an increasing index in its name will be created.
*/
#include <TFile.h>
#include <TTree.h>
#include <TObjArray.h>

#include "../../core/TGNGEdge.h"
#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"

#include "TGNGUnitLoggerExt.h"


TGNGUnitLoggerExt::TGNGUnitLoggerExt() : 
	TGNGExtension(),
	fLogFile(0),
	fLogTree(0),
	fNextTreeIdx(0),
	fLogInterval(1),
	fInpCnt(0),
	fUnits(0)
{}

TGNGUnitLoggerExt::TGNGUnitLoggerExt(const char *file) : 
	TGNGExtension(),
	fLogFile(0),
	fLogTree(0),
	fNextTreeIdx(0),
	fLogInterval(1),
	fInpCnt(0),
	fUnits(0)
{
	SetLogFile(file);
}

TGNGUnitLoggerExt::~TGNGUnitLoggerExt()
{
	CloseLogFile();	
}

void TGNGUnitLoggerExt::SetLogFile(const char *file)
{
	// set a new logfile. A previously opened logfile will be closed.
	CloseLogFile();
	fLogFile = TFile::Open(file,"recreate");
}

void TGNGUnitLoggerExt::CloseLogFile()
{
	if (fLogFile) {
		fLogFile->Close();
		delete fLogFile;
		fLogFile = 0;
		fLogTree = 0; // an existing tree will be freed by logfile close
	}
}

void  TGNGUnitLoggerExt::SetLogInterval(Int_t step)
{
	fLogInterval = step;
}

Int_t TGNGUnitLoggerExt::GetLogInterval()
{
	return fLogInterval;
}


void TGNGUnitLoggerExt::OnGNGInit(TGNG *gng)
{
	if (fLogFile == 0) {
		gng->RaiseError(TGNG::ecExtension,"fLogFile is null on init! Fatal!");
		return;
	}
	if (fLogTree) {
		gng->RaiseError(TGNG::ecExtension,"fLogTree is non null on init!");
		delete fLogTree;
		fLogTree = 0;		
	}
	fLogTree = new TTree(Form("unitTree_%d",fNextTreeIdx++),"units of a gng");
	fLogTree->Branch("inputCnt",&fInpCnt,"idx/l");
	fLogTree->Branch("units","TObjArray",&fUnits,32000,0);

	fInpCnt = 0;
	fUnits  = &(gng->fUnits);

	fLogTree->Fill();
}

void TGNGUnitLoggerExt::OnGNGReset(TGNG*)
{
	if (fLogTree) {
		fLogTree->Write();
		delete fLogTree;
		fLogTree = 0;
	}	
}

void TGNGUnitLoggerExt::OnPeekResult(TGNG *gng, TGNGUnit*, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	LogUnits(gng);
}
void TGNGUnitLoggerExt::OnFeedResult(TGNG *gng, TGNGUnit*, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	LogUnits(gng);
}

void TGNGUnitLoggerExt::LogUnits(TGNG *gng)
{
	fInpCnt = gng->GetInputCount();
	fUnits  = &(gng->fUnits);
	if ((fLogFile != 0) && (fLogTree != 0) && (fInpCnt % fLogInterval == 0)) {
		fLogTree->Fill();		
	}
}