/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a hierarchical clustering visualization
based on simple dendrograms. It uses the hierarchical clustering generated by
TGNGHierarchicalClusterExt.
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>
#include <TRandom.h>
#include <TVector2.h>
#include <TMath.h>
#include <TStyle.h>
#include <TPad.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../core/TGNGDataElement.h"
#include "../functional/TGNGIdExt.h"
#include "../cluster/TGNGHierarchicalClusterExt.h"
#include "../cluster/TGNGHcNode.h"

#include "TGNGHiClusterVis.h"

TGNGHiClusterVis::TGNGHiClusterVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fCurrentPad(0),
	fUpdateInterval(1),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kBlue+2),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fTableTextTopOffsetX(0.0),
	fTableTextTopOffsetY(-0.1),
	fTableTextColor(kBlue+2),
	fTableTextFont(100),
	fTableTextSize(0.8),
	fCandidates()
{

}


void TGNGHiClusterVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e  - do not draw lines
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGHiClusterVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e   - do not draw lines
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	if ((fCurrentGNG == 0) || (fCurrentPad == 0) || (fCurrentHC == 0)) {
		fLastUpdate = TTimeStamp();
		return;
	}

	TGNGHcNode *rn = fCurrentHC->GetRoot();

	if (rn == 0) {
		fLastUpdate = TTimeStamp();
		return;
	}	

	Int_t nrOfUnits = rn->fLeafList.GetEntriesFast();
	Int_t depth     = rn->fDepth;

	if (nrOfUnits == 0) {
		fLastUpdate = TTimeStamp();
		return;
	}

	Double_t dimX = (Double_t)(nrOfUnits);
	Double_t dimY = TMath::Max((Double_t)(depth + 2),15.0);

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();


	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(21);
		SetTextColor(fTableTextColor);
		SetTextFont(fTableTextFont);
		SetTextSize((1.0 / dimX) * fTableTextSize);
		TAttText::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(rn->fLeafList.At(i));

			Int_t idx = -1;
			// get unit id
			if (opt.Contains("tr")) {
				idx = fCurrentGNG->fUnits.IndexOf(curUnit);
			} else {
				TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
				if (uid == 0)
					continue;
				idx = uid->fID;
			}

			Double_t x1 = ((Double_t)(i+fTableTextTopOffsetX) + 0.5) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((1.0 + fTableTextTopOffsetY) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",idx);
			fCurrentPad->PaintText(x1,y1,buffer);
		}
	}

	// draw lines
	if (!opt.Contains("!e")) {

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		fCandidates.Add(rn);

		fCurrentPad->PaintLine(0.5 * mw + left_m,1.0 - ((depth+2) / dimY * mh) - top_m,
			                   0.5 * mw + left_m,1.0 - ((depth+1) / dimY * mh) - top_m);

		Int_t entryCnt = 0;
		while ((entryCnt = fCandidates.GetEntriesFast()) > 0) {
			TGNGHcNode *curNode = static_cast<TGNGHcNode*>(fCandidates.RemoveAt(entryCnt-1));

			if ((curNode->fChildA == 0) || (curNode->fChildB == 0))
				continue;

			Double_t leafCntA = (Double_t)(curNode->fChildA->fLeafList.GetEntriesFast());
			Double_t leafCntB = (Double_t)(curNode->fChildB->fLeafList.GetEntriesFast());

			Double_t startPos = (Double_t)(rn->fLeafList.IndexOf(curNode->fLeafList.At(0)));

			Double_t xA = (startPos + (leafCntA / 2.0)) / dimX * mw + left_m;
			Double_t yA = 1.0-((Double_t)(curNode->fChildA->fDepth + 1) / dimY * mh) - top_m;

			Double_t xB = (startPos + leafCntA + (leafCntB / 2.0)) / dimX * mw + left_m;
			Double_t yB = 1.0-((Double_t)(curNode->fChildB->fDepth + 1) / dimY * mh) - top_m;

			//Double_t x1 = (xA + xB) / 2.0;
			Double_t y1 = 1.0-((Double_t)(curNode->fDepth + 1) / dimY * mh) - top_m;

			fCurrentPad->PaintLine(xA,y1,xB,y1);
			fCurrentPad->PaintLine(xA,y1,xA,yA);
			fCurrentPad->PaintLine(xB,y1,xB,yB);

			fCandidates.Add(curNode->fChildA);
			fCandidates.Add(curNode->fChildB);
		}

		SetLineColor(kGray+1);
		SetLineStyle(1);
		SetLineWidth(1);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {

			Double_t x1 = i+1;
			Double_t y1 = 0;

			Double_t x2 = i+1;
			Double_t y2 = 1.5;

			fCurrentPad->PaintLine(x1 / dimX * mw + left_m,
							1.0 - (y1 / dimY * mh) - top_m,
				            x2 / dimX * mw + left_m,
				            1.0 - (y2 / dimY * mh) - top_m);
		}


	}
	
	fLastUpdate = TTimeStamp();

}

void TGNGHiClusterVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	if (gng != fCurrentGNG) {
		fCurrentGNG = gng;
		fCurrentHC  = static_cast<TGNGHierarchicalClusterExt*>(fCurrentGNG->GetExtension("TGNGHierarchicalClusterExt"));

		if (fCurrentHC == 0) {
			fCurrentGNG->RaiseError(TGNG::ecExtension,"GNG has no extension \"TGNGHierarchicalClusterExt\"!");
		}
	}
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGHiClusterVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	if (gng != fCurrentGNG) {
		fCurrentGNG = gng;
		fCurrentHC  = static_cast<TGNGHierarchicalClusterExt*>(fCurrentGNG->GetExtension("TGNGHierarchicalClusterExt"));

		if (fCurrentHC == 0) {
			fCurrentGNG->RaiseError(TGNG::ecExtension,"GNG has no extension \"TGNGHierarchicalClusterExt\"!");
		}
	}
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGHiClusterVis::OnGNGReset(TGNG*)
{
	fCurrentGNG = 0;
	fCurrentHC  = 0;
	fCurrentBMU = 0;
}

void TGNGHiClusterVis::SetUpdateInterval(Double_t sec)
{
	fUpdateInterval = sec;
}


void TGNGHiClusterVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGHiClusterVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGHiClusterVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}

void TGNGHiClusterVis::SetTableTextTopOffsetX(Double_t ox)
{
	fTableTextTopOffsetX = ox;
}

void TGNGHiClusterVis::SetTableTextTopOffsetY(Double_t oy)
{
	fTableTextTopOffsetY = oy;
}

void TGNGHiClusterVis::SetTableTextColor(Color_t color)
{
	fTableTextColor = color;
}

void TGNGHiClusterVis::SetTableTextFont(Font_t font)
{
	fTableTextFont = font;
}

void TGNGHiClusterVis::SetTableTextSize(Float_t size)
{
	fTableTextSize = size;
}

void TGNGHiClusterVis::UpdateVis()
{
	TTimeStamp current;
	Double_t delta = current - fLastUpdate;
	if (delta < fUpdateInterval)
		return;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	

	//fLastUpdate = current;
}

