/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a gng visualization that plots the gng
units in an incidence matrix. The matrix can be colored by distance, can 
indicate edges between units, and can draw a small miniature local input space
histogram for each edge. See Draw()-Method for more info.
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>
#include <TRandom.h>
#include <TVector2.h>
#include <TMath.h>
#include <TStyle.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../datatypes/TGNGVectorData.h"
#include "../functional/TGNGIdExt.h"

#include "TGNGMatrixVis.h"

TGNGMatrixVis::TGNGMatrixVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kGray),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fAdjMarkerColor(kBlack),
	fAdjMarkerSize(1),
	fAdjMarkerStyle(34),
	fTableTextTopOffsetX(0.0),
	fTableTextTopOffsetY(-0.1),
	fTableTextLeftOffsetX(-0.1),
	fTableTextLeftOffsetY(0.0),
	fTableTextColor(kBlue+2),
	fTableTextFont(100),
	fTableTextSize(0.8),
	fMinX(-1),
	fMaxX(-1),
	fMinY(-1),
	fMaxY(-1),
	fLishBgColor(kBlack),
	fLishBgColorAlpha(1.0),
	fLishBgStyle(1001),
	fLishLineColor(kOrange-9),
	fLishErrorLineColor(kRed),
	fLishLineWidth(1),
	fLishLineStyle(1),
	fPaletteA(51)
{
	SetPalette(fPaletteA);
}


void TGNGMatrixVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e  - do not draw matrix lines
	// col - distances as colored boxes
	// ecol - distances as colored boxes, only edges
	// m    - mark edges
	// li   - draw lish (only sensible if few nodes are viewed)
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGMatrixVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e   - do not draw matrix lines
	// col  - distances as colored boxes
	// ecol - distances as colored boxes, only edges
	// m    - mark edges
	// li   - draw lish (only sensible if few nodes are viewed)
	// t    - print node IDs 
	// tr   - print raw node idx (position in TObjArray) instead of IDs 

	if ((fCurrentGNG == 0) || (fCurrentPad == 0))
		return;

	Int_t nrOfUnits = fCurrentGNG->fUnits.GetEntriesFast();	

	Int_t maxXUnit = nrOfUnits - 1;
	if ((fMaxX > 0) && (fMaxX < nrOfUnits))
		maxXUnit = fMaxX;

	Int_t minXUnit = 0;
	if ((fMinX > 0) && (fMinX <= fMaxX))
		minXUnit = fMinX;

	Int_t maxYUnit = nrOfUnits - 1;
	if ((fMaxY > 0) && (fMaxY < nrOfUnits))
		maxYUnit = fMaxY;

	Int_t minYUnit = 0;
	if ((fMinY > 0) && (fMinY <= fMaxY))
		minYUnit = fMinY;

	Double_t dimX = (Double_t)(maxXUnit - minXUnit + 1) + 2.0;
	Double_t dimY = (Double_t)(maxYUnit - minYUnit + 1) + 2.0;

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(21);
		SetTextColor(fTableTextColor);
		SetTextFont(fTableTextFont);
		SetTextSize((1.0 / dimX) * fTableTextSize);
		TAttText::Modify();

		for (Int_t i = minXUnit; i <= maxXUnit; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t idx = -1;
			// get unit id
			if (opt.Contains("tr")) {
				idx = i;
			} else {
				TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
				if (uid == 0)
					continue;
				idx = uid->fID;
			}

			Double_t x1 = ((Double_t)(i-minXUnit+1+fTableTextTopOffsetX) + 0.5) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((1.0+fTableTextTopOffsetY) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",idx);
			fCurrentPad->PaintText(x1,y1,buffer);

		}

		SetTextAlign(32);
		SetTextSize((1.0 / dimY) * fTableTextSize);
		TAttText::Modify();

		for (Int_t i = minYUnit; i <= maxYUnit; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t idx = -1;
			// get unit id
			if (opt.Contains("tr")) {
				idx = i;
			} else {
				TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
				if (uid == 0)
					continue;
				idx = uid->fID;
			}

			Double_t x1 = (1.0+fTableTextLeftOffsetX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - (((Double_t)(i-minYUnit+1+fTableTextLeftOffsetY) + 0.5) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",idx);
			fCurrentPad->PaintText(x1,y1,buffer);

		}

	}

	// draw distances as colors
	if (opt.Contains("ecol")) {
		TMatrixD* distances = fCurrentGNG->GetDistanceMatrix();

		SetFillStyle(1001);

		Double_t maxDist  = distances->Max();
		//Int_t palSize = gStyle->GetNumberOfColors();
		Int_t palSize = fPalA.size();
		Double_t colScale = maxDist / (Double_t)(palSize);
		if (colScale <= 0)
			colScale = 1;

		for (Int_t i = minXUnit; i <= maxXUnit; ++i) {

			Double_t x1 = (Double_t)(i-minXUnit+1) / dimX * mw + left_m;
			Double_t x2 = (Double_t)(i-minXUnit+2) / dimX * mw + left_m;

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;

				Int_t otherIdx = fCurrentGNG->fUnits.IndexOf(otherUnit);

				if ((otherIdx < minYUnit) || (otherIdx > maxYUnit))
					continue;

				Double_t y1 = 1.0 - ((Double_t)(otherIdx-minYUnit+1) / dimY * mh) - top_m;
				Double_t y2 = 1.0 - ((Double_t)(otherIdx-minYUnit+2) / dimY * mh) - top_m;

				Int_t colorIdx = TMath::Floor(((*distances)(i,otherIdx) / colScale) + 0.5);
				if (colorIdx >= palSize)
					colorIdx = palSize-1;

				//SetFillColor(gStyle->GetColorPalette(colorIdx));
				SetFillColor(fPalA[colorIdx]);
				TAttFill::Modify();

				fCurrentPad->PaintBox(x1,y1,x2,y2);

			}
		}

		delete distances;
	} else if (opt.Contains("col")) {
		TMatrixD* distances = fCurrentGNG->GetDistanceMatrix();

		SetFillStyle(1001);

		Double_t maxDist  = distances->Max();

		//printf("\nMaxDist: %f                    \n", maxDist);

		//Int_t palSize = gStyle->GetNumberOfColors();
		Int_t palSize = fPalA.size();
		Double_t colScale = maxDist / (Double_t)(palSize);
		if (colScale <= 0)
			colScale = 1;

		for (Int_t i = minXUnit; i <= maxXUnit; ++i) {

			Double_t x1 = (Double_t)(i-minXUnit+1) / dimX * mw + left_m;
			Double_t x2 = (Double_t)(i-minXUnit+2) / dimX * mw + left_m;

			for (Int_t j = minYUnit; j <= maxYUnit; ++j) {
				Double_t y1 = 1.0 - ((Double_t)(j-minYUnit+1) / dimY * mh) - top_m;
				Double_t y2 = 1.0 - ((Double_t)(j-minYUnit+2) / dimY * mh) - top_m;

				Int_t colorIdx = TMath::Floor(((*distances)(i,j) / colScale) + 0.5);
				if (colorIdx >= palSize)
					colorIdx = palSize-1;

				//SetFillColor(gStyle->GetColorPalette(colorIdx));
				SetFillColor(fPalA[colorIdx]);
				TAttFill::Modify();

				fCurrentPad->PaintBox(x1,y1,x2,y2);
			}

		}

		delete distances;
	}

	// draw matrix lines
	if (!opt.Contains("!e")) {

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		for (Int_t i = minXUnit; i <= maxXUnit + 1; ++i) {

			Double_t x1 = i-minXUnit+1;
			Double_t y1 = 0;

			Double_t x2 = i-minXUnit+1;
			Double_t y2 = dimY;

			fCurrentPad->PaintLine(x1 / dimX * mw + left_m,
							1.0 - (y1 / dimY * mh) - top_m,
				            x2 / dimX * mw + left_m,
				            1.0 - (y2 / dimY * mh) - top_m);
		}
		for (Int_t i = minYUnit; i <= maxYUnit + 1; ++i) {
			Double_t x1 = 0;
			Double_t y1 = i-minYUnit+1;

			Double_t x2 = dimX;
			Double_t y2 = i-minYUnit+1;

			fCurrentPad->PaintLine(x1 / dimX * mw + left_m,
							1.0 - (y1 / dimY * mh) - top_m,
				            x2 / dimX * mw + left_m,
				            1.0 - (y2 / dimY * mh) - top_m);

		}

	}

	if (opt.Contains("m")) {

		SetMarkerColor(fAdjMarkerColor);
		SetMarkerSize(fAdjMarkerSize);
		SetMarkerStyle(fAdjMarkerStyle);
		TAttMarker::Modify();				

		for (Int_t i = minXUnit; i <= maxXUnit; ++i) {

			Double_t x1 = ((Double_t)(i-minXUnit+1) + 0.5) / dimX * mw + left_m;

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;

				Int_t otherIdx = fCurrentGNG->fUnits.IndexOf(otherUnit);

				if ((otherIdx < minYUnit) || (otherIdx > maxYUnit))
					continue;

				Double_t y1 = 1.0 - (((Double_t)(otherIdx-minYUnit+1) + 0.5) / dimY * mh) - top_m;

				fCurrentPad->PaintPolyMarker(1,&x1,&y1);

			}
		}

	}

	
	// draw LISHs	
	if (opt.Contains("li")) {
		TObjString key("LISH");

		SetFillColorAlpha(fLishBgColor,fLishBgColorAlpha);
		SetFillStyle(fLishBgStyle);
		TAttFill::Modify();

		SetLineColor(fLishLineColor);
		SetLineStyle(fLishLineStyle);
		SetLineWidth(fLishLineWidth);
		TAttLine::Modify();

		for (Int_t i = minXUnit; i <= maxXUnit; ++i) {

			Double_t outerXMargin = 0.1;
			Double_t outerYMargin = 0.1;

			Double_t x1 = (Double_t)(i-minXUnit+1+outerXMargin) / dimX * mw + left_m;
			Double_t x2 = (Double_t)(i-minXUnit+2-outerXMargin) / dimX * mw + left_m;

			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));

			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));

				TGNGUnit *otherUnit = (curEdge->fUnitA != curUnit) ? curEdge->fUnitA : curEdge->fUnitB;

				Int_t otherIdx = fCurrentGNG->fUnits.IndexOf(otherUnit);

				if ((otherIdx < minYUnit) || (otherIdx > maxYUnit))
					continue;

				Double_t y1 = 1.0 - ((Double_t)(otherIdx-minYUnit+1+outerYMargin) / dimY * mh) - top_m;
				Double_t y2 = 1.0 - ((Double_t)(otherIdx-minYUnit+2-outerYMargin) / dimY * mh) - top_m;

				// get lish
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&key));
				if (lish == 0) {
					SetFillColor(kRed);
					TAttFill::Modify();
					fCurrentPad->PaintBox(x1,y1,x2,y2);
					SetFillColorAlpha(fLishBgColor,fLishBgColorAlpha);
					TAttFill::Modify();
					continue;
				}


				// paint black box
				if (fLishBgColorAlpha > 0)
					fCurrentPad->PaintBox(x1,y1,x2,y2);

				Double_t bw = x2-x1;
				Double_t bh = y2-y1;

				Double_t xmargin = 0.002;
				Double_t ymargin = 0.002;

				// draw mini histogram
				Int_t nrOfBins  = lish->GetNbinsX();
				Int_t maxBin    = lish->GetMaximumBin();
				Double_t maxVal = lish->GetBinContent(maxBin);
				if (maxVal == 0.0) maxVal = 1.0;
				Double_t xFrac  = (bw - 2*xmargin)  / (Double_t)nrOfBins;
				Double_t yFact  = (bh + 2*ymargin)  / maxVal; // bh is negative

				Bool_t reverse = (curEdge->fUnitA != curUnit);

				SetLineColor(fLishLineColor);
				TAttLine::Modify();
				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = !reverse ? lish->GetBinContent(k+1) : 
											     lish->GetBinContent(nrOfBins-k);
					// draw line
					Double_t tmpX = x1 + (Double_t)k * xFrac + xFrac / 2.0 + xmargin;
					fCurrentPad->PaintLine(tmpX,y2 + ymargin,tmpX,y2 - (curVal * yFact) + ymargin);	
				}

				SetLineColor(fLishErrorLineColor);
				TAttLine::Modify();
				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = !reverse ? lish->GetBinError(k+1) : 
											     lish->GetBinError(nrOfBins-k);
					// draw line
					Double_t tmpX = x1 + (Double_t)k * xFrac + xFrac / 2.0 + xmargin;
					fCurrentPad->PaintLine(tmpX,y2 + ymargin,tmpX,y2 - (curVal * yFact) + ymargin);	
				}

			}
		}


	}
	
	fLastUpdate = TTimeStamp();

}

void TGNGMatrixVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGMatrixVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t, TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGMatrixVis::SetUpdateInterval(Double_t sec)
{
	fUpdateInterval = sec;
}

void TGNGMatrixVis::SetRange(Int_t minX, Int_t maxX, Int_t minY, Int_t maxY)
{
	// sets the range of units (by unit index) that are visible in the matrix

	fMinX = minX;
	fMaxX = maxX;
	fMinY = minY;
	fMaxY = maxY;
}

void TGNGMatrixVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGMatrixVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGMatrixVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}


void TGNGMatrixVis::SetAdjMarkerColor(Color_t color)
{
	fAdjMarkerColor = color;
}

void TGNGMatrixVis::SetAdjMarkerSize(Size_t size)
{
	fAdjMarkerSize = size;
}

void TGNGMatrixVis::SetAdjMarkerStyle(Style_t style)
{
	fAdjMarkerStyle = style;
}

void TGNGMatrixVis::SetTableTextTopOffsetX(Double_t ox)
{
	fTableTextTopOffsetX = ox;
}

void TGNGMatrixVis::SetTableTextTopOffsetY(Double_t oy)
{
	fTableTextTopOffsetY = oy;
}

void TGNGMatrixVis::SetTableTextLeftOffsetX(Double_t ox)
{
	fTableTextLeftOffsetX = ox;
}

void TGNGMatrixVis::SetTableTextLeftOffsetY(Double_t oy)
{
	fTableTextLeftOffsetY = oy;
}

void TGNGMatrixVis::SetTableTextColor(Color_t color)
{
	fTableTextColor = color;
}

void TGNGMatrixVis::SetTableTextFont(Font_t font)
{
	fTableTextFont = font;
}

void TGNGMatrixVis::SetTableTextSize(Float_t size)
{
	fTableTextSize = size;
}

void TGNGMatrixVis::SetLishBgColor(Color_t color, Float_t alpha)
{
	fLishBgColor = color;
	fLishBgColorAlpha = alpha;
}

void TGNGMatrixVis::SetLishBgStyle(Style_t style)
{
	fLishBgStyle = style;
}

void TGNGMatrixVis::SetLishLineColor(Color_t color)
{
	fLishLineColor = color;
}

void TGNGMatrixVis::SetLishErrorLineColor(Color_t color)
{
	fLishErrorLineColor = color;
}

void TGNGMatrixVis::SetLishLineWidth(Width_t width)
{
	fLishLineWidth = width;
}

void TGNGMatrixVis::SetLishLineStyle(Style_t style)
{
	fLishLineStyle = style;
}



void TGNGMatrixVis::UpdateVis()
{
	TTimeStamp current;
	Double_t delta = current - fLastUpdate;
	if (delta < fUpdateInterval)
		return;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	

	//fLastUpdate = current;
}

void TGNGMatrixVis::SetPalette(Int_t a)
{
	fPaletteA = a;

	fPalA.clear();

	gStyle->SetPalette(fPaletteA);
	Int_t palSize = gStyle->GetNumberOfColors();
	for (Int_t i = 0; i < palSize; ++i)
		fPalA.push_back(gStyle->GetColorPalette(i));

}
