/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a simple GNG network visualization if
the GNG contains only two-dimensional input vectors of type TGNGVectorData
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../datatypes/TGNGVectorData.h"
#include "../../datatypes/TGNGRecursiveData.h"
#include "../functional/TGNGIdExt.h"

#include "TGNGRecursive2DVis.h"

const Color_t TGNGRecursive2DVis::fColors[] = {kBlue,kRed,kGreen,kOrange,
	kMagenta,kGreen+3,kAzure+1,kOrange+1,kCyan,kGray};

const Int_t TGNGRecursive2DVis::fNrOfColors = 10;

TGNGRecursive2DVis::TGNGRecursive2DVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fUnitSize(1),
	fUnitStyle(21),
	fUnitTextOffsetX(0.015),
	fUnitTextOffsetY(-0.015),
	fUnitTextColor(kBlue+2),
	fUnitTextFont(100),
	fUnitTextSize(0.02),
	fAutoRange(kTRUE),
	fXRangeMin(0.0),
	fXRangeMax(1.0),
	fYRangeMin(0.0),
	fYRangeMax(1.0)

{}


void TGNGRecursive2DVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGRecursive2DVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	if ((fCurrentGNG == 0) || (fCurrentPad == 0))
		return;

	Int_t nrOfTLUnits = fCurrentGNG->fUnits.GetEntriesFast();

	// get space dimension
	Double_t minX = numeric_limits<Double_t>::max();
	Double_t maxX = numeric_limits<Double_t>::min();
	Double_t minY = numeric_limits<Double_t>::max();
	Double_t maxY = numeric_limits<Double_t>::min();

	if (fAutoRange)	{

		for (Int_t tlu = 0; tlu < nrOfTLUnits; ++tlu) {
			TGNGUnit *curTLU = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(tlu));
			TGNGRecursiveData *rData = dynamic_cast<TGNGRecursiveData*>(curTLU->fPrototype);
			if (rData == 0) {
				fCurrentGNG->RaiseError(TGNG::ecIncomp,"Top Level Data is not of type TGNGRecursiveData!");
				return;
			}

			TGNG *fLocalGNG = &(rData->fGNG);

			Int_t nrOfUnits = fLocalGNG->fUnits.GetEntriesFast();
			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit       *curUnit = static_cast<TGNGUnit*>(fLocalGNG->fUnits.At(i));
				TGNGVectorData *curData = dynamic_cast<TGNGVectorData*>(curUnit->fPrototype);
				if (curData == 0) {
					fCurrentGNG->RaiseError(TGNG::ecIncomp,"Data is not of type TGNGVectorData!");
					return;
				}
				/*
				if (curData->fVector.GetNrows() != 2) {
					fCurrentGNG->RaiseError(TGNG::ecIncomp,"Data has dimension other than 2!");
					return;			
				}
				*/
				Double_t curX = curData->fVector[0];
				Double_t curY = curData->fVector[1];

				Int_t rowCnt = curData->fVector.GetNrows();
				if (rowCnt > 2) {
					// determine x and y as maximum values in lower and upper half
					Double_t mxv = 0;
					Int_t    mxi = 0;
					Double_t myv = 0;
					Int_t    myi = rowCnt / 2;
					for (Int_t r = 0; r < rowCnt / 2; ++r) {
						if (curData->fVector[r] > mxv) {
							mxv = curData->fVector[r];
							mxi = r;
						}
						Int_t ru = r + (rowCnt / 2);
						if (curData->fVector[ru] > myv) {
							myv = curData->fVector[ru];
							myi = ru;
						}					
					}
					myi -= rowCnt / 2;
					curX = (Double_t)mxi / (Double_t)(rowCnt / 2);
					curY = (Double_t)myi / (Double_t)(rowCnt / 2);
				}

				if (curX < minX) minX = curX;
				if (curX > maxX) maxX = curX;
				if (curY < minY) minY = curY;
				if (curY > maxY) maxY = curY;
			}
		}

	} else {
		minX = fXRangeMin;
		maxX = fXRangeMax;
		minY = fYRangeMin;
		maxY = fYRangeMax;
	}

	Double_t dimX = maxX - minX;
	Double_t dimY = maxY - minY;

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw edges
	if (!opt.Contains("!e")) {

		for (Int_t tlu = 0; tlu < nrOfTLUnits; ++tlu) {
			TGNGUnit *curTLU = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(tlu));
			TGNGRecursiveData *rData = dynamic_cast<TGNGRecursiveData*>(curTLU->fPrototype);

			TGNG *fLocalGNG = &(rData->fGNG);

			Int_t nrOfUnits = fLocalGNG->fUnits.GetEntriesFast();

			Int_t colIdx = tlu;
			if (colIdx >= fNrOfColors) colIdx = fNrOfColors - 1;

			SetLineColor(fColors[colIdx]);
			SetLineStyle(fEdgeStyle);
			SetLineWidth(fEdgeWidth);
			TAttLine::Modify();

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fLocalGNG->fUnits.At(i));
				// get unit pos
				TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
				Double_t x1 = curData->fVector[0];
				Double_t y1 = curData->fVector[1];

				Int_t rowCnt = curData->fVector.GetNrows();
				if (rowCnt > 2) {
					// determine x and y as maximum values in lower and upper half
					Double_t mxv = 0;
					Int_t    mxi = 0;
					Double_t myv = 0;
					Int_t    myi = rowCnt / 2;
					for (Int_t r = 0; r < rowCnt / 2; ++r) {
						if (curData->fVector[r] > mxv) {
							mxv = curData->fVector[r];
							mxi = r;
						}
						Int_t ru = r + (rowCnt / 2);
						if (curData->fVector[ru] > myv) {
							myv = curData->fVector[ru];
							myi = ru;
						}					
					}
					myi -= rowCnt / 2;
					x1 = (Double_t)mxi / (Double_t)(rowCnt / 2);
					y1 = (Double_t)myi / (Double_t)(rowCnt / 2);
				}

				Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
				for (Int_t j = 0; j < edgeCnt; ++j) {
					TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
					if (curEdge->fUnitA != curUnit)
						continue;
					// get other units positions
					TGNGVectorData *otherData = static_cast<TGNGVectorData*>(curEdge->fUnitB->fPrototype);
					Double_t x2 = otherData->fVector[0];
					Double_t y2 = otherData->fVector[1];

					Int_t rowCnt = otherData->fVector.GetNrows();
					if (rowCnt > 2) {
						// determine x and y as maximum values in lower and upper half
						Double_t mxv = 0;
						Int_t    mxi = 0;
						Double_t myv = 0;
						Int_t    myi = rowCnt / 2;
						for (Int_t r = 0; r < rowCnt / 2; ++r) {
							if (otherData->fVector[r] > mxv) {
								mxv = otherData->fVector[r];
								mxi = r;
							}
							Int_t ru = r + (rowCnt / 2);
							if (otherData->fVector[ru] > myv) {
								myv = otherData->fVector[ru];
								myi = ru;
							}					
						}
						myi -= rowCnt / 2;
						x2 = (Double_t)mxi / (Double_t)(rowCnt / 2);
						y2 = (Double_t)myi / (Double_t)(rowCnt / 2);
					}

					// draw line
					fCurrentPad->PaintLine((x1-minX) / dimX * mw + left_m,
									1.0 - ((y1-minY) / dimY * mh) - top_m,
						            (x2-minX) / dimX * mw + left_m,
						            1.0 - ((y2-minY) / dimY * mh) - top_m);
				}			
			}

		}
	}

	// draw LISHs
	
	if (opt.Contains("l")) {
		TObjString key("LISH");

		SetFillColor(kGray);
		SetFillStyle(1001);
		TAttFill::Modify();

		SetLineColor(kBlack);
		SetLineStyle(1);
		SetLineWidth(2);
		TAttLine::Modify();

		for (Int_t tlu = 0; tlu < nrOfTLUnits; ++tlu) {
			TGNGUnit *curTLU = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(tlu));
			TGNGRecursiveData *rData = dynamic_cast<TGNGRecursiveData*>(curTLU->fPrototype);

			TGNG *fLocalGNG = &(rData->fGNG);

			Int_t nrOfUnits = fLocalGNG->fUnits.GetEntriesFast();

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fLocalGNG->fUnits.At(i));
				// get unit pos
				TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
				Double_t x1 = curData->fVector[0];
				Double_t y1 = curData->fVector[1];

				Int_t rowCnt = curData->fVector.GetNrows();
				if (rowCnt > 2) {
					// determine x and y as maximum values in lower and upper half
					Double_t mxv = 0;
					Int_t    mxi = 0;
					Double_t myv = 0;
					Int_t    myi = rowCnt / 2;
					for (Int_t r = 0; r < rowCnt / 2; ++r) {
						if (curData->fVector[r] > mxv) {
							mxv = curData->fVector[r];
							mxi = r;
						}
						Int_t ru = r + (rowCnt / 2);
						if (curData->fVector[ru] > myv) {
							myv = curData->fVector[ru];
							myi = ru;
						}					
					}
					myi -= rowCnt / 2;
					x1 = (Double_t)mxi / (Double_t)(rowCnt / 2);
					y1 = (Double_t)myi / (Double_t)(rowCnt / 2);
				}

				Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
				for (Int_t j = 0; j < edgeCnt; ++j) {
					TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
					if (curEdge->fUnitA != curUnit)
						continue;
					// get other units positions
					TGNGVectorData *otherData = static_cast<TGNGVectorData*>(curEdge->fUnitB->fPrototype);
					Double_t x2 = otherData->fVector[0];
					Double_t y2 = otherData->fVector[1];

					Int_t rowCnt = otherData->fVector.GetNrows();
					if (rowCnt > 2) {
						// determine x and y as maximum values in lower and upper half
						Double_t mxv = 0;
						Int_t    mxi = 0;
						Double_t myv = 0;
						Int_t    myi = rowCnt / 2;
						for (Int_t r = 0; r < rowCnt / 2; ++r) {
							if (otherData->fVector[r] > mxv) {
								mxv = otherData->fVector[r];
								mxi = r;
							}
							Int_t ru = r + (rowCnt / 2);
							if (otherData->fVector[ru] > myv) {
								myv = otherData->fVector[ru];
								myi = ru;
							}					
						}
						myi -= rowCnt / 2;
						x2 = (Double_t)mxi / (Double_t)(rowCnt / 2);
						y2 = (Double_t)myi / (Double_t)(rowCnt / 2);
					}

					// get lish
					TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&key));
					if (lish == 0)
						continue;

					// center pos on line
					Double_t cpx = ((x1+x2)/2.0 - minX) / dimX * mw + left_m;
					Double_t cpy = 1.0 - (((y1+y2)/2.0 - minY) / dimY * mh) - top_m;

					//Double_t bw2 = 0.015;
					//Double_t bh2 = 0.01;

					Double_t bw2 = 0.0375;
					Double_t bh2 = 0.0375;

					Double_t xmargin = 0.002;
					Double_t ymargin = 0.002;

					// draw black box centered on the line
					fCurrentPad->PaintBox(cpx-bw2,cpy+bh2,cpx+bw2,cpy-bh2);

					cpx -= bw2 - xmargin;
					cpy -= bh2 - ymargin;

					// draw mini histogram
					Int_t nrOfBins  = lish->GetNbinsX();
					Double_t maxVal = lish->GetBinContent(lish->GetMaximumBin());
					if (maxVal == 0.0) maxVal = 1.0;
					Double_t xFrac  = ((bw2 - xmargin) * 2.0) / (Double_t)nrOfBins;
					Double_t yFact  = ((bh2 - ymargin) * 2.0) / maxVal;

					for (Int_t k = 0; k < nrOfBins; ++k) {
						Double_t curVal = (x1 < x2) ? lish->GetBinContent(k+1) : 
													  lish->GetBinContent(nrOfBins-k);
						// draw line
						Double_t tmpX = cpx + (Double_t)k * xFrac;
						fCurrentPad->PaintLine(tmpX,cpy,tmpX,cpy + (curVal * yFact));	
					}


				}			
			}
		}
	}
	

	// draw units
	if (!opt.Contains("!u")) {

		for (Int_t tlu = 0; tlu < nrOfTLUnits; ++tlu) {
			TGNGUnit *curTLU = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(tlu));
			TGNGRecursiveData *rData = dynamic_cast<TGNGRecursiveData*>(curTLU->fPrototype);

			TGNG *fLocalGNG = &(rData->fGNG);

			Int_t nrOfUnits = fLocalGNG->fUnits.GetEntriesFast();

			Int_t colIdx = tlu;
			if (colIdx >= fNrOfColors) colIdx = fNrOfColors - 1;

			SetMarkerColor(fColors[colIdx]);
			SetMarkerSize(fUnitSize);
			if (colIdx == fNrOfColors - 1)
				SetMarkerStyle(7);
			else
				SetMarkerStyle(fUnitStyle);
			TAttMarker::Modify();				

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fLocalGNG->fUnits.At(i));
				// get unit pos
				TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);

				Double_t xr1 = curData->fVector[0];
				Double_t yr1 = curData->fVector[1];

				Int_t rowCnt = curData->fVector.GetNrows();
				if (rowCnt > 2) {
					// determine x and y as maximum values in lower and upper half
					Double_t mxv = 0;
					Int_t    mxi = 0;
					Double_t myv = 0;
					Int_t    myi = rowCnt / 2;
					for (Int_t r = 0; r < rowCnt / 2; ++r) {
						if (curData->fVector[r] > mxv) {
							mxv = curData->fVector[r];
							mxi = r;
						}
						Int_t ru = r + (rowCnt / 2);
						if (curData->fVector[ru] > myv) {
							myv = curData->fVector[ru];
							myi = ru;
						}					
					}
					myi -= rowCnt / 2;
					xr1 = (Double_t)mxi / (Double_t)(rowCnt / 2);
					yr1 = (Double_t)myi / (Double_t)(rowCnt / 2);
				}

				Double_t x1 = (xr1 - minX) / dimX * mw + left_m;
				Double_t y1 = 1.0 - ((yr1 - minY) / dimY * mh) - top_m;
				fCurrentPad->PaintPolyMarker(1,&x1,&y1);

			}

		}

	}

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(22);
		SetTextColor(fUnitTextColor);
		SetTextFont(fUnitTextFont);
		SetTextSize(fUnitTextSize);
		TAttText::Modify();

		for (Int_t tlu = 0; tlu < nrOfTLUnits; ++tlu) {
			TGNGUnit *curTLU = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(tlu));
			TGNGRecursiveData *rData = dynamic_cast<TGNGRecursiveData*>(curTLU->fPrototype);

			TGNG *fLocalGNG = &(rData->fGNG);

			Int_t nrOfUnits = fLocalGNG->fUnits.GetEntriesFast();

			for (Int_t i = 0; i < nrOfUnits; ++i) {
				TGNGUnit *curUnit = static_cast<TGNGUnit*>(fLocalGNG->fUnits.At(i));
				// get unit id
				TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
				if (uid == 0)
					continue;

				// get unit pos
				TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);

				Double_t xr1 = curData->fVector[0];
				Double_t yr1 = curData->fVector[1];

				Int_t rowCnt = curData->fVector.GetNrows();
				if (rowCnt > 2) {
					// determine x and y as maximum values in lower and upper half
					Double_t mxv = 0;
					Int_t    mxi = 0;
					Double_t myv = 0;
					Int_t    myi = rowCnt / 2;
					for (Int_t r = 0; r < rowCnt / 2; ++r) {
						if (curData->fVector[r] > mxv) {
							mxv = curData->fVector[r];
							mxi = r;
						}
						Int_t ru = r + (rowCnt / 2);
						if (curData->fVector[ru] > myv) {
							myv = curData->fVector[ru];
							myi = ru;
						}					
					}
					myi -= rowCnt / 2;
					xr1 = (Double_t)mxi / (Double_t)(rowCnt / 2);
					yr1 = (Double_t)myi / (Double_t)(rowCnt / 2);
				}

				Double_t x1 = (xr1 - minX) / dimX * mw + left_m;
				Double_t y1 = 1.0 - ((yr1 - minY) / dimY * mh) - top_m;

				std::sprintf(buffer,"%d",uid->fID);
				fCurrentPad->PaintText(x1+fUnitTextOffsetX,
					            y1+fUnitTextOffsetY,buffer);
			}
		}
	}


	//fCurrentGNG = 0;
	//fCurrentBMU = 0;
}

void TGNGRecursive2DVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t,
									  TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGRecursive2DVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t,
									  TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGRecursive2DVis::SetUpdateInterval(Double_t sec)
{
	// The update interval in seconds determines how often the gng visualization
	// is refreshed while the gng receives inputs.

	fUpdateInterval = sec;
}

void TGNGRecursive2DVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGRecursive2DVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}


void TGNGRecursive2DVis::SetUnitSize(Size_t size)
{
	fUnitSize = size;
}

void TGNGRecursive2DVis::SetUnitStyle(Style_t style)
{
	fUnitStyle = style;
}

void TGNGRecursive2DVis::SetUnitTextOffsetX(Double_t ox)
{
	fUnitTextOffsetX = ox;
}

void TGNGRecursive2DVis::SetUnitTextOffsetY(Double_t oy)
{
	fUnitTextOffsetY = oy;
}

void TGNGRecursive2DVis::SetUnitTextColor(Color_t color)
{
	fUnitTextColor = color;
}

void TGNGRecursive2DVis::SetUnitTextFont(Font_t font)
{
	fUnitTextFont = font;
}

void TGNGRecursive2DVis::SetUnitTextSize(Float_t size)
{
	fUnitTextSize = size;
}



void TGNGRecursive2DVis::UpdateVis()
{
	TTimeStamp current;
	if (current - fLastUpdate < fUpdateInterval)
		return;

	fLastUpdate = current;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	
}

void TGNGRecursive2DVis::SetAutoRange(Bool_t enable)
{
	fAutoRange = enable;
}

void TGNGRecursive2DVis::SetRange(Double_t minX, Double_t maxX, Double_t minY, Double_t maxY)
{
	fXRangeMin = minX;
	fXRangeMax = maxX;
	fYRangeMin = minY;
	fYRangeMax = maxY;
}
