/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 International, 
         see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.

This file contains the definition of a simple GNG network visualization if
the GNG contains only two-dimensional input vectors of type TGNGVectorData
*/
#include <limits>
#include <cstdio>
#include <TSystem.h>
#include <TObjString.h>
#include <TH1D.h>

#include "../../core/TGNGUnit.h"
#include "../../core/TGNG.h"
#include "../../datatypes/TGNGVectorData.h"
#include "../functional/TGNGIdExt.h"

#include "TGNGVector2DVis.h"


TGNGVector2DVis::TGNGVector2DVis() :
	TGNGExtension(),
	TAttLine(),
	TAttMarker(),
	TAttText(),
	TAttFill(),
	fCurrentPad(0),
	fUpdateInterval(0.05),
	fLastUpdate(),
	fCurrentGNG(0),
	fCurrentBMU(0),
	fEdgeColor(kBlack),
	fEdgeStyle(1),
	fEdgeWidth(1),
	fUnitColor(kAzure+1),
	fUnitSize(1),
	fUnitStyle(21),
	fUnitMColor(kOrange+1),
	fUnitMSize(2),
	fUnitMStyle(21),
	fUnitTextOffsetX(0.015),
	fUnitTextOffsetY(-0.015),
	fUnitTextColor(kBlue+2),
	fUnitTextFont(100),
	fUnitTextSize(0.02)	
{}


void TGNGVector2DVis::Draw(Option_t* option)
{
	// draws the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	// clear current Pad
	if (gPad)
		gPad->Clear(); 

	TObject::Draw(option);
	fCurrentPad = gPad;
}

void TGNGVector2DVis::Paint(Option_t *option)
{
	// paints the GNG
	//
	// draws by default edges and units
	//
	// options:
	// !e - do not draw edges
	// !u - do not draw units
	// m - mark current BMU
	// t - draw node IDs (if attribute "ID" is present)
	// l - draw LISHs (if attribute "LISH" is present)

	if ((fCurrentGNG == 0) || (fCurrentPad == 0))
		return;

	// get space dimension
	Double_t minX = numeric_limits<Double_t>::max();
	Double_t maxX = numeric_limits<Double_t>::min();
	Double_t minY = numeric_limits<Double_t>::max();
	Double_t maxY = numeric_limits<Double_t>::min();

	Int_t nrOfUnits = fCurrentGNG->fUnits.GetEntriesFast();
	for (Int_t i = 0; i < nrOfUnits; ++i) {
		TGNGUnit       *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
		TGNGVectorData *curData = dynamic_cast<TGNGVectorData*>(curUnit->fPrototype);
		if (curData == 0) {
			fCurrentGNG->RaiseError(TGNG::ecIncomp,"Data is not of type TGNGVectorData!");
			return;
		}
		if (curData->fVector.GetNrows() != 2) {
			fCurrentGNG->RaiseError(TGNG::ecIncomp,"Data has dimension other than 2!");
			return;			
		}
		Double_t curX = curData->fVector[0];
		Double_t curY = curData->fVector[1];
		if (curX < minX) minX = curX;
		if (curX > maxX) maxX = curX;
		if (curY < minY) minY = curY;
		if (curY > maxY) maxY = curY;
	}

	Double_t dimX = maxX - minX;
	Double_t dimY = maxY - minY;

	// get pad dimensions
	Double_t to_w = fCurrentPad->UtoPixel(1.);
	Double_t to_h = fCurrentPad->VtoPixel(0.);

	Double_t left_m   = fCurrentPad->GetLeftMargin();
	Double_t right_m  = fCurrentPad->GetLeftMargin();
	Double_t top_m    = fCurrentPad->GetTopMargin();
	Double_t bottom_m = fCurrentPad->GetBottomMargin();

	Double_t mw = 1.0 - left_m - right_m;
	Double_t mh = 1.0 - top_m - bottom_m;

	TString opt = option;
	opt.ToLower();

	// draw edges
	if (!opt.Contains("!e")) {

		SetLineColor(fEdgeColor);
		SetLineStyle(fEdgeStyle);
		SetLineWidth(fEdgeWidth);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
			Double_t x1 = curData->fVector[0];
			Double_t y1 = curData->fVector[1];
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TGNGVectorData *otherData = static_cast<TGNGVectorData*>(curEdge->fUnitB->fPrototype);
				Double_t x2 = otherData->fVector[0];
				Double_t y2 = otherData->fVector[1];
				// draw line
				fCurrentPad->PaintLine((x1-minX) / dimX * mw + left_m,
								1.0 - ((y1-minY) / dimY * mh) - top_m,
					            (x2-minX) / dimX * mw + left_m,
					            1.0 - ((y2-minY) / dimY * mh) - top_m);
			}			
		}

	}

	// draw LISHs
	
	if (opt.Contains("l")) {
		TObjString key("LISH");

		SetFillColor(kGray);
		SetFillStyle(1001);
		TAttFill::Modify();

		SetLineColor(kBlack);
		SetLineStyle(1);
		SetLineWidth(2);
		TAttLine::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
			Double_t x1 = curData->fVector[0];
			Double_t y1 = curData->fVector[1];
			Int_t edgeCnt = curUnit->fEdges.GetEntriesFast();
			for (Int_t j = 0; j < edgeCnt; ++j) {
				TGNGEdge *curEdge = static_cast<TGNGEdge*>(curUnit->fEdges.At(j));
				if (curEdge->fUnitA != curUnit)
					continue;
				// get other units positions
				TGNGVectorData *otherData = static_cast<TGNGVectorData*>(curEdge->fUnitB->fPrototype);
				Double_t x2 = otherData->fVector[0];
				Double_t y2 = otherData->fVector[1];

				// get lish
				TH1D *lish = static_cast<TH1D*>(curEdge->fAttributes.GetValue(&key));
				if (lish == 0)
					continue;

				// center pos on line
				Double_t cpx = ((x1+x2)/2.0 - minX) / dimX * mw + left_m;
				Double_t cpy = 1.0 - (((y1+y2)/2.0 - minY) / dimY * mh) - top_m;

				//Double_t bw2 = 0.015;
				//Double_t bh2 = 0.01;

				Double_t bw2 = 0.0375;
				Double_t bh2 = 0.0375;

				Double_t xmargin = 0.002;
				Double_t ymargin = 0.002;

				// draw black box centered on the line
				fCurrentPad->PaintBox(cpx-bw2,cpy+bh2,cpx+bw2,cpy-bh2);

				cpx -= bw2 - xmargin;
				cpy -= bh2 - ymargin;

				// draw mini histogram
				Int_t nrOfBins  = lish->GetNbinsX();
				Double_t maxVal = lish->GetBinContent(lish->GetMaximumBin());
				if (maxVal == 0.0) maxVal = 1.0;
				Double_t xFrac  = ((bw2 - xmargin) * 2.0) / (Double_t)nrOfBins;
				Double_t yFact  = ((bh2 - ymargin) * 2.0) / maxVal;

				for (Int_t k = 0; k < nrOfBins; ++k) {
					Double_t curVal = (x1 < x2) ? lish->GetBinContent(k+1) : 
												  lish->GetBinContent(nrOfBins-k);
					// draw line
					Double_t tmpX = cpx + (Double_t)k * xFrac;
					fCurrentPad->PaintLine(tmpX,cpy,tmpX,cpy + (curVal * yFact));	
				}


			}			
		}
	}
	

	// draw units
	if (!opt.Contains("!u")) {


		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit pos
			TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
			Double_t x1 = (curData->fVector[0] - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((curData->fVector[1] - minY) / dimY * mh) - top_m;
			// later: set style dependend on attributes
			if ((curUnit == fCurrentBMU) && (opt.Contains("m"))) {
				SetMarkerColor(fUnitMColor);
				SetMarkerSize(fUnitMSize);
				SetMarkerStyle(fUnitMStyle);
			} else {
				SetMarkerColor(fUnitColor);
				SetMarkerSize(fUnitSize);
				SetMarkerStyle(fUnitStyle);
			}
			TAttMarker::Modify();				
			fCurrentPad->PaintPolyMarker(1,&x1,&y1);

		}

	}

	// draw uint IDs 
	if (opt.Contains("t")) {
		TObjString key("ID");

		char buffer[10]; // buffer for node IDs
		SetTextAlign(22);
		SetTextColor(fUnitTextColor);
		SetTextFont(fUnitTextFont);
		SetTextSize(fUnitTextSize);
		TAttText::Modify();

		for (Int_t i = 0; i < nrOfUnits; ++i) {
			TGNGUnit *curUnit = static_cast<TGNGUnit*>(fCurrentGNG->fUnits.At(i));
			// get unit id
			TGNGId *uid = static_cast<TGNGId*>(curUnit->fAttributes.GetValue(&key));
			if (uid == 0)
				continue;

			// get unit pos
			TGNGVectorData *curData = static_cast<TGNGVectorData*>(curUnit->fPrototype);
			Double_t x1 = (curData->fVector[0] - minX) / dimX * mw + left_m;
			Double_t y1 = 1.0 - ((curData->fVector[1] - minY) / dimY * mh) - top_m;

			std::sprintf(buffer,"%d",uid->fID);
			fCurrentPad->PaintText(x1+fUnitTextOffsetX,
				            y1+fUnitTextOffsetY,buffer);
		}
	}


	//fCurrentGNG = 0;
	//fCurrentBMU = 0;
}

void TGNGVector2DVis::OnPeekResult(TGNG *gng, TGNGUnit *bmu, Double_t,
								   TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGVector2DVis::OnFeedResult(TGNG *gng, TGNGUnit *bmu, Double_t,
								   TGNGUnit*, Double_t, const TGNGDataElement*)
{
	fCurrentGNG = gng;
	fCurrentBMU = bmu;
	UpdateVis();
}

void TGNGVector2DVis::SetUpdateInterval(Double_t sec)
{
	// The update interval in seconds determines how often the gng visualization
	// is refreshed while the gng receives inputs.

	fUpdateInterval = sec;
}

void TGNGVector2DVis::SetEdgeColor(Color_t color)
{
	fEdgeColor = color;
}

void TGNGVector2DVis::SetEdgeStyle(Style_t style)
{
	fEdgeStyle = style;
}

void TGNGVector2DVis::SetEdgeWidth(Width_t width)
{
	fEdgeWidth = width;
}


void TGNGVector2DVis::SetUnitColor(Color_t color)
{
	fUnitColor = color;
}

void TGNGVector2DVis::SetUnitSize(Size_t size)
{
	fUnitSize = size;
}

void TGNGVector2DVis::SetUnitStyle(Style_t style)
{
	fUnitStyle = style;
}

void TGNGVector2DVis::SetUnitTextOffsetX(Double_t ox)
{
	fUnitTextOffsetX = ox;
}

void TGNGVector2DVis::SetUnitTextOffsetY(Double_t oy)
{
	fUnitTextOffsetY = oy;
}

void TGNGVector2DVis::SetUnitTextColor(Color_t color)
{
	fUnitTextColor = color;
}

void TGNGVector2DVis::SetUnitTextFont(Font_t font)
{
	fUnitTextFont = font;
}

void TGNGVector2DVis::SetUnitTextSize(Float_t size)
{
	fUnitTextSize = size;
}



void TGNGVector2DVis::UpdateVis()
{
	TTimeStamp current;
	if (current - fLastUpdate < fUpdateInterval)
		return;

	fLastUpdate = current;

	if (fCurrentPad) {
		fCurrentPad->Modified();
		fCurrentPad->Update();
		gSystem->ProcessEvents();
	}	
}

