/*
FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
License: Creative Commons Attribution-Noncommercial-Share Alike 4.0 
         International, see README file that comes with this work
Author: J. Kerdels

The use of this code for military and/or intelligence purposes is disapproved of.
--------------------------------------------------------------------------------

This macro contains many functions to view the analysis data of RGNG grid cell 
simulations, i.e., the functions in this macro operate on the output data that
is generated by the AnalyseFile and AnalyseDir functions in the rgng_analysis.C
macro.

Common usage in the ROOT command line would be, e.g.:

root [0] .x tgng_load.C
root [1] .L macros/rgng_inspect.C

root [2] any of the functions below (see documenation at each function)

*/


#include <vector>
#include <utility>
#include <queue>

#include <TTree.h>
#include <TObjArray.h>
#include <TH2D.h>
#include <TH3D.h>
#include <TMatrixT.h>
#include <TMath.h>
#include <TVector2.h>
#include <TFile.h>
#include <TSystemFile.h>
#include <TSystem.h>
#include <TSystemDirectory.h>
#include <TList.h>
#include <TRandom.h>
#include <TCanvas.h>
#include <TROOT.h>
#include <TStyle.h>
#include <TLatex.h>
#include <TLine.h>
#include <TGraph.h>
#include <TColor.h>
#include <THStack.h>

#include "src/core/TGNG.h"
#include "src/core/TGNGUnit.h"
#include "src/core/TNG.h"
#include "src/core/TNGUnit.h"
#include "src/core/TGNGP.h"
#include "src/distmeasures/TGNGMinkowskiDistance.h"
#include "src/datatypes/TGNGVectorData.h"
#include "src/datatypes/TGNGRecursiveData.h"
#include "src/tools/EMD.h"



/*
	This functions plots w x h rate maps of a given RGNG unit (nID) from the
	analysis file (file) over time with the drawing option "option" onto the
	canvas "can". If no canvas is provided (can = 0), a new canvas is created.
	The ObjArray mOut will receive, if provided, each rate map as a TMatrixD.
*/
void plotRatemaps(const char *file, int nID, int w, int h, Option_t* option = "", TCanvas *can = 0, TObjArray *mOut = 0)
{
	if (can == 0) {
		can = new TCanvas();
		can->Divide(w,h);
	}

	gStyle->SetOptStat(0);

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of entries for that node
    Int_t nodeEntries = 0;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID == nID)
    		++nodeEntries;
    }

    if (nodeEntries == 0) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return;
    }    

    if (nodeEntries > w*h)
    	nodeEntries = w*h;

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

    int j = 1;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID != nID)
    		continue;
    	
    	TMatrixD *rmapCl = (TMatrixD*)rmap->Clone();

    	can->cd(j);
    	rmapCl->Draw(option);

    	if (mOut) mOut->Add(rmapCl);

    	//delete rmap;
    	//rmap = 0;

    	if (j++ == nodeEntries)
    		break;
    }

    f->Close();
    delete f;

}



/*
	This functions plots w x h autocorrelations of the rate maps of a given RGNG
	unit (nID) from the analysis file (file) over time with the drawing option 
	"option" onto the canvas "can". If no canvas is provided (can = 0), a new 
	canvas is created. The ObjArray mOut will receive, if provided, each rate 
	map autocorrelation as a TMatrixD.
*/
void plotAutomaps(const char *file, int nID, int w, int h, Option_t* option = "", TCanvas *can = 0, TObjArray *mOut = 0)
{
	if (can == 0) {
		can = new TCanvas();
	}

	gStyle->SetOptStat(0);

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of entries for that node
    Int_t nodeEntries = 0;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID == nID)
    		++nodeEntries;
    }

    if (nodeEntries == 0) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return;
    }    

    if (nodeEntries > w*h)
    	nodeEntries = w*h;

    can->Divide(w,h);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("automap",1);
    rm->SetBranchAddress("automap",&rmap);

    int j = 1;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID != nID)
    		continue;
    	
    	TMatrixD *rmapCl = (TMatrixD*)rmap->Clone();

    	can->cd(j);
    	rmapCl->Draw(option);

    	if (mOut) mOut->Add(rmapCl);

    	//delete rmap;
    	//rmap = 0;

    	if (j++ == nodeEntries)
    		break;
    }

    f->Close();
    delete f;

}


/*
	This function plots 2D grid orientation histograms (orientation 
	distributions over time) of multiple input files. The input files are
	specified using a base string that contains placeholders as they are
	used in sprintf. The vectors topL and botL contain the numeric values
	that are dropped into the base string resulting in |topL| x |botL| input
	files.

	Example:
	base = "analysis_%dTL_%dBL.root"
	topL = {10,50,100}
	botL = {20,40,80}

	results in input files:
	analysis_10TL_20BL.root
	analysis_10TL_40BL.root
	analysis_10TL_80BL.root

	analysis_50TL_20BL.root
	analysis_50TL_40BL.root
	analysis_50TL_80BL.root

	analysis_100TL_20BL.root
	analysis_100TL_40BL.root
	analysis_100TL_80BL.root

	The histograms are plotted using option "option" on Canvas "can". If no
	canvas is provided, a new Canvas is created.
*/
void plotOrientationTLvsBL(const char *base, std::vector<int> &topL, std::vector<int> &botL, Option_t* option = "", TCanvas *can = 0){
	if (can == 0) {
		can = new TCanvas();
	}

	int ts = topL.size();
	int bs = botL.size();

	can->Divide(ts,bs);

	gStyle->SetOptStat(0);

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("orientdist");

		    TH2D *gdc = (TH2D*)gd->Clone();
		    gdc->SetDirectory(gROOT);

		    f->Close();

		    delete f;

		    sprintf(logfile,"%dTL %dBL",tLevel,bLevel);

		    can->cd(bl * ts + tl + 1);

		    gdc->SetTitle(logfile);

		    gdc->Draw(option);

		}
	}

}

/*
	This function plots the first rate maps of all RGNG units after the given
	time threshold (mintime) that are found in a given RGNG analysis file 
	(file). The rate maps are plotted in a raster of w x h maps with option
	"option" on Canvas "can". If no canvas is provided, a new canvas is created.
	The rate maps are stored in the ObjArray "mOut" as TMatrixD if mOut is 
	provided.
*/
void plotRatemaps(ULong64_t mintime, const char *file, int w, int h, Option_t* option = "", TCanvas *can = 0, TObjArray *mOut = 0)
{
	if (can == 0) {
		can = new TCanvas();
	    can->Divide(w,h);
	}

	gStyle->SetOptStat(0);

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of entries for that node
    Int_t nodeEntries = 0;

    int maxentries = w*h;
    std::vector<Bool_t> nodeShown(maxentries,kFALSE);

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((nodeID < maxentries) && (nodeShown[nodeID] == kFALSE)) {
    		nodeShown[nodeID] = kTRUE;
    		++nodeEntries;
    	}
    }

    if (nodeEntries == 0) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return;
    }    

    if (nodeEntries > maxentries)
    	nodeEntries = maxentries;

    printf("nr of nodeEntries: %d\n",nodeEntries);

//    std::vector<Bool_t> nodeShown(nodeEntries,kFALSE);


    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);



    int j = 1;

    for (int i = 0; i < nrOfEntries; ++i) {
//    for (int i = nrOfEntries-1; i > 0; ++i) {
    	rm->GetEntry(i);

    	if ((nodeID >= nodeEntries) || (nodeShown[nodeID] == kFALSE) || (time < mintime))
    		continue;
    	
    	TH2D *rmapCl = new TH2D(*rmap);
    	rmapCl->SetDirectory(0);

    	can->cd(nodeID+1);

    	rmapCl->GetZaxis()->SetRangeUser(-0.000000001,rmapCl->GetMaximum());

    	rmapCl->Draw(option);
    	nodeShown[nodeID] = kFALSE;

    	if (mOut) mOut->Add(rmapCl);

    	//delete rmap;
    	//rmap = 0;

    	if (j++ == nodeEntries)
    		break;
    }

    f->Close();
    delete f;

}

/*
	This function merges the rate maps of all RGNG units that are found first
	after the time threshold "mintime" in the analysis file "file". The merge
	uses either the SUM-approach (mode = 0) or the MAX-approach (mode = 1).
*/
TMatrixD* mergeRatemaps(const char *file, ULong64_t mintime, int mode = 0)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of nodes
    Int_t maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    // vector to store ratemaps
    std::vector<TMatrixD*> rmaps(maxNodeID+1,0);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);

    int j = 1;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

    	if ((rmaps[nodeID] != 0) || (time < mintime))
    		continue;
    	
    	rmaps[nodeID] = (TMatrixD*)rmap->Clone();

    	if (j++ > maxNodeID)
    		break;
    }

    f->Close();
    delete f;

    TMatrixD *result = 0;

    for (int i = 0; i <= maxNodeID; ++i) {
    	TMatrixD *curMat = rmaps[i];
    	if (curMat == 0)
    		continue;
    	if (result == 0) {
    		result = curMat;
    		continue;
    	}
    	if (mode == 0) {
    		*result += *curMat;
    	} else
    	if (mode == 1) {
    		Int_t rc = result->GetNrows();
    		Int_t cc = result->GetNcols();
    		for (int r = 0; r < rc; ++r) {
    			for (int c = 0; c < cc; ++c) {
    				if (TMath::IsNaN((*curMat)(r,c)))
    					continue;
    				if ((*curMat)(r,c) > (*result)(r,c)) {
    					(*result)(r,c) = (*curMat)(r,c);
    				}
    				if ((*result)(r,c) > 1.0)
    					(*result)(r,c) = 1.0;
    			}
    		}
    	}
    	delete curMat;
    }

    return result;
}

/*
	This method calculated the evenness of Matrix "mr" based on the approximated
	earth mover's distance. It assumes matrix dimensions of 50x50.
*/
Double_t getEveness(TMatrixD *mr)
{

	// fast hack to make 50x50 maps power of two
	Double_t avr = mr->Sum() / (50.0*50.0);
	TMatrixD *mr64 = new TMatrixD(64,64);
	for (Int_t i = 0; i < 64; ++i) {
		for (Int_t j = 0; j < 64; ++j) {
			if ((i < 50) && (j < 50)) {
				(*mr64)(i,j) = (*mr)(i,j);
			} else {
				(*mr64)(i,j) = avr;
			}
		}
	}
	TMatrixD *comp = new TMatrixD(*mr64);

	//TMatrixD *comp = new TMatrixD(*mr);
	*comp = 1.0;

	Double_t result = EMD::ApproxEMD(mr64,comp);

	//Double_t result = EMD::ApproxEMD(mr,comp);
	delete mr64;

	delete comp;
	return result;
}

/*
This function estimates the amount of movement between two rate maps
"mr1" and "mr2" using the approximated earth mover's distance. The function
assumes that mr1 and mr2 have a dimension of 50x50
*/
Double_t getMovement(TMatrixD *mr1, TMatrixD *mr2)
{

	// fast hack to make 50x50 maps power of two
	TMatrixD *mr164 = new TMatrixD(64,64);
	for (Int_t i = 0; i < 64; ++i) {
		for (Int_t j = 0; j < 64; ++j) {
			if ((i < 50) && (j < 50)) {
				(*mr164)(i,j) = (*mr1)(i,j);
			} else {
				(*mr164)(i,j) = 0;
			}
		}
	}

	TMatrixD *mr264 = new TMatrixD(64,64);
	for (Int_t i = 0; i < 64; ++i) {
		for (Int_t j = 0; j < 64; ++j) {
			if ((i < 50) && (j < 50)) {
				(*mr264)(i,j) = (*mr2)(i,j);
			} else {
				(*mr264)(i,j) = 0;
			}
		}
	}


	Double_t result = EMD::ApproxEMD(mr164,mr264,kFALSE);

	delete mr164;
	delete mr264;

	return result;
}

/*
	This function calculates the absolute difference between matrices "mr1"
	and "mr2". It assumes that mr1 and mr2 have the same dimensions.
*/
Double_t getAbsDiff(TMatrixD *mr1, TMatrixD *mr2)
{
	Int_t rCnt = mr1->GetNrows();
	Int_t cCnt = mr1->GetNcols();

	Double_t result = 0;

	for (Int_t r = 0; r < rCnt; ++r)
		for (Int_t c = 0; c < cCnt; ++c)
			result += TMath::Abs((*mr1)(r,c) - (*mr2)(r,c));

	return result;
}

/*
	This function calculates the coverage of the merged rate maps after time
	"mintime" in the given analysis file "file" for the given merge mode "mode".
	The file assumes that the rate maps have a resolution of 50 x 50
*/
Double_t getRmapsCoverage(const char *file, ULong64_t mintime, int mode = 0)
{
	TMatrixD *mr = mergeRatemaps(file,mintime,mode);

	// fast hack to make 50x50 maps power of two
	Double_t avr = mr->Sum() / (50.0*50.0);
	TMatrixD *mr64 = new TMatrixD(64,64);
	for (Int_t i = 0; i < 64; ++i) {
		for (Int_t j = 0; j < 64; ++j) {
			if ((i < 50) && (j < 50)) {
				(*mr64)(i,j) = (*mr)(i,j);
			} else {
				(*mr64)(i,j) = avr;
			}
		}
	}
	TMatrixD *comp = new TMatrixD(*mr64);

	//TMatrixD *comp = new TMatrixD(*mr);
	*comp = 1.0;

	Double_t result = EMD::ApproxEMD(mr64,comp);

	//Double_t result = EMD::ApproxEMD(mr,comp);
	delete mr;

	delete mr64;

	delete comp;
	return result;
}



/*
	This function plots the development of the firing fields coverage over time
	from time "from" to time "to" of the rate maps in the analysis file "file"
	using "steps" number of steps and merge mode "mode".
*/
TH1D* getRmapsCovDev(const char *file, ULong64_t from, ULong64_t to, int steps, int mode = 0)
{
	TH1D *result = new TH1D("RmapsCovDev","coverage development",steps,(double)from,(double)to);
	result->SetDirectory(0);
	ULong64_t inc = (to - from) / steps;
	if (inc == 0) inc = 1;
	if (inc > steps) inc = steps;
	int bin = 1;
	for (ULong64_t i = from; i <= to; i += inc) {
		result->SetBinContent(bin++,getRmapsCoverage(file,i,mode));
	}
	return result;
}



/*
	This function plots 2D gridness distributions (gridness distributions over
	time) for multiple input files. The input files are
	specified using a base string that contains placeholders as they are
	used in sprintf. The vectors topL and botL contain the numeric values
	that are dropped into the base string resulting in |topL| x |botL| input
	files.

	Example:
	base = "analysis_%dTL_%dBL.root"
	topL = {10,50,100}
	botL = {20,40,80}

	results in input files:
	analysis_10TL_20BL.root
	analysis_10TL_40BL.root
	analysis_10TL_80BL.root

	analysis_50TL_20BL.root
	analysis_50TL_40BL.root
	analysis_50TL_80BL.root

	analysis_100TL_20BL.root
	analysis_100TL_40BL.root
	analysis_100TL_80BL.root

	The histograms are plotted using option "option" on Canvas "can". If no
	canvas is provided, a new Canvas is created. The parameter "trl" is added
	to the title as trial number.
*/
void plotGridnessTLvsBL2(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, Option_t* option = "", TCanvas *can = 0)
{
	int ts = topL.size();
	int bs = botL.size();

	if (can == 0) {
		can = new TCanvas();
	}

	gStyle->SetOptStat(0);

	TH2D *axHist = new TH2D("GridnessTLvsBL",Form("Gridness TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);
	axHist->SetDirectory(gROOT);

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		char *foo = Form("%d",tLevel);
		axHist->GetXaxis()->SetBinLabel(tl+1,foo);
	}

	for (int bl = 0; bl < bs; ++bl) {
		int bLevel = botL[bl];
		char *bar = Form("%d",bLevel);
		axHist->GetYaxis()->SetBinLabel(bl+1,bar);
	}

	axHist->GetXaxis()->SetTicks("-");
	axHist->GetXaxis()->SetTickLength(0.015);
	axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
	axHist->GetYaxis()->SetTicks("+");
	axHist->GetYaxis()->SetTickLength(0.015);
	axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
	axHist->Draw("AXIS");

	TPad *tiling = new TPad("tilepad","",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("griddist");

		    TH2D *gdc = (TH2D*)gd->Clone();
		    gdc->SetDirectory(0);

		    f->Close();

		    delete f;

		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			//curPad->SetGridy(1);

			gdc->GetXaxis()->SetDrawOption("B");
			gdc->GetXaxis()->SetTickLength(0.);
			gdc->GetXaxis()->SetLabelColor(kWhite);
			gdc->GetXaxis()->SetLabelOffset(1);
			gdc->GetYaxis()->SetDrawOption("B");
			gdc->GetYaxis()->SetTickLength(0.);
			gdc->GetYaxis()->SetLabelColor(kWhite);
			gdc->GetYaxis()->SetLabelOffset(1);

		    gdc->SetTitle("");	    

		    gdc->GetZaxis()->SetRangeUser(-0.00000001,gdc->GetMaximum());

		    gdc->Draw(option);


		}
	}

}


/*
	This function sets a custom palette. Currently only palID = 0 is
	supported. It generates an inverted grey palette.
*/
void setCustomPal(int palID) 
{
	// inverted grey
	if (palID == 0) {
		const Int_t nRGBs = 3;
    	Double_t stops[nRGBs] = { 0.00, 0.50, 1.00};
    	Double_t red[nRGBs]   = { 1.00, 0.50, 0.00};
    	Double_t green[nRGBs] = { 1.00, 0.50, 0.00};
    	Double_t blue[nRGBs]  = { 1.00, 0.50, 0.00};
    	TColor::CreateGradientColorTable(nRGBs, stops, red, green, blue, 255);
    	return;
	}
}




/*
	This function plots a projection of the gridness distributions that occur
	in the last third of a simulation for multiple input files. The input
	files are generated using the base string and the vectors topL and botL as
	described above, e.g., for plotGridnessTLvsBL2(). The trl index is used
	in the title as trial number. If "noTitle" is true, the diagram gets no
	title. If noAxis is true, the diagram gets no axis. If no canvas is 
	supplied, a new one is created.
*/
void plotGridnessPrTLvsBL4(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, TVirtualPad *can = 0, Bool_t noTitle = kFALSE, Bool_t noAxis = kFALSE)
{
	int ts = topL.size();
	int bs = botL.size();

	if (can == 0) {
		can = new TCanvas();
	} else {
		can->Clear();
	}

	gStyle->SetOptStat(0);

	if (noAxis == kFALSE) {
		TH2D *axHist = 0;
		if (trl >= 0)
			axHist = new TH2D(Form("GridnessTLvsBL_%d",trl),Form("Gridness TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);	
		else
			axHist = new TH2D("GridnessTLvsBL","Gridness TL vs. BL",ts,0,1,bs,0,1);	
		axHist->SetDirectory(0);

		if (noTitle == kTRUE)
			axHist->SetTitle("");

		for (int tl = 0; tl < ts; ++tl) {
			int tLevel = topL[tl];
			char *foo = Form("%d",tLevel);
			axHist->GetXaxis()->SetBinLabel(tl+1,foo);
		}

		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			char *bar = Form("%d",bLevel);
			axHist->GetYaxis()->SetBinLabel(bl+1,bar);
		}

		axHist->GetXaxis()->SetTitle("top layer");
		axHist->GetXaxis()->SetTitleOffset(-1.2);
		axHist->GetXaxis()->CenterTitle();
		axHist->GetXaxis()->SetTicks("-");
		axHist->GetXaxis()->SetTickLength(0.015);
		axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
		axHist->GetXaxis()->SetLabelSize(0.05);

		axHist->GetYaxis()->SetTitle("bottom layer");
		axHist->GetYaxis()->SetTitleOffset(-1.2);
		axHist->GetYaxis()->CenterTitle();
		axHist->GetYaxis()->SetTicks("+");
		axHist->GetYaxis()->SetTickLength(0.015);
		axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
		axHist->GetYaxis()->SetLabelSize(0.05);

		gStyle->SetTitleAlign(13);

		axHist->Draw("AXIS");
	}

	can->SetFrameBorderMode(0);
	can->SetFrameLineColor(0);

	TPad *tiling = new TPad(Form("tilepad_%d",trl),"",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	Double_t tpYlow = tiling->GetYlowNDC();
	Double_t tpH    = tiling->GetHNDC();

	Double_t ipXlow  = 0;
	Double_t ipW     = 0;

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("griddist");

		    Int_t nob = gd->GetNbinsX();

			TH1D *pr = gd->ProjectionY(Form("pr_griddist_%d_%d_%d",trl,tl,bl),nob*2/3,nob);
		    pr->SetDirectory(0);

		    pr->SetFillStyle(1001);
		    pr->SetFillColor(kGray+2);


		    TH1D *highGrids = new TH1D(*pr);
		    highGrids->SetDirectory(0);

		    for (int i = 1; i <= 40; ++i)
		    	if (highGrids->GetBinCenter(i) < 0.4)
		    		highGrids->SetBinContent(i,0);

		    highGrids->SetFillStyle(1001);
		    highGrids->SetFillColor(TColor::GetColor(90,120,200));


		    TH1D *posGrids = new TH1D(*pr);
		    posGrids->SetDirectory(0);

		    for (int i = 1; i <= 40; ++i)
		    	if ((posGrids->GetBinCenter(i) < 0.0) || 
		    		(posGrids->GetBinCenter(i) >= 0.4))
		    		posGrids->SetBinContent(i,0);

		    posGrids->SetFillStyle(1001);
		    posGrids->SetFillColor(TColor::GetColor(200,170,90));


		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			curPad->SetGridx(1);

			if ((tl == 0) && (bl == 0)) {
				ipXlow  = curPad->GetXlowNDC();
				ipW     = curPad->GetWNDC();
			}

			//pr->GetXaxis()->SetRange(0,41);

			TH1D *bghist = new TH1D(Form("bg_hist_%d_%d_%d",trl,tl,bl),"",40,-2.0,2.0);
			bghist->SetDirectory(0);
			for (int i = 21; i <= 40; ++i)
				bghist->SetBinContent(i,pr->GetMaximum()*2);
		
			bghist->GetXaxis()->SetDrawOption("B");
			bghist->GetXaxis()->SetTickLength(0.);
			bghist->GetXaxis()->SetLabelColor(kWhite);
			bghist->GetXaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetDrawOption("B");
			bghist->GetYaxis()->SetTickLength(0.);
			bghist->GetYaxis()->SetLabelColor(kWhite);
			bghist->GetYaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetRangeUser(0,pr->GetMaximum()*1.1);

		    bghist->SetFillStyle(1001);
		    bghist->SetFillColorAlpha(kGray+1,0.5);

			bghist->Draw("B");

		    pr->SetTitle("");	    

		    pr->Draw("B SAME");
		    posGrids->Draw("B0 SAME");
		    highGrids->Draw("B0 SAME");



		    f->Close();

		    delete f;

		}
	}

	if (noAxis == kFALSE) {

		can->cd();

		Double_t lineX1 = ipXlow;
		Double_t lineX2 = lineX1 + ipW;
		Double_t lineY  = tpYlow + tpH + 0.11;

		TLine *line = new TLine(lineX1,lineY-0.01,lineX1,lineY+0.01);
		line->Draw();
		line = new TLine(lineX2,lineY-0.01,lineX2,lineY+0.01);
		line->Draw();
		line = new TLine(lineX1,lineY,lineX2,lineY);
		line->Draw();



	//	TLatex *tex = new TLatex(-0.00534,1.03,"-2");
		TLatex *tex = new TLatex(lineX1,1.03,"-2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();

	//	tex = new TLatex(0.1466,1.03,"2");
		tex = new TLatex(lineX2,1.03,"2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();
	}

}


/*
	This function plots a projection of the orientation distributions that occur
	in the last third of a simulation for multiple input files. The input
	files are generated using the base string and the vectors topL and botL as
	described above, e.g., for plotGridnessTLvsBL2(). The trl index is used
	in the title as trial number. If "noTitle" is true, the diagram gets no
	title. If noAxis is true, the diagram gets no axis. If no canvas is 
	supplied, a new one is created. The rbin parameter allows to rebin the
	orientation distributions.
*/
void plotOrientationPrTLvsBL4(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, TVirtualPad *can = 0, int rbin = 2, Bool_t noTitle = kFALSE, Bool_t noAxis = kFALSE)
{
	int ts = topL.size();
	int bs = botL.size();

	if (can == 0) {
		printf("creating new canvas...\n");
		can = new TCanvas();
	} else {
		can->Clear();
	}

	gStyle->SetOptStat(0);

	if (noAxis == kFALSE) {

		TH2D *axHist = 0;
		if (trl >= 0)
			axHist = new TH2D(Form("OrientationTLvsBL_%d",trl),Form("Orientation TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);
		else
			axHist = new TH2D("OrientationTLvsBL","Orientation TL vs. BL",ts,0,1,bs,0,1);
		axHist->SetDirectory(0);

		if (noTitle == kTRUE)
			axHist->SetTitle("");

		for (int tl = 0; tl < ts; ++tl) {
			int tLevel = topL[tl];
			char *foo = Form("%d",tLevel);
			axHist->GetXaxis()->SetBinLabel(tl+1,foo);
		}

		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			char *bar = Form("%d",bLevel);
			axHist->GetYaxis()->SetBinLabel(bl+1,bar);
		}

		axHist->GetXaxis()->SetTitle("top layer");
		axHist->GetXaxis()->SetTitleOffset(-1.2);
		axHist->GetXaxis()->CenterTitle();
		axHist->GetXaxis()->SetTicks("-");
		axHist->GetXaxis()->SetTickLength(0.015);
		axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
		axHist->GetXaxis()->SetLabelSize(0.05);

		axHist->GetYaxis()->SetTitle("bottom layer");
		axHist->GetYaxis()->SetTitleOffset(-1.2);
		axHist->GetYaxis()->CenterTitle();
		axHist->GetYaxis()->SetTicks("+");
		axHist->GetYaxis()->SetTickLength(0.015);
		axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
		axHist->GetYaxis()->SetLabelSize(0.05);

		gStyle->SetTitleAlign(13);

		axHist->Draw("AXIS");
	}

	can->SetFrameBorderMode(0);
	can->SetFrameLineColor(0);

	TPad *tiling = new TPad(Form("tilepad_%d",trl),"",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	Double_t tpYlow = tiling->GetYlowNDC();
	Double_t tpH    = tiling->GetHNDC();

	Double_t ipXlow  = 0;
	Double_t ipW     = 0;

	Double_t *quantiles = new Double_t[2];
	quantiles[0]=0.; quantiles[1]=0.;
	Double_t *prob = new Double_t[2];
	prob[0]=0.10; prob[1]=0.90;


	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("orientdist");

		    Int_t nob = gd->GetNbinsX();

			//TH1D *pr = gd->ProjectionY(Form("pr_orientdist_%d_%d_%d",trl,tl,bl),nob*2/3,nob);
			TH1D *pr = gd->ProjectionY(Form("pr_orientdist_%d_%d_%d",trl,tl,bl),nob-1,nob-1);
		    pr->SetDirectory(0);

		    pr->SetFillStyle(1001);
		    pr->SetFillColor(kGray+2);

		    pr->Rebin(rbin);    


		    // get individual gridness values out of ratemaps tree
			TTree *rm = (TTree*)f->Get("ratemaps");

			// get maximum input cnt
		    rm->SetBranchStatus("*",0);
		    rm->SetBranchStatus("inputCnt",1);

		    ULong64_t inpCnt   = 0;
		    ULong64_t mxInpCnt = 0;
		    rm->SetBranchAddress("inputCnt",&inpCnt);
		    Int_t nrOfEntries = rm->GetEntries();
		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt > mxInpCnt)
		    		mxInpCnt = inpCnt;
		    }
		    ULong64_t inpCntThres = mxInpCnt * 2 / 3;

		    Double_t gridness;
		    rm->SetBranchStatus("gridness",1);
		    rm->SetBranchAddress("gridness",&gridness);

			Double_t orientation;
		    rm->SetBranchStatus("orientation",1);
		    rm->SetBranchAddress("orientation",&orientation);

		    TH1D *highGrids = new TH1D(Form("prhg_orientdist_%d_%d_%d",trl,tl,bl),"",180,-TMath::Pi()/2.0,TMath::Pi()/2.0);
		    highGrids->SetDirectory(0);

		    TH1D *posGrids = new TH1D(Form("prpg_orientdist_%d_%d_%d",trl,tl,bl),"",180,-TMath::Pi()/2.0,TMath::Pi()/2.0);
		    posGrids->SetDirectory(0);

		    TH1D *negGrids = new TH1D(Form("prng_orientdist_%d_%d_%d",trl,tl,bl),"",180,-TMath::Pi()/2.0,TMath::Pi()/2.0);
		    negGrids->SetDirectory(0);

		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt < inpCntThres)
		    		continue;
		    	if (gridness > 0.4) {
		    		highGrids->Fill(orientation);
		    	} else if (gridness > 0.0) {
		    		posGrids->Fill(orientation);
		    	} else {
		    		negGrids->Fill(orientation);
		    	}
		    }

		    highGrids->SetFillStyle(1001);
		    highGrids->SetFillColor(TColor::GetColor(90,120,200));

		    highGrids->Rebin(rbin);    

		    posGrids->SetFillStyle(1001);
		    posGrids->SetFillColor(TColor::GetColor(200,170,90));

		    posGrids->Rebin(rbin);    

		    negGrids->SetFillStyle(1001);
		    negGrids->SetFillColor(kGray+2);

		    negGrids->Rebin(rbin);    

		    f->Close();

		    delete f;

		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			curPad->SetGridx(1);

			if ((tl == 0) && (bl == 0)) {
				ipXlow  = curPad->GetXlowNDC();
				ipW     = curPad->GetWNDC();
			}

			
			TH1D *bghist = new TH1D(Form("bg_hist_%d_%d_%d",trl,tl,bl),"",180,-TMath::Pi()/2.0,TMath::Pi()/2.0);
			bghist->SetDirectory(0);

			if (highGrids->GetEntries() > (pr->GetEntries() / 10.0)) {
			    highGrids->GetQuantiles(2,quantiles,prob);
			    if (quantiles[1] - quantiles[0] < TMath::Pi()/18.0)
					for (int i = 1; i <= 180; ++i)
						bghist->SetBinContent(i,pr->GetMaximum()*2);
			}
		
			bghist->GetXaxis()->SetDrawOption("B");
			bghist->GetXaxis()->SetTickLength(0.);
			bghist->GetXaxis()->SetLabelColor(kWhite);
			bghist->GetXaxis()->SetLabelOffset(1);

			bghist->GetYaxis()->SetDrawOption("B");
			bghist->GetYaxis()->SetTickLength(0.);
			bghist->GetYaxis()->SetLabelColor(kWhite);
			bghist->GetYaxis()->SetLabelOffset(1);

			bghist->GetYaxis()->SetRangeUser(0,pr->GetMaximum()*1.1);

		    bghist->SetFillStyle(1001);
		    bghist->SetFillColorAlpha(kGray+1,0.5);

		    bghist->GetXaxis()->SetRangeUser(0,TMath::Pi()/2.0);

			bghist->Draw("B");


		    THStack *hst = new THStack(Form("hs_%d_%d_%d",trl,tl,bl),"");

		    hst->Add(highGrids);
		    hst->Add(posGrids);
		    hst->Add(negGrids);

		    hst->Draw("B SAME");

		}
	}

	if (noAxis == kFALSE) {

		can->cd();

		Double_t lineX1 = ipXlow;
		Double_t lineX2 = lineX1 + ipW;
		Double_t lineY  = tpYlow + tpH + 0.11;

		TLine *line = new TLine(lineX1,lineY-0.01,lineX1,lineY+0.01);
		line->Draw();
		line = new TLine(lineX2,lineY-0.01,lineX2,lineY+0.01);
		line->Draw();
		line = new TLine(lineX1,lineY,lineX2,lineY);
		line->Draw();

	//	TLatex *tex = new TLatex(0.0045,1.03,"0");
		TLatex *tex = new TLatex(lineX1,1.03,"0");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextAlign(21);
		tex->SetTextSize(0.03);
		tex->SetLineWidth(2);
		tex->Draw();

	//	tex = new TLatex(0.14,1.03,"#pi/2");
		tex = new TLatex(lineX2,1.03,"#pi/2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextAlign(21);
		tex->SetTextSize(0.03);
		tex->SetLineWidth(2);
		tex->Draw();

	}	
}



/*
	This function plots a projection of the grid spacing distributions that 
	occur in the last third of a simulation for multiple input files. The input
	files are generated using the base string and the vectors topL and botL as
	described above, e.g., for plotGridnessTLvsBL2(). The trl index is used
	in the title as trial number. If "noTitle" is true, the diagram gets no
	title. If noAxis is true, the diagram gets no axis. If no canvas is 
	supplied, a new one is created. The rbin parameter allows to rebin the
	spacing distributions. The as parameter defines the range used for the
	spacing histograms.
*/

void plotSpacingPrTLvsBL4(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, int as = 64, TVirtualPad *can = 0, int rbin = 2, Bool_t noTitle = kFALSE, Bool_t noAxis = kFALSE)
{
	int ts = topL.size();
	int bs = botL.size();

	int as2 = as * 2;

	if (can == 0) {
		printf("creating new canvas...\n");
		can = new TCanvas();
	} else {
		can->Clear();
	}

	gStyle->SetOptStat(0);

	if (noAxis == kFALSE) {

		TH2D *axHist = new TH2D(Form("SpacingTLvsBL_%d",trl),Form("Grid Spacing TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);
		axHist->SetDirectory(0);

		if (noTitle == kTRUE)
			axHist->SetTitle("");

		for (int tl = 0; tl < ts; ++tl) {
			int tLevel = topL[tl];
			char *foo = Form("%d",tLevel);
			axHist->GetXaxis()->SetBinLabel(tl+1,foo);
		}

		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			char *bar = Form("%d",bLevel);
			axHist->GetYaxis()->SetBinLabel(bl+1,bar);
		}

		axHist->GetXaxis()->SetTitle("top level");
		axHist->GetXaxis()->SetTitleOffset(-1.2);
		axHist->GetXaxis()->CenterTitle();
		axHist->GetXaxis()->SetTicks("-");
		axHist->GetXaxis()->SetTickLength(0.015);
		axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
		axHist->GetXaxis()->SetLabelSize(0.05);

		axHist->GetYaxis()->SetTitle("bottom level");
		axHist->GetYaxis()->SetTitleOffset(-1.2);
		axHist->GetYaxis()->CenterTitle();
		axHist->GetYaxis()->SetTicks("+");
		axHist->GetYaxis()->SetTickLength(0.015);
		axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
		axHist->GetYaxis()->SetLabelSize(0.05);

		gStyle->SetTitleAlign(13);

		axHist->Draw("AXIS");
	}

	can->SetFrameBorderMode(0);
	can->SetFrameLineColor(0);

	TPad *tiling = new TPad(Form("tilepad_%d",trl),"",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	Double_t tpYlow = tiling->GetYlowNDC();
	Double_t tpH    = tiling->GetHNDC();

	Double_t ipXlow  = 0;
	Double_t ipW     = 0;

	Double_t *quantiles = new Double_t[2];
	quantiles[0]=0.; quantiles[1]=0.;
	Double_t *prob = new Double_t[2];
	prob[0]=0.10; prob[1]=0.90;


	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("spacedist");

		    Int_t nob = gd->GetNbinsX();

			TH1D *pr = gd->ProjectionY(Form("pr_spacedist_%d_%d_%d",trl,tl,bl),nob*2/3,nob);
		    pr->SetDirectory(0);

		    pr->SetFillStyle(1001);
		    pr->SetFillColor(kGray+2);

		    pr->Rebin(rbin);    


		    // get individual gridness values out of ratemaps tree
			TTree *rm = (TTree*)f->Get("ratemaps");

			// get maximum input cnt
		    rm->SetBranchStatus("*",0);
		    rm->SetBranchStatus("inputCnt",1);

		    ULong64_t inpCnt   = 0;
		    ULong64_t mxInpCnt = 0;
		    rm->SetBranchAddress("inputCnt",&inpCnt);
		    Int_t nrOfEntries = rm->GetEntries();
		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt > mxInpCnt)
		    		mxInpCnt = inpCnt;
		    }
		    ULong64_t inpCntThres = mxInpCnt * 2 / 3;

		    Double_t gridness;
		    rm->SetBranchStatus("gridness",1);
		    rm->SetBranchAddress("gridness",&gridness);

			Double_t spacing;
		    rm->SetBranchStatus("spacing",1);
		    rm->SetBranchAddress("spacing",&spacing);

		    TH1D *highGrids = new TH1D(Form("prhg_spacedist_%d_%d_%d",trl,tl,bl),"",as2,0,as2);
		    highGrids->SetDirectory(0);

		    TH1D *posGrids = new TH1D(Form("prpg_spacedist_%d_%d_%d",trl,tl,bl),"",as2,0,as2);
		    posGrids->SetDirectory(0);

		    TH1D *negGrids = new TH1D(Form("prng_spacedist_%d_%d_%d",trl,tl,bl),"",as2,0,as2);
		    negGrids->SetDirectory(0);

		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt < inpCntThres)
		    		continue;
		    	if (gridness > 0.4) {
		    		highGrids->Fill(spacing);
		    	} else if (gridness > 0.0) {
		    		posGrids->Fill(spacing);
		    	} else {
		    		negGrids->Fill(spacing);
		    	}
		    }

		    highGrids->SetFillStyle(1001);
		    highGrids->SetFillColor(TColor::GetColor(90,120,200));

		    highGrids->Rebin(rbin);    

		    posGrids->SetFillStyle(1001);
		    posGrids->SetFillColor(TColor::GetColor(200,170,90));

		    posGrids->Rebin(rbin);    

		    negGrids->SetFillStyle(1001);
		    negGrids->SetFillColor(kGray+2);

		    negGrids->Rebin(rbin);    

		    f->Close();

		    delete f;

		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			curPad->SetGridx(1);

			if ((tl == 0) && (bl == 0)) {
				ipXlow  = curPad->GetXlowNDC();
				ipW     = curPad->GetWNDC();
			}

			
			TH1D *bghist = new TH1D(Form("bg_hist_%d_%d_%d",trl,tl,bl),"",as2,0,as2);
			bghist->SetDirectory(0);

			/*
			if (highGrids->GetEntries() > (pr->GetEntries() / 10.0)) {
			    highGrids->GetQuantiles(2,quantiles,prob);
			    if (quantiles[1] - quantiles[0] < TMath::Pi()/18.0)
					for (int i = 1; i <= 180; ++i)
						bghist->SetBinContent(i,pr->GetMaximum()*2);
			}
			*/
		
			bghist->GetXaxis()->SetDrawOption("B");
			bghist->GetXaxis()->SetTickLength(0.);
			bghist->GetXaxis()->SetLabelColor(kWhite);
			bghist->GetXaxis()->SetLabelOffset(1);

			bghist->GetYaxis()->SetDrawOption("B");
			bghist->GetYaxis()->SetTickLength(0.);
			bghist->GetYaxis()->SetLabelColor(kWhite);
			bghist->GetYaxis()->SetLabelOffset(1);

			bghist->GetYaxis()->SetRangeUser(0,pr->GetMaximum()*1.1);

		    bghist->SetFillStyle(1001);
		    bghist->SetFillColorAlpha(kGray+1,0.5);

		    bghist->GetXaxis()->SetRangeUser(0,as);

			bghist->Draw("B");

		    THStack *hst = new THStack(Form("hs_%d_%d_%d",trl,tl,bl),"");

		    hst->Add(highGrids);
		    hst->Add(posGrids);
		    hst->Add(negGrids);

		    hst->Draw("B SAME");

		}
	}

	if (noAxis == kFALSE) {

		can->cd();

		Double_t lineX1 = ipXlow;
		Double_t lineX2 = lineX1 + ipW;
		Double_t lineY  = tpYlow + tpH + 0.11;

		TLine *line = new TLine(lineX1,lineY-0.01,lineX1,lineY+0.01);
		line->Draw();
		line = new TLine(lineX2,lineY-0.01,lineX2,lineY+0.01);
		line->Draw();
		line = new TLine(lineX1,lineY,lineX2,lineY);
		line->Draw();

	//	TLatex *tex = new TLatex(0.0045,1.03,"0");
		TLatex *tex = new TLatex(lineX1,1.03,"0");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextAlign(21);
		tex->SetTextSize(0.03);
		tex->SetLineWidth(2);
		tex->Draw();

	//	tex = new TLatex(0.14,1.03,"#pi/2");
		tex = new TLatex(lineX2,1.03,"1");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextAlign(21);
		tex->SetTextSize(0.03);
		tex->SetLineWidth(2);
		tex->Draw();

	}	
}

/*
	This function returns the first rate map of the TL unit "nID" after the
	time threshold "minTime" from analysis file "file". If "chMin" is true, all
	zero entries will be set to a tiny value just below 0. This ensures that 
	all cells are drawn when using the "COL" drawing option.
*/
TMatrixD* getRatemap(const char *file, int nID, ULong64_t minTime, Bool_t chMin = kFALSE)
{
	TMatrixD *result = 0;

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

 	ULong64_t inpCnt = 0;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&inpCnt);

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((inpCnt < minTime) || (nodeID != nID))
    		continue;

    	result = (TMatrixD*)rmap->Clone();
    	break;
    }

    f->Close();
    delete f;

	if (chMin == kTRUE) {
		int nr = result->GetNrows();
		int nc = result->GetNcols();
		for (int r = 0; r < nr; ++r)
			for (int c = 0; c < nc; ++c)
				if ( (*result)(r,c) == 0 )
					(*result)(r,c) = -0.0000000001;
	}

    return result;
}


/*
	Returns the first rate map autocorrelation of unit "nID" after time 
	threshold "minTime" from analysis file "file".
*/
TMatrixD* getAutomap(const char *file, int nID, ULong64_t minTime)
{
	TMatrixD *result = 0;

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("automap",1);
    rm->SetBranchAddress("automap",&rmap);

 	ULong64_t inpCnt = 0;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&inpCnt);

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((inpCnt < minTime) || (nodeID != nID))
    		continue;

    	result = (TMatrixD*)rmap->Clone();
    	break;
    }

    f->Close();
    delete f;

    return result;
}

/*
	searches for local maxima in matrix "ac" within neighborhoods of size
	(2nb+1)^2 (-nb to +nb) and returns these as TVector2-Objects in the array
	"maxPoints". The array is cleared before adding the maxima.
*/
void GetPeaks(TMatrixD &ac, Int_t nb, TObjArray *maxPoints)
{
	Int_t nCols = ac.GetNcols();
	Int_t nRows = ac.GetNrows();

	Int_t xmid = nCols / 2;
	Int_t ymid = nRows / 2;

	maxPoints->Clear();

	Int_t x,y,dx,dy;
	for (y = nb; y < nRows-nb; ++y)
		for (x = nb; x < nCols-nb; ++x) {
			Double_t mv = ac(y,x);
			for (dy = -nb; dy <= nb; ++dy)
				for (dx = -nb; dx <= nb; ++dx)
					if ((ac(y+dy,x+dx) >= mv) && ((dx != 0) || (dy != 0))) {
						mv = 0;
						break;
					}
			if (mv > 0)
				maxPoints->Add(new TVector2((Double_t)(x-xmid),(Double_t)(y-ymid)));
		}
}

/*
	This function returns the first rate maps of all TL units after the time
	threshold "minTime" from file "file" as an TObjArray holding 
	TMatrixD-Objects. If "chMin" is true, every map entry that is 0 is changed
	to a very small value just below zero.
*/
TObjArray* getRatemaps(const char *file, ULong64_t minTime, Bool_t chMin = kFALSE)
{
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of nodes
    Int_t maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    // vector to store ratemaps   
    std::vector<int> nodeCheck(maxNodeID+1,0);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

 	ULong64_t inpCnt = 0;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&inpCnt);

    //Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((inpCnt < minTime) || (nodeCheck[nodeID] != 0))
    		continue;

		nodeCheck[nodeID] = 1;

		TMatrixD *tmp = (TMatrixD*)rmap->Clone();

		if (chMin == kTRUE) {
			int nr = tmp->GetNrows();
			int nc = tmp->GetNcols();
			for (int r = 0; r < nr; ++r)
				for (int c = 0; c < nc; ++c)
					if ( (*tmp)(r,c) == 0 )
						(*tmp)(r,c) = -0.0000000001;
		}

    	result->Add(tmp);

    	if (--maxNodeID < 0) break;
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function returns the average maximum activation across all rate maps
	of a simulation run after the time threshold "minTime".
*/
Double_t getAvrMaxAct(const char *file, ULong64_t minTime)
{
	TObjArray *rmaps = getRatemaps(file, minTime);

	Double_t result = 0;
	Int_t nrOfMaps = rmaps->GetEntriesFast();
	for (Int_t i = 0; i < nrOfMaps; ++i) {
		TMatrixD *rmap = (TMatrixD*)rmaps->At(i);
		result += rmap->Max();
	}
	result /= (Double_t)nrOfMaps;

	delete rmaps;

	return result;
}

/*
	This function returns the average minimum activation across all rate maps
	of a simulation run after the time threshold "minTime". If "excludeZero"
	is true, only non-zero activation values are used.
*/
Double_t getAvrMinAct(const char *file, ULong64_t minTime, Bool_t excludeZero = kFALSE)
{
	TObjArray *rmaps = getRatemaps(file, minTime);

	Double_t result = 0;
	Int_t nrOfMaps = rmaps->GetEntriesFast();
	for (Int_t i = 0; i < nrOfMaps; ++i) {
		TMatrixD *rmap = (TMatrixD*)rmaps->At(i);

		if (excludeZero == kFALSE) {
			result += rmap->Min();
			continue;
		}

		Double_t minVal = rmap->Max();
		Int_t rows = rmap->GetNrows();
		Int_t cols = rmap->GetNcols();
		for (Int_t r = 0; r < rows; ++r) {
			for (Int_t c = 0; c < cols; ++c) {
				Double_t curVal = (*rmap)(r,c);
				if ((curVal < minVal) && (curVal > 0))
					minVal = curVal;
			}
		}
		result += minVal;
	}
	result /= (Double_t)nrOfMaps;

	delete rmaps;

	return result;
}


/*
	This function returns the first rate map autocorrelations of all TL units 
	after the time threshold "minTime" from file "file" as an TObjArray holding 
	TMatrixD-Objects.
*/
TObjArray* getAutomaps(const char *file, ULong64_t minTime)
{
	TObjArray *result = new TObjArray();

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of nodes
    Int_t maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    // vector to store ratemaps   
    std::vector<int> nodeCheck(maxNodeID+1,0);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("automap",1);
    rm->SetBranchAddress("automap",&rmap);

 	ULong64_t inpCnt = 0;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&inpCnt);

    //Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((inpCnt < minTime) || (nodeCheck[nodeID] != 0))
    		continue;

		nodeCheck[nodeID] = 1;

		TMatrixD *tmp = (TMatrixD*)rmap->Clone();

    	result->Add(tmp);

    	if (--maxNodeID < 0) break;
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function retrieves the first value "valuename" from the analysis file 
	"file" after the time threshold "minTime" for all TL units.
*/
std::vector<double>* getVals(const char *file, ULong64_t minTime, const char *valname)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    // get nr of nodes
    Int_t maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    // vector used to check if the value for a node was already received   
    std::vector<int> nodeCheck(maxNodeID+1,0);

	std::vector<double> *result = new std::vector<double>(maxNodeID+1,0.0);

    double stuff = 0;
    rm->SetBranchStatus(valname,1);
    rm->SetBranchAddress(valname,&stuff);

 	ULong64_t inpCnt = 0;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&inpCnt);

    //Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if ((inpCnt < minTime) || (nodeCheck[nodeID] != 0))
    		continue;

		nodeCheck[nodeID] = 1;

    	(*result)[nodeID] = stuff;

    	if (--maxNodeID < 0) break;
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function plots a projection of the gridness distributions that occur
	in the last third of a simulation for multiple input files. The input
	files are generated using the base string and the vectors topL and botL as
	described above, e.g., for plotGridnessTLvsBL2(). The trl index is used
	in the title as trial number. If "noTitle" is true, the diagram gets no
	title. If noAxis is true, the diagram gets no axis. If no canvas is 
	supplied, a new one is created. In addition to plotGridnessPrTLvsBL4 it 
	plots also the percentage of units that changed from being classified as
	grid cell to not being classified as grid cell and vice versa within the
	last third of the simulation.
*/
void plotGridnessPrTLvsBL5(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, TVirtualPad *can = 0, Bool_t noTitle = kFALSE, Bool_t noAxis = kFALSE)
{
	int ts = topL.size();
	int bs = botL.size();

	if (can == 0) {
		can = new TCanvas();
	} else {
		can->Clear();
	}

	gStyle->SetOptStat(0);

	if (noAxis == kFALSE) {
		TH2D *axHist = 0;
		if (trl >= 0)
			axHist = new TH2D(Form("GridnessTLvsBL_%d",trl),Form("Gridness TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);	
		else
			axHist = new TH2D("GridnessTLvsBL","Gridness TL vs. BL",ts,0,1,bs,0,1);	
		axHist->SetDirectory(0);

		if (noTitle == kTRUE)
			axHist->SetTitle("");

		for (int tl = 0; tl < ts; ++tl) {
			int tLevel = topL[tl];
			char *foo = Form("%d",tLevel);
			axHist->GetXaxis()->SetBinLabel(tl+1,foo);
		}

		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			char *bar = Form("%d",bLevel);
			axHist->GetYaxis()->SetBinLabel(bl+1,bar);
		}

		axHist->GetXaxis()->SetTitle("top layer");
		axHist->GetXaxis()->SetTitleOffset(-1.2);
		axHist->GetXaxis()->CenterTitle();
		axHist->GetXaxis()->SetTicks("-");
		axHist->GetXaxis()->SetTickLength(0.015);
		axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
		axHist->GetXaxis()->SetLabelSize(0.05);

		axHist->GetYaxis()->SetTitle("bottom layer");
		axHist->GetYaxis()->SetTitleOffset(-1.2);
		axHist->GetYaxis()->CenterTitle();
		axHist->GetYaxis()->SetTicks("+");
		axHist->GetYaxis()->SetTickLength(0.015);
		axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
		axHist->GetYaxis()->SetLabelSize(0.05);

		gStyle->SetTitleAlign(13);

		axHist->Draw("AXIS");
	}

	can->SetFrameBorderMode(0);
	can->SetFrameLineColor(0);

	TPad *tiling = new TPad(Form("tilepad_%d",trl),"",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	Double_t tpYlow = tiling->GetYlowNDC();
	Double_t tpH    = tiling->GetHNDC();

	Double_t ipXlow  = 0;
	Double_t ipW     = 0;

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    TH2D *gd = (TH2D*)f->Get("griddist");

		    Int_t nob = gd->GetNbinsX();

			TH1D *pr = gd->ProjectionY(Form("pr_griddist_%d_%d_%d",trl,tl,bl),nob*2/3,nob);
		    pr->SetDirectory(0);

		    pr->SetFillStyle(1001);
		    pr->SetFillColor(kGray+2);


		    TH1D *highGrids = new TH1D(*pr);
		    highGrids->SetDirectory(0);

		    for (int i = 1; i <= 40; ++i)
		    	if (highGrids->GetBinCenter(i) < 0.4)
		    		highGrids->SetBinContent(i,0);

		    highGrids->SetFillStyle(1001);
		    highGrids->SetFillColor(TColor::GetColor(90,120,200));


		    TH1D *posGrids = new TH1D(*pr);
		    posGrids->SetDirectory(0);

		    for (int i = 1; i <= 40; ++i)
		    	if ((posGrids->GetBinCenter(i) < 0.0) || 
		    		(posGrids->GetBinCenter(i) >= 0.4))
		    		posGrids->SetBinContent(i,0);

		    posGrids->SetFillStyle(1001);
		    posGrids->SetFillColor(TColor::GetColor(200,170,90));




		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			curPad->SetGridx(1);

			if ((tl == 0) && (bl == 0)) {
				ipXlow  = curPad->GetXlowNDC();
				ipW     = curPad->GetWNDC();
			}


			TH1D *bghist = new TH1D(Form("bg_hist_%d_%d_%d",trl,tl,bl),"",40,-2.0,2.0);
			bghist->SetDirectory(0);
			for (int i = 21; i <= 40; ++i)
				bghist->SetBinContent(i,pr->GetMaximum()*2);
		
			bghist->GetXaxis()->SetDrawOption("B");
			bghist->GetXaxis()->SetTickLength(0.);
			bghist->GetXaxis()->SetLabelColor(kWhite);
			bghist->GetXaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetDrawOption("B");
			bghist->GetYaxis()->SetTickLength(0.);
			bghist->GetYaxis()->SetLabelColor(kWhite);
			bghist->GetYaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetRangeUser(0,pr->GetMaximum()*1.1);

		    bghist->SetFillStyle(1001);
		    bghist->SetFillColorAlpha(kGray+1,0.5);

			bghist->Draw("B");

		    pr->SetTitle("");	    

		    pr->Draw("B SAME");
		    posGrids->Draw("B0 SAME");
		    highGrids->Draw("B0 SAME");


		    // get individual gridness values out of ratemaps tree
			TTree *rm = (TTree*)f->Get("ratemaps");

			// get maximum input cnt
		    rm->SetBranchStatus("*",0);
		    rm->SetBranchStatus("inputCnt",1);
		    rm->SetBranchStatus("nodeID",1);

		    ULong64_t inpCnt   = 0;
		    ULong64_t mxInpCnt = 0;
		    rm->SetBranchAddress("inputCnt",&inpCnt);

		    Int_t nodeID    = 0;
		    Int_t maxNodeID = 0;
		    rm->SetBranchAddress("nodeID",&nodeID);

		    Int_t nrOfEntries = rm->GetEntries();
		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt > mxInpCnt)
		    		mxInpCnt = inpCnt;
		    	if (nodeID > maxNodeID)
		    		maxNodeID = nodeID;
		    }
		    ULong64_t inpCntThres = mxInpCnt * 2 / 3;

		    std::vector<std::vector<Double_t>*> gridVals(maxNodeID+1,0);


		    Double_t gridness;
		    rm->SetBranchStatus("gridness",1);
		    rm->SetBranchAddress("gridness",&gridness);

		    Int_t nrOfSteps = 0;

		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt < inpCntThres)
		    		continue;

		    	std::vector<Double_t> *nodeGridVals = gridVals[nodeID];

		    	if (nodeGridVals == 0) {
		    		nodeGridVals = new std::vector<Double_t>();
		    		gridVals[nodeID] = nodeGridVals;
		    	}

				nodeGridVals->push_back(gridness);
				if (nodeGridVals->size() > nrOfSteps)
					nrOfSteps = nodeGridVals->size();
		    }


		    double percChanged = 0;
		    for (int i = 0; i < nrOfSteps-1; ++i) {
			    double totalGCs   = 0;
			    double changedGCs = 0;
		    	for (int j = 0; j < gridVals.size(); ++j) {
		    		std::vector<Double_t> *nodeGridVals = gridVals[j];
		    		if (nodeGridVals == 0)
		    			continue;
		    		Double_t g1 = nodeGridVals->at(i);
		    		Double_t g2 = nodeGridVals->at(i+1);
		    		totalGCs += 1;
		    		if (((g1 >  0.4) && (g2 <= 0.4)) ||
		    			((g1 <= 0.4) && (g2 > 0.4)))
		    			changedGCs += 1;
		    	}
		    	percChanged += changedGCs * 100.0 / totalGCs;
		    }

		    percChanged /= (double)(nrOfSteps-1);

		    for (int j = 0; j < gridVals.size(); ++j)
		    	if (gridVals[j] != 0) delete gridVals[j];

			TText *text = new TText();
			text->SetNDC();
			text->SetTextFont(42);
			//text->SetTextColor();
			text->SetTextSize(0.66);
			text->SetTextAlign(13);
			text->DrawText(0.05,0.85,Form("%2d",(int)percChanged));

		    f->Close();

		    delete f;

		}
	}

	if (noAxis == kFALSE) {

		can->cd();

		Double_t lineX1 = ipXlow;
		Double_t lineX2 = lineX1 + ipW;
		Double_t lineY  = tpYlow + tpH + 0.11;

		TLine *line = new TLine(lineX1,lineY-0.01,lineX1,lineY+0.01);
		line->Draw();
		line = new TLine(lineX2,lineY-0.01,lineX2,lineY+0.01);
		line->Draw();
		line = new TLine(lineX1,lineY,lineX2,lineY);
		line->Draw();



	//	TLatex *tex = new TLatex(-0.00534,1.03,"-2");
		TLatex *tex = new TLatex(lineX1,1.03,"-2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();

	//	tex = new TLatex(0.1466,1.03,"2");
		tex = new TLatex(lineX2,1.03,"2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();
	}

}

/*
	This function plots the distribution of average gridness values within the
	last third of each simulation run provided by multiple input files. The 
	input files are generated using the base string and the vectors topL and 
	botL as described above, e.g., for plotGridnessTLvsBL2().
*/
void plotGridnessAvgTLvsBL(const char *base, int trl, std::vector<int> &topL, std::vector<int> &botL, TVirtualPad *can = 0, Bool_t noTitle = kFALSE, Bool_t noAxis = kFALSE)
{
	int ts = topL.size();
	int bs = botL.size();

	if (can == 0) {
		can = new TCanvas();
	} else {
		can->Clear();
	}

	gStyle->SetOptStat(0);

	if (noAxis == kFALSE) {
		TH2D *axHist = 0;
		if (trl >= 0)
			axHist = new TH2D(Form("GridnessTLvsBL_%d",trl),Form("Gridness TL vs. BL | trial %d",trl),ts,0,1,bs,0,1);	
		else
			axHist = new TH2D("GridnessTLvsBL","Gridness TL vs. BL",ts,0,1,bs,0,1);	
		axHist->SetDirectory(0);

		if (noTitle == kTRUE)
			axHist->SetTitle("");

		for (int tl = 0; tl < ts; ++tl) {
			int tLevel = topL[tl];
			char *foo = Form("%d",tLevel);
			axHist->GetXaxis()->SetBinLabel(tl+1,foo);
		}

		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			char *bar = Form("%d",bLevel);
			axHist->GetYaxis()->SetBinLabel(bl+1,bar);
		}

		axHist->GetXaxis()->SetTitle("top layer");
		axHist->GetXaxis()->SetTitleOffset(-1.2);
		axHist->GetXaxis()->CenterTitle();
		axHist->GetXaxis()->SetTicks("-");
		axHist->GetXaxis()->SetTickLength(0.015);
		axHist->GetXaxis()->SetNdivisions(ts*100,kFALSE);
		axHist->GetXaxis()->SetLabelSize(0.05);

		axHist->GetYaxis()->SetTitle("bottom layer");
		axHist->GetYaxis()->SetTitleOffset(-1.2);
		axHist->GetYaxis()->CenterTitle();
		axHist->GetYaxis()->SetTicks("+");
		axHist->GetYaxis()->SetTickLength(0.015);
		axHist->GetYaxis()->SetNdivisions(bs*100,kFALSE);
		axHist->GetYaxis()->SetLabelSize(0.05);

		gStyle->SetTitleAlign(13);

		axHist->Draw("AXIS");
	}

	can->SetFrameBorderMode(0);
	can->SetFrameLineColor(0);

	TPad *tiling = new TPad(Form("tilepad_%d",trl),"",0.1,0.1,0.9,0.9);
	tiling->Divide(ts,bs);
	tiling->SetFillStyle(0);

	tiling->Draw();

	Double_t tpYlow = tiling->GetYlowNDC();
	Double_t tpH    = tiling->GetHNDC();

	Double_t ipXlow  = 0;
	Double_t ipW     = 0;

	for (int tl = 0; tl < ts; ++tl) {
		int tLevel = topL[tl];
		for (int bl = 0; bl < bs; ++bl) {
			int bLevel = botL[bl];
			
		    char logfile[200];
		    sprintf(logfile,base,tLevel,bLevel);

		    TFile *f = TFile::Open(logfile);

		    // get individual gridness values out of ratemaps tree
			TTree *rm = (TTree*)f->Get("ratemaps");

			// get maximum input cnt
		    rm->SetBranchStatus("*",0);
		    rm->SetBranchStatus("inputCnt",1);
		    rm->SetBranchStatus("nodeID",1);

		    ULong64_t inpCnt   = 0;
		    ULong64_t mxInpCnt = 0;
		    rm->SetBranchAddress("inputCnt",&inpCnt);

		    Int_t nodeID    = 0;
		    Int_t maxNodeID = 0;
		    rm->SetBranchAddress("nodeID",&nodeID);

		    Int_t nrOfEntries = rm->GetEntries();
		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt > mxInpCnt)
		    		mxInpCnt = inpCnt;
		    	if (nodeID > maxNodeID)
		    		maxNodeID = nodeID;
		    }
		    ULong64_t inpCntThres = mxInpCnt * 2 / 3;

		    std::vector<std::vector<Double_t>*> gridVals(maxNodeID+1,0);

		    Double_t gridness;
		    rm->SetBranchStatus("gridness",1);
		    rm->SetBranchAddress("gridness",&gridness);

		    Int_t nrOfSteps = 0;

		    for (int i = 0; i < nrOfEntries; ++i) {
		    	rm->GetEntry(i);
		    	if (inpCnt < inpCntThres)
		    		continue;

		    	std::vector<Double_t> *nodeGridVals = gridVals[nodeID];

		    	if (nodeGridVals == 0) {
		    		nodeGridVals = new std::vector<Double_t>();
		    		gridVals[nodeID] = nodeGridVals;
		    	}

				nodeGridVals->push_back(gridness);
				if (nodeGridVals->size() > nrOfSteps)
					nrOfSteps = nodeGridVals->size();
		    }


		    double percChanged = 0;
		    for (int i = 0; i < nrOfSteps-1; ++i) {
			    double totalGCs   = 0;
			    double changedGCs = 0;
		    	for (int j = 0; j < gridVals.size(); ++j) {
		    		std::vector<Double_t> *nodeGridVals = gridVals[j];
		    		if (nodeGridVals == 0)
		    			continue;
		    		Double_t g1 = nodeGridVals->at(i);
		    		Double_t g2 = nodeGridVals->at(i+1);
		    		totalGCs += 1;
		    		if (((g1 >  0.4) && (g2 <= 0.4)) ||
		    			((g1 <= 0.4) && (g2 > 0.4)))
		    			changedGCs += 1;
		    	}
		    	percChanged += changedGCs * 100.0 / totalGCs;
		    }

		    percChanged /= (double)(nrOfSteps-1);

		    TH1D *highGrids = new TH1D("gridDist_high","",40,-2.0,2.0);
		    highGrids->SetDirectory(0);
		    highGrids->SetFillStyle(1001);
		    highGrids->SetFillColor(TColor::GetColor(90,120,200));


		    TH1D *posGrids = new TH1D("gridDist_pos","",40,-2.0,2.0);
		    posGrids->SetDirectory(0);
		    posGrids->SetFillStyle(1001);
		    posGrids->SetFillColor(TColor::GetColor(200,170,90));

		    TH1D *negGrids = new TH1D("gridDist_neg","",40,-2.0,2.0);
		    negGrids->SetDirectory(0);
		    negGrids->SetFillStyle(1001);
		    negGrids->SetFillColor(kGray+2);

		    for (int j = 0; j < gridVals.size(); ++j) {
		    	std::vector<Double_t> *nodeGridVals = gridVals[j];
		    	if (nodeGridVals == 0)
		    		continue;
		    	Double_t avgGrid  = 0;
		    	Double_t avgCount = 0;
		    	for (int i = 0; i < nodeGridVals->size(); ++i) {
		    		avgGrid  += nodeGridVals->at(i);
		    		avgCount += 1.0;
		    	}
		    	if (avgCount > 0)
		    		avgGrid /= avgCount;
		    	if (avgGrid < 0) {
		    		negGrids->Fill(avgGrid);
		    	} else if (avgGrid > 0.4) {
		    		highGrids->Fill(avgGrid);
		    	} else {
		    		posGrids->Fill(avgGrid);
		    	}
		    }

		    // clean mem
		    for (int j = 0; j < gridVals.size(); ++j) {
		    	if (gridVals[j] != 0) delete gridVals[j];
		    }

		    TVirtualPad *curPad = tiling->cd((bs-(bl+1)) * ts + tl + 1);

			curPad->SetMargin(0.01,0.01,0.01,0.01);

			curPad->SetGridx(1);

			if ((tl == 0) && (bl == 0)) {
				ipXlow  = curPad->GetXlowNDC();
				ipW     = curPad->GetWNDC();
			}

			//pr->GetXaxis()->SetRange(0,41);
			Double_t maxVal = highGrids->GetMaximum();
			if (posGrids->GetMaximum() > maxVal)
				maxVal = posGrids->GetMaximum();
			if (negGrids->GetMaximum() > maxVal)
				maxVal = negGrids->GetMaximum();

			TH1D *bghist = new TH1D(Form("bg_hist_%d_%d_%d",trl,tl,bl),"",40,-2.0,2.0);
			bghist->SetDirectory(0);
			for (int i = 21; i <= 40; ++i)
				bghist->SetBinContent(i,maxVal*2);
		
			bghist->GetXaxis()->SetDrawOption("B");
			bghist->GetXaxis()->SetTickLength(0.);
			bghist->GetXaxis()->SetLabelColor(kWhite);
			bghist->GetXaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetDrawOption("B");
			bghist->GetYaxis()->SetTickLength(0.);
			bghist->GetYaxis()->SetLabelColor(kWhite);
			bghist->GetYaxis()->SetLabelOffset(1);
			bghist->GetYaxis()->SetRangeUser(0,maxVal*1.1);

		    bghist->SetFillStyle(1001);
		    bghist->SetFillColorAlpha(kGray+1,0.5);

			bghist->Draw("B");

		    negGrids->Draw("B SAME");
		    posGrids->Draw("B0 SAME");
		    highGrids->Draw("B0 SAME");

			TText *text = new TText();
			text->SetNDC();
			text->SetTextFont(42);
			//text->SetTextColor();
			text->SetTextSize(0.66);
			text->SetTextAlign(13);
			text->DrawText(0.05,0.85,Form("%2d",(int)percChanged));

		    f->Close();

		    delete f;

		}
	}

	if (noAxis == kFALSE) {

		can->cd();

		Double_t lineX1 = ipXlow;
		Double_t lineX2 = lineX1 + ipW;
		Double_t lineY  = tpYlow + tpH + 0.11;

		TLine *line = new TLine(lineX1,lineY-0.01,lineX1,lineY+0.01);
		line->Draw();
		line = new TLine(lineX2,lineY-0.01,lineX2,lineY+0.01);
		line->Draw();
		line = new TLine(lineX1,lineY,lineX2,lineY);
		line->Draw();



	//	TLatex *tex = new TLatex(-0.00534,1.03,"-2");
		TLatex *tex = new TLatex(lineX1,1.03,"-2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();

	//	tex = new TLatex(0.1466,1.03,"2");
		tex = new TLatex(lineX2,1.03,"2");
		tex->SetTextFont(42);
	//	tex->SetTextSize(0.02350265);
		tex->SetTextSize(0.03);
		tex->SetTextAlign(21);
		tex->SetLineWidth(2);
		tex->Draw();
	}

}

/*
	This function calculates the average gridness of all TL units for the
	first rate maps encountered after "mintime" in the analysis file "file".
*/
Double_t getAvgGridness(ULong64_t mintime, const char *file)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    int maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    if (maxNodeID == -1) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return -3.0;
    }

    std::vector<Double_t> nodeGridness(maxNodeID+1,-3.0);

    printf("nr of nodes: %d",maxNodeID+1);

    Double_t gridness = 0;
    rm->SetBranchStatus("gridness",1);
    rm->SetBranchAddress("gridness",&gridness);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);

    int j = 0;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

    	if ((nodeGridness[nodeID] != -3.0) || (time < mintime))
    		continue;
    	
		nodeGridness[nodeID] = gridness;

    	if (j++ >= maxNodeID)
    		break;
    }

    Double_t gdSum = 0;
    Double_t gdCnt = 0;

    for (Double_t gd : nodeGridness) {
    	if (gd != -3.0) {
    		gdSum += gd;
    		gdCnt += 1.0;
    	}
    }

    if (gdCnt > 0)
    	gdSum /= gdCnt;

    f->Close();
    delete f;

	printf(" with average gridness of %f\n",gdSum);

    return gdSum;
}


/*
	This function plots the gridness values at or after mintime across multiple 
	episodes for multiple simulations (each simulation as its own graph). The 
	"titles" are used for the plot legend. The 	input files are generated using 
	the base string and the vectors "sims" and "epis" similar to previous 
	functions, e.g., for plotGridnessTLvsBL2().
*/
void plotAvgGridness(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Double_t *x  = new Double_t[nrOfEpisodes];
	Double_t **y = new Double_t*[nrOfSims];

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		x[e] = e+1;
	}

	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {
		y[s] = new Double_t[nrOfEpisodes];
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			Int_t sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    y[s][e] = getAvgGridness(mintime,logfile);		    
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TGraph *curGraph = new TGraph(nrOfEpisodes,x,y[s]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[s]);

		// set styles
		switch (s) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		Int_t eVal = epis[e];
		xAxis->SetBinLabel(xAxis->FindBin(e+1),Form("%d",eVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of BL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(-1.0,1.5);
	yAxis->SetTitle("gridness");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t s = 1; s < nrOfSims; ++s) {
		TGraph *curGraph = (TGraph*)graphs.At(s);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t s = 0; s < nrOfSims; ++s) {
		delete[] y[s];
	}
	delete[] y;
}


/*
	This function plots the gridness values at or after mintime across multiple 
	episodes for multiple simulations (each simulation as its own graph). The 
	"titles" are used for the plot legend. The 	input files are generated using 
	the base string and the vectors "sims" and "epis" similar to previous 
	functions, e.g., for plotGridnessTLvsBL2().
*/
void plotAvgGridness(const char *base, std::vector<const char*> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Double_t *x  = new Double_t[nrOfEpisodes];
	Double_t **y = new Double_t*[nrOfSims];

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		x[e] = e+1;
	}

	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {
		y[s] = new Double_t[nrOfEpisodes];
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			const char *sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    y[s][e] = getAvgGridness(mintime,logfile);		    
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TGraph *curGraph = new TGraph(nrOfEpisodes,x,y[s]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[s]);

		// set styles
		switch (s) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		Int_t eVal = epis[e];
		xAxis->SetBinLabel(xAxis->FindBin(e+1),Form("%d",eVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of BL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(-1.0,1.5);
	yAxis->SetTitle("gridness");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t s = 1; s < nrOfSims; ++s) {
		TGraph *curGraph = (TGraph*)graphs.At(s);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t s = 0; s < nrOfSims; ++s) {
		delete[] y[s];
	}
	delete[] y;
}

/*
	This function plots the first rate maps of all TL units found after the time
	threshold "mintime" from several input files. The input filenames are 
	assembled via sprintf(logfile,filebase,tVal,rVal) where tVal and rVal are
	instances of the vectors "trials" and "runs", respectively. The parameter
	string "filebase" contains the printf-style placeholders "%d" for tVal and
	rVal. Each plot is stored in a pdf described by "pdfbase" which is augmented
	by tVal and rVal, i.e., by sprintf(pdfname,pdfbase,tVal,rVal).
	If the parameter "store" is provided, all rate maps of each plot are stored
	within a TObjArray in a root file named using the parameter "store". The 
	name of each TObjArray is constructed by sprintf(name,namebase,rVal,tVal).
	Please note the swap in rVal and tVal.
*/
void plotRatemaps(ULong64_t mintime, const char *filebase, const char *namebase, const char *pdfbase, std::vector<int> &trials, std::vector<int> &runs, int w, int h, TCanvas *can, const char *store = 0)
{
	TFile *f = 0;
	if (store) {
		f = TFile::Open(store,"new");
	}

	TObjArray allMaps;
	allMaps.SetOwner(kTRUE);
	for (Int_t t = 0; t < trials.size(); ++t) {

		Int_t tVal = trials[t];

		for (Int_t r = 0; r < runs.size(); ++r) {

			Int_t rVal = runs[r];

			TObjArray *curMaps = new TObjArray();
			curMaps->SetOwner(kTRUE);
			allMaps.Add(curMaps);

		    char logfile[200];
		    sprintf(logfile,filebase,tVal,rVal);

			plotRatemaps(mintime, logfile, w, h, "ACOL", can, curMaps);

		    char pdfname[200];
		    sprintf(pdfname,pdfbase,tVal,rVal);

		    can->SaveAs(pdfname);

		    if (store) {
		    	char name[200];
		    	sprintf(name,namebase,rVal,tVal);
		    	f->cd();
		    	curMaps->Write(name,TObject::kSingleKey);
		    }

		}
	}

	if (store) {
		f->Close();
		delete f;
	}
 	
}

/*
	This function plots merged rate maps based on the first rate maps of all TL 
	units found after the time threshold "mintime" using merge mode "mergemode"
	from several input files. The input filenames are assembled via 
	sprintf(logfile,filebase,tVal,rVal) where tVal and rVal are
	instances of the vectors "trials" and "runs", respectively. The parameter
	string "filebase" contains the printf-style placeholders "%d" for tVal and
	rVal. Each plot is stored in a pdf described by "pdfbase" which is augmented
	by tVal and rVal, i.e., by sprintf(pdfname,pdfbase,tVal,rVal).
	If the parameter "store" is provided, all rate maps of each plot are stored
	within a TObjArray in a root file named using the parameter "store". The 
	name of each TObjArray is constructed by sprintf(name,namebase,rVal,tVal).
	Please note the swap in rVal and tVal.
*/
void mergeRatemaps(ULong64_t mintime, const char *filebase, const char *namebase, const char *pdfbase, std::vector<int> &trials, std::vector<int> &runs, TCanvas *can, Option_t* option = "ACOL", const char *store = 0, int mergemode = 0)
{
	TFile *f = 0;
	if (store) {
		f = TFile::Open(store,"new");
	}

	for (Int_t t = 0; t < trials.size(); ++t) {

		Int_t tVal = trials[t];

		for (Int_t r = 0; r < runs.size(); ++r) {

			Int_t rVal = runs[r];

		    char logfile[200];
		    sprintf(logfile,filebase,tVal,rVal);

		    TMatrixD *mergedMap = mergeRatemaps(logfile,mintime,mergemode);

		    // temporary histogram to have better control over drawing
		    TH2D mmHist(*mergedMap);

		    if (mergemode == 0) {
		    	mmHist.GetZaxis()->SetRangeUser(0.0,mergedMap->Max());
		    } else if (mergemode == 1) {
		    	mmHist.GetZaxis()->SetRangeUser(0.0,1.0);
		    }

		    if (can) {
		    	can->cd();
		    	mmHist.Draw(option);
			}

			if ((pdfbase != 0) && (can != 0)) {
		    	char pdfname[200];
		    	sprintf(pdfname,pdfbase,tVal,rVal);
		    	can->SaveAs(pdfname);
			}

		    if ((store != 0) && (namebase != 0)) {
		    	char name[200];
		    	sprintf(name,namebase,rVal,tVal);
		    	f->cd();
		    	mergedMap->Write(name);
		    }

		}
	}

	if (store) {
		f->Close();
		delete f;
	}
 	
}


/*
	This function returns a new Canvas that has width "w" and height "h" in
	pixels and is tiled "tx" x "ty" times. The margin between the tiles is 
	defined by "margin". In addition, the function sets the global palette to
	"pal" and the number of contours, i.e., the number of palette colors, to
	"cont". The default value for "pal" (55) is the common palette used for
	rate maps in the literature. Lastly, the function disables the display of
	overlay statistics if, e.g., histograms are drawn.
*/
TCanvas* getCanvas(int w, int h, int tx, int ty, Double_t margin = 0.0025, int cont = 99, int pal = 55)
{
	TCanvas *result = new TCanvas();
	result->SetCanvasSize(w,h);
	result->Divide(tx,ty,margin,margin);
	gStyle->SetNumberContours(cont);
	gStyle->SetPalette(pal);
	gStyle->SetOptStat(0);
	return result;
}



/*
	This function plots the earth mover's distance (EMD) between the merged
	rate maps (merged using "mergemode") and a flat map of multiple simulations
	as graphs (one graph for each element of "sims") against a number of 
	episodes ("epis"). The merged rate maps are base on the first rate maps 
	found in each simulation analysis file after "mintime" time steps. The input
	files are derived by sprintf(logfile,base,sVal,eVal) where "base" is a 
	string that contains two placeholders "%d" which are filled with "sVal" and
	"eVal" which are both instances from the vectors "sims" and "epis", 
	respectively. The strings in the "titles" vector are used to label the 
	individual graphs, i.e., the size of sims and the size of titles should
	match ;-).
*/
void plotEMDCover(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime, int mergemode)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Double_t *x  = new Double_t[nrOfEpisodes];
	Double_t **y = new Double_t*[nrOfSims];

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		x[e] = e+1;
	}

	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {
		y[s] = new Double_t[nrOfEpisodes];
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			Int_t sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

	    	y[s][e] = getRmapsCoverage(logfile,mintime,mergemode);
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TGraph *curGraph = new TGraph(nrOfEpisodes,x,y[s]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[s]);

		// set styles
		switch (s) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		Int_t eVal = epis[e];
		xAxis->SetBinLabel(xAxis->FindBin(e+1),Form("%d",eVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.65);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,0.05);
	yAxis->SetTitle("approx. EMD");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(1.1);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t s = 1; s < nrOfSims; ++s) {
		TGraph *curGraph = (TGraph*)graphs.At(s);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t s = 0; s < nrOfSims; ++s) {
		delete[] y[s];
	}
	delete[] y;
}


/*
	This function plots the average number of firing fields found in the first
	rate maps after "mintime" time steps in multiple simulations. The input 
	files are determined by sprintf(logfile,base,sVal,eVal) where "base" is a 
	string that contains two placeholders "%d" which are filled with "sVal" and
	"eVal" which are both instances from the vectors "sims" and "epis", 
	respectively. For each simulation indicated by "sims" a graph is drawn 
	against each variation of this simulation indicated by "epis". The graphs
	are labeled with the strings provided in "titles". If the ObjArray "graphs"
	is provided, the created graphs are placed there. If not Canvas "can" is
	provided, a new Canvas is created.
*/
void plotAvrPeaksCover(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime, TObjArray *graphs = 0, TCanvas *can = 0)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Double_t *x  = new Double_t[nrOfEpisodes];
	Double_t **y = new Double_t*[nrOfSims];

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		x[e] = e+1;
	}

	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {
		y[s] = new Double_t[nrOfEpisodes];
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			Int_t sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    TObjArray *rmaps = getRatemaps(logfile,mintime);
		    rmaps->SetOwner(kTRUE);

		    TObjArray peaks;

		    Int_t nMaps = rmaps->GetEntriesFast();
		    y[s][e] = 0.0;
		    for (Int_t n = 0; n < nMaps; ++n) {
		    	TMatrixD *curMap = static_cast<TMatrixD*>(rmaps->At(n));
		    	GetPeaks(*curMap,2,&peaks);
		    	y[s][e] += (Double_t)(peaks.GetEntriesFast());
		    }
		    if (nMaps > 0)
		    	y[s][e] /= (Double_t)nMaps;

		    delete rmaps;

		}
	}

	// building graphs
	Bool_t gal = kFALSE;
	if (graphs == 0) {
		graphs = new TObjArray();
		graphs->SetOwner(kTRUE);
		gal = kTRUE;
	}

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TGraph *curGraph = new TGraph(nrOfEpisodes,x,y[s]);
		graphs->Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[s]);

		// set styles
		switch (s) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	if (can == 0) {
		can = new TCanvas();
	}

	// get first graph
	TGraph *curGraph = (TGraph*)graphs->At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		Int_t eVal = epis[e];
		xAxis->SetBinLabel(xAxis->FindBin(e+1),Form("%d",eVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.65);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,80.0);
	yAxis->SetTitle("# of firing fields");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(1.1);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t s = 1; s < nrOfSims; ++s) {
		TGraph *curGraph = (TGraph*)graphs->At(s);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs->At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t s = 0; s < nrOfSims; ++s) {
		delete[] y[s];
	}
	delete[] y;

	if (gal == kTRUE) {
		delete graphs;
	}
}


/*
	This function plots the average evenness (as calculated by getEveness) of 
	the first rate maps after "mintime" time steps in multiple simulations. The 
	input files are determined by sprintf(logfile,base,sVal,eVal) where "base" 
	is a string that contains two placeholders "%d" which are filled with "sVal"
	and	"eVal" which are both instances from the vectors "sims" and "epis", 
	respectively. For each simulation indicated by "sims" a graph is drawn 
	against each variation of this simulation indicated by "epis". The graphs
	are labeled with the strings provided in "titles". If the ObjArray "graphs"
	is provided, the created graphs are placed there. If not Canvas "can" is
	provided, a new Canvas is created.
*/
void plotAvrEveness(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime, TObjArray *graphs = 0, TCanvas *can = 0)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Double_t *x  = new Double_t[nrOfEpisodes];
	Double_t **y = new Double_t*[nrOfSims];

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		x[e] = e+1;
	}

	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {
		y[s] = new Double_t[nrOfEpisodes];
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			Int_t sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    TObjArray *rmaps = getRatemaps(logfile,mintime);
		    rmaps->SetOwner(kTRUE);

		    TObjArray peaks;

		    Int_t nMaps = rmaps->GetEntriesFast();
		    y[s][e] = 0.0;
		    for (Int_t n = 0; n < nMaps; ++n) {
		    	TMatrixD *curMap = static_cast<TMatrixD*>(rmaps->At(n));		    	
		    	y[s][e] += getEveness(curMap);
		    }
		    if (nMaps > 0)
		    	y[s][e] /= (Double_t)nMaps;

		    delete rmaps;

		}
	}

	// building graphs
	Bool_t gal = kFALSE;
	if (graphs == 0) {
		graphs = new TObjArray();
		graphs->SetOwner(kTRUE);
		gal = kTRUE;
	}

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TGraph *curGraph = new TGraph(nrOfEpisodes,x,y[s]);
		graphs->Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[s]);

		// set styles
		switch (s) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	if (can == 0) {
		can = new TCanvas();
	}

	// get first graph
	TGraph *curGraph = (TGraph*)graphs->At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t e = 0; e < nrOfEpisodes; ++e) {
		Int_t eVal = epis[e];
		xAxis->SetBinLabel(xAxis->FindBin(e+1),Form("%d",eVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.65);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,0.05);
	yAxis->SetTitle("approx. EMD");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(1.1);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t s = 1; s < nrOfSims; ++s) {
		TGraph *curGraph = (TGraph*)graphs->At(s);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs->At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t s = 0; s < nrOfSims; ++s) {
		delete[] y[s];
	}
	delete[] y;

	if (gal == kTRUE) {
		delete graphs;
	}
}


/*
	This function creates candle plots of the distributions of activity values
	in MAX-based merged rate maps across several simulations and episodes. The 
	input files are determined by sprintf(logfile,base,sVal,eVal) where "base" 
	is a string that contains two placeholders "%d" which are filled with "sVal"
	and	"eVal" which are both instances from the vectors "sims" and "epis", 
	respectively. The merged rate maps are based on the first rate maps found
	in the corresponding analysis file after "mintime" time steps. The candle
	plots are grouped by episodes as to facilitate an easy comparison between
	different simulation runs. The simulation runs are labeled by the strings
	in "titles". If not canvas "can" is provided, a new one is created. 
*/
TObjArray* getMergedCover(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, ULong64_t mintime, TCanvas *can = 0)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Int_t xBinCnt = (nrOfSims+1)*nrOfEpisodes+1;
	Int_t yBinCnt = 20;

	// setting up histogram
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

	for (Int_t s = 0; s < nrOfSims; ++s) {
		TH2D *curHist = new TH2D(Form("mergedCoverHist_%d",s),"",xBinCnt,0.0,(Double_t)xBinCnt,yBinCnt,0.0,1.000001);		
		curHist->SetDirectory(0);
		for (Int_t i = 1; i <= xBinCnt; ++i)
			curHist->GetXaxis()->SetBinLabel(i,"");

		// set all labels
		for (Int_t s2 = 0; s2 < nrOfSims; ++s2) {
			for (Int_t e = 0; e < nrOfEpisodes; ++e) {
				curHist->GetXaxis()->SetBinLabel(s2 + (nrOfSims+1) * e + 2,titles[s2]);
			}
		}

		// some formatting
		curHist->GetXaxis()->LabelsOption("v");

		curHist->GetXaxis()->SetTitle("# of TL units");
		curHist->GetXaxis()->SetTitleOffset(3.0);
		curHist->GetXaxis()->SetTitleSize(0.06);
		curHist->GetXaxis()->SetLabelSize(0.07);
		curHist->GetXaxis()->SetLabelOffset(0.01);

		curHist->GetXaxis()->SetTickLength(-0.21);

		curHist->GetYaxis()->SetTitle("activity");
		curHist->GetYaxis()->SetTitleOffset(0.80);
		curHist->GetYaxis()->SetTitleSize(0.06);
		curHist->GetYaxis()->SetLabelSize(0.05);

		curHist->GetYaxis()->SetRangeUser(-0.1,1.1);

		switch (s) {
			case 0 : {				
				curHist->SetLineColor(kBlue);				
				//curHist->GetXaxis()->SetLabelColor(kBlue);
			} break;

			case 1 : {				
				curHist->SetLineColor(kRed+1);				
				//curHist->GetXaxis()->SetLabelColor(kRed+1);
			} break;

			case 2 : {
				curHist->SetLineColor(kOrange+2);		
				//curHist->GetXaxis()->SetLabelColor(kOrange+2);		
			} break;

			case 3 : {
				curHist->SetLineColor(kGreen+3);		
				//curHist->GetXaxis()->SetLabelColor(kGreen+3);		
			} break;

			default : {
				curHist->SetLineColor(kGray);
				//curHist->GetXaxis()->SetLabelColor(kGray);
			}
		}		

		result->Add(curHist);
	}


	// getting data
	for (Int_t s = 0; s < nrOfSims; ++s) {

		// get cur hist
		TH2D *curHist = static_cast<TH2D*>(result->At(s));

		for (Int_t e = 0; e < nrOfEpisodes; ++e) {
			Int_t sVal = sims[s];
			Int_t eVal = epis[e];

		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    TMatrixD *mMap = mergeRatemaps(logfile,mintime,1);

		    Int_t mmRows = mMap->GetNrows();
		    Int_t mmCols = mMap->GetNcols();

			Double_t xVal = (Double_t)(s + (nrOfSims+1) * e) + 1.5;
		    for (Int_t r = 0; r < mmRows; ++r) {
		    	for (Int_t c = 0; c < mmCols; ++c) {
		    		if ((*mMap)(r,c) > 0)
		    			curHist->Fill(xVal,(*mMap)(r,c));
		    	}
		    }

		    delete mMap;
		}		
	}

	if (can == 0) {
		can = new TCanvas();
	}

	// draw
	gStyle->SetOptStat(0);

	can->SetLeftMargin(0.1);
	can->SetRightMargin(0.02);
	can->SetTopMargin(0.035);
	can->SetBottomMargin(0.33);

	// draw first
	TH2D *curHist = static_cast<TH2D*>(result->At(0));
	curHist->Draw("CANDLE");

	// draw the rest
	for (Int_t s = 1; s < nrOfSims; ++s) {
		curHist = static_cast<TH2D*>(result->At(s));
		curHist->Draw("CANDLE SAME");
	}

	return result;
}

/*
	This function returns the maximum TL unit node ID that can be found
	in the analysis file "file".
*/
Int_t getMaxNodeID(const char *file)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    Int_t result = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (result < nodeID)
    		result = nodeID;
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function returns all rate maps of a specific TL unit ("nID") from an
	analysis file "file" as TMatrixD objects in an ObjArray. If "ignoreFirst"
	is true, the first rate map, which is usually empty, will not be included.
*/
TObjArray* getAllRatemaps(const char *file, int nID, Bool_t ignoreFirst = kTRUE)
{
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID != nID)
    		continue;

		if (ignoreFirst == kTRUE) {
			ignoreFirst = kFALSE;
			continue;
		}

    	result->Add(rmap->Clone());
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function returns all rate maps contained in an analysis file. The
	resulting TObjArray contains for each TL unit a TObjArray that in turn
	contains all rate maps of that TL unit as TMatrixD objects. If "ignoreFirst"
	is true the first rate map of each TL unit is not returned as these maps are
	typically empty.
*/
TObjArray* getAllRatemaps(const char *file, Bool_t ignoreFirst = kTRUE)
{
	Int_t mxID = getMaxNodeID(file);

	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

    Bool_t *firstMap = new Bool_t[mxID+1];

	for (Int_t nid = 0; nid <= mxID; ++nid) {
		TObjArray *curArray = new TObjArray();
		curArray->SetOwner(kTRUE);
		result->Add(curArray);
		firstMap[nid] = kTRUE;
	}

    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    rm->SetBranchStatus("*",0);

    Int_t nodeID = 0;
    rm->SetBranchStatus("nodeID",1);
    rm->SetBranchAddress("nodeID",&nodeID);

    TMatrixD *rmap = 0;
    rm->SetBranchStatus("ratemap",1);
    rm->SetBranchAddress("ratemap",&rmap);

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

		if (firstMap[nodeID] == kTRUE) {
			firstMap[nodeID] = kFALSE;
			if (ignoreFirst == kTRUE)
				continue;
		}
    	
    	// get array
    	TObjArray *curArray = static_cast<TObjArray*>(result->At(nodeID));
    	curArray->Add(rmap->Clone());
    }

    f->Close();
    delete f;

    delete[] firstMap;

    return result;
}

/*
	This function creates candle plots of the movement between consecutive rate
	maps as determined by the getMovement() function across several simulations 
	and episodes. The input files are determined by sprintf(logfile,base,sVal,eVal)
	where "base" is a string that contains two placeholders "%d" which are 
	filled with "sVal" and "eVal" which are both instances from the vectors 
	"sims" and "epis", respectively. The candle plots are grouped by episodes as
	to facilitate an easy comparison between different simulation runs. The 
	simulation runs are labeled by the strings in "titles". If not canvas "can" 
	is provided, a new one is created. 
*/
TObjArray* getMovementDist(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, TCanvas *can = 0)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Int_t xBinCnt = (nrOfSims+1)*nrOfEpisodes+1;
	Int_t yBinCnt = 50;

	// setting up histogram
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

	printf("building histograms\n");
	for (Int_t s = 0; s < nrOfSims; ++s) {
		TH2D *curHist = new TH2D(Form("movementHist_%d",s),"",xBinCnt,0.0,(Double_t)xBinCnt,yBinCnt,0.0,50.0);		
		curHist->SetDirectory(0);
		for (Int_t i = 1; i <= xBinCnt; ++i)
			curHist->GetXaxis()->SetBinLabel(i,"");

		// set all labels
		for (Int_t s2 = 0; s2 < nrOfSims; ++s2) {
			for (Int_t e = 0; e < nrOfEpisodes; ++e) {
				curHist->GetXaxis()->SetBinLabel(s2 + (nrOfSims+1) * e + 2,titles[s2]);
			}
		}

		// some formatting
		curHist->GetXaxis()->LabelsOption("v");

		curHist->GetXaxis()->SetTitle("# of TL units");
		curHist->GetXaxis()->SetTitleOffset(3.0);
		curHist->GetXaxis()->SetTitleSize(0.06);
		curHist->GetXaxis()->SetLabelSize(0.07);
		curHist->GetXaxis()->SetLabelOffset(0.01);

		curHist->GetXaxis()->SetTickLength(-0.21);

		curHist->GetYaxis()->SetTitle("approx. EMD");
		curHist->GetYaxis()->SetTitleOffset(0.80);
		curHist->GetYaxis()->SetTitleSize(0.06);
		curHist->GetYaxis()->SetLabelSize(0.05);

		curHist->GetYaxis()->SetRangeUser(-0.1,50.1);

		switch (s) {
			case 0 : {				
				curHist->SetLineColor(kBlue);				
				//curHist->GetXaxis()->SetLabelColor(kBlue);
			} break;

			case 1 : {				
				curHist->SetLineColor(kRed+1);				
				//curHist->GetXaxis()->SetLabelColor(kRed+1);
			} break;

			case 2 : {
				curHist->SetLineColor(kOrange+2);		
				//curHist->GetXaxis()->SetLabelColor(kOrange+2);		
			} break;

			case 3 : {
				curHist->SetLineColor(kGreen+3);		
				//curHist->GetXaxis()->SetLabelColor(kGreen+3);		
			} break;

			default : {
				curHist->SetLineColor(kGray);
				//curHist->GetXaxis()->SetLabelColor(kGray);
			}
		}		

		result->Add(curHist);
	}


	// getting data
	printf("getting data\n");
	for (Int_t s = 0; s < nrOfSims; ++s) {

		// get cur hist
		TH2D *curHist = static_cast<TH2D*>(result->At(s));

		for (Int_t e = 0; e < nrOfEpisodes; ++e) {

			Int_t sVal = sims[s];
			Int_t eVal = epis[e];


		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    Double_t xVal = (Double_t)(s + (nrOfSims+1) * e) + 1.5;

		    TObjArray* maps = getAllRatemaps(logfile);
		    Int_t maxID = maps->GetEntriesFast();

		    for (Int_t nid = 0; nid < maxID; ++nid) {

				printf("sim %d episode %d node %d                                      \r",sVal,eVal,nid+1);
				fflush(stdout);

				// get array
				TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
		    	
		    	Int_t mapsCnt = curArray->GetEntriesFast();
		    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

		    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
		    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

		    		Double_t mvmnt = getMovement(m1,m2);

					curHist->Fill(xVal,mvmnt);
		    	}
		    }

	    	delete maps;
		}		
	}
	printf("done.                                          \n");

	if (can == 0) {
		can = new TCanvas();
	}

	// draw
	gStyle->SetOptStat(0);

	can->SetLeftMargin(0.1);
	can->SetRightMargin(0.02);
	can->SetTopMargin(0.035);
	can->SetBottomMargin(0.33);

	// draw first
	TH2D *curHist = static_cast<TH2D*>(result->At(0));
	curHist->Draw("CANDLE");

	// draw the rest
	for (Int_t s = 1; s < nrOfSims; ++s) {
		curHist = static_cast<TH2D*>(result->At(s));
		curHist->Draw("CANDLE SAME");
	}

	return result;
}


/*
	This function creates violin plots of the distribution of absolute 
	differences between consecutive rate maps as determined by the getAbsDiff() 
	function across several simulations and episodes. The input files are 
	determined by sprintf(logfile,base,sVal,eVal) where "base" is a string that 
	contains two placeholders "%d" which are filled with "sVal" and "eVal" which
	are both instances from the vectors "sims" and "epis", respectively. The 
	violin plots are grouped by episodes as	to facilitate an easy comparison 
	between different simulation runs. The simulation runs are labeled by the 
	strings in "titles". If not canvas "can" is provided, a new one is created.
	If "layout" is 0, the X-Axis title is closer to the axis.
*/
TObjArray* getDeltaDist(const char *base, std::vector<int> &sims, std::vector<const char*> &titles, std::vector<int> &epis, Int_t layout = -1, TCanvas *can = 0)
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	Int_t xBinCnt = (nrOfSims+1)*nrOfEpisodes+1;
	Int_t yBinCnt = 50;
	Double_t maxY = 475.0;

	// setting up histogram
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

	printf("building histograms\n");
	for (Int_t s = 0; s < nrOfSims; ++s) {
		TH2D *curHist = new TH2D(Form("movementHist_%d",s),"",xBinCnt,0.0,(Double_t)xBinCnt,yBinCnt,0.0,maxY);		
		curHist->SetDirectory(0);
		for (Int_t i = 1; i <= xBinCnt; ++i)
			curHist->GetXaxis()->SetBinLabel(i,"");

		// set all labels
		for (Int_t s2 = 0; s2 < nrOfSims; ++s2) {
			for (Int_t e = 0; e < nrOfEpisodes; ++e) {
				curHist->GetXaxis()->SetBinLabel(s2 + (nrOfSims+1) * e + 2,titles[s2]);
			}
		}

		// some formatting
		switch (layout) {
			case 0 : {
				curHist->GetXaxis()->LabelsOption("v");

				curHist->GetXaxis()->SetTitle("# of TL units");
				curHist->GetXaxis()->SetTitleOffset(1.95);
				curHist->GetXaxis()->SetTitleSize(0.05);
				curHist->GetXaxis()->SetLabelSize(0.05);
				curHist->GetXaxis()->SetLabelOffset(0.006);

				curHist->GetXaxis()->SetTickLength(-0.11);

				curHist->GetYaxis()->SetTitle("abs. difference");
				curHist->GetYaxis()->SetTitleOffset(1.45);
				curHist->GetYaxis()->SetTitleSize(0.05);
				curHist->GetYaxis()->SetLabelSize(0.04);

				curHist->GetYaxis()->SetRangeUser(-0.1,maxY + 0.1);				
			} break;

			default : {
				curHist->GetXaxis()->LabelsOption("v");

				curHist->GetXaxis()->SetTitle("# of TL units");
				curHist->GetXaxis()->SetTitleOffset(3.0);
				curHist->GetXaxis()->SetTitleSize(0.06);
				curHist->GetXaxis()->SetLabelSize(0.07);
				curHist->GetXaxis()->SetLabelOffset(0.01);

				curHist->GetXaxis()->SetTickLength(-0.21);

				curHist->GetYaxis()->SetTitle("abs. difference");
				curHist->GetYaxis()->SetTitleOffset(1.00);
				curHist->GetYaxis()->SetTitleSize(0.06);
				curHist->GetYaxis()->SetLabelSize(0.05);

				curHist->GetYaxis()->SetRangeUser(-0.1,maxY + 0.1);				
			}
		}

		switch (s) {
			case 0 : {				
				curHist->SetLineColor(kBlue);				
				//curHist->GetXaxis()->SetLabelColor(kBlue);
			} break;

			case 1 : {				
				curHist->SetLineColor(kRed+1);				
				//curHist->GetXaxis()->SetLabelColor(kRed+1);
			} break;

			case 2 : {
				curHist->SetLineColor(kOrange+2);		
				//curHist->GetXaxis()->SetLabelColor(kOrange+2);		
			} break;

			case 3 : {
				curHist->SetLineColor(kGreen+3);		
				//curHist->GetXaxis()->SetLabelColor(kGreen+3);		
			} break;

			default : {
				curHist->SetLineColor(kGray);
				//curHist->GetXaxis()->SetLabelColor(kGray);
			}
		}		

		result->Add(curHist);
	}


	// getting data
	printf("getting data\n");
	for (Int_t s = 0; s < nrOfSims; ++s) {

		// get cur hist
		TH2D *curHist = static_cast<TH2D*>(result->At(s));

		for (Int_t e = 0; e < nrOfEpisodes; ++e) {

			Int_t sVal = sims[s];
			Int_t eVal = epis[e];


		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    Double_t xVal = (Double_t)(s + (nrOfSims+1) * e) + 1.5;

		    TObjArray* maps = getAllRatemaps(logfile);
		    Int_t maxID = maps->GetEntriesFast();

		    for (Int_t nid = 0; nid < maxID; ++nid) {

				printf("sim %d episode %d node %d                                      \r",sVal,eVal,nid+1);
				fflush(stdout);

				// get array
				TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
		    	
		    	Int_t mapsCnt = curArray->GetEntriesFast();
		    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

		    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
		    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

		    		Double_t mvmnt = getAbsDiff(m1,m2);

					curHist->Fill(xVal,mvmnt);
		    	}

		    }

	    	delete maps;
		}		
	}
	printf("done.                                          \n");

	if (can == 0) {
		can = new TCanvas();
	}

	// draw
	gStyle->SetOptStat(0);

	switch (layout) {
		case 0 : {
			can->SetCanvasSize(670,950);

			can->SetLeftMargin(0.14);
			can->SetRightMargin(0.02);
			can->SetTopMargin(0.035);
			can->SetBottomMargin(0.18);
		} break;

		default : {

			can->SetLeftMargin(0.13);
			can->SetRightMargin(0.02);
			can->SetTopMargin(0.035);
			can->SetBottomMargin(0.33);
		}
	}

	// draw first
	TH2D *curHist = static_cast<TH2D*>(result->At(0));
	curHist->Draw("VIOLIN");

	// draw the rest
	for (Int_t s = 1; s < nrOfSims; ++s) {
		curHist = static_cast<TH2D*>(result->At(s));
		curHist->Draw("VIOLIN SAME");
	}

	return result;
}

/*
	This function creates a histogram that correlates the maximum absolute 
	difference between consecutive rate maps of a TL unit with the outgrad of
	that unit at/after "mintime" time steps. The function requires not only
	the analysis file "anafile", but also the original "logfile" of the 
	simulation.
*/
TH2D* getDeltaVsGrad(const char* logfile, const char* anafile, ULong64_t mintime)
{
	TObjArray* maps = getAllRatemaps(anafile);

	Int_t nrOfNodes = maps->GetEntriesFast();

	TFile *inpFile = TFile::Open(logfile);

	TTree *gngTree = (TTree*)inpFile->Get("gngTree_0");

	gngTree->SetBranchStatus("*",0);

	ULong64_t  time;
	gngTree->SetBranchStatus("inputCnt",1);
	gngTree->SetBranchAddress("inputCnt",&time);

	Int_t treeSize = gngTree->GetEntries();

	TGNG *gng = 0;

	for (Int_t i = 0; i < treeSize; ++i) {
		gngTree->GetEntry(i);
		if (time < mintime)
			continue;

		gngTree->SetBranchStatus("gng",1);
		gngTree->SetBranchAddress("gng",&gng);

		gngTree->GetEntry(i);
		break;
	}

	TMatrixD *am = gng->GetAdjacencyMatrix();
	Int_t amColCnt = am->GetNcols();
	Int_t amRowCnt = am->GetNrows();

	if (amRowCnt != nrOfNodes)
		printf("warning: numbers of nodes do not match! %d vs. %d\n",amRowCnt,nrOfNodes);

	delete gng;

	inpFile->Close();
	delete inpFile;

	Int_t yBins = nrOfNodes / 10;
	if (yBins < 10)
		yBins = 10;

	// generate histogram
	TH2D *result = new TH2D("deltavsgrad","",50,0.0,475.0,yBins,0.0,nrOfNodes+1);
	result->SetDirectory(0);

	for (Int_t nid = 0; nid < nrOfNodes; ++nid) {
		// get delta
		Double_t maxDelta = 0;
		//Double_t curDelta = 0;
		//Double_t curDeltaCnt = 0;

		TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
    	
    	Int_t mapsCnt = curArray->GetEntriesFast();
    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

    		Double_t curDelta = getAbsDiff(m1,m2);
    		if (curDelta > maxDelta)
    			maxDelta = curDelta;
    		
    		//curDelta += getAbsDiff(m1,m2);
			//curDeltaCnt += 1.0;
			
    	}
    	//if (curDeltaCnt > 0)
    	//	curDelta /= curDeltaCnt;

    	// get grad
    	Double_t grad = 0;
    	for (Int_t c = 0; c < amColCnt; ++c)
    		grad += (*am)(nid,c);

    	// fill
    	//result->Fill(curDelta,grad);
    	result->Fill(maxDelta,grad);
	}

	delete maps;
	delete am;

	return result;
}

/*
	This function creates a histogram that correlates the maximum absolute 
	difference between consecutive rate maps of a TL unit with the normalized 
	ratio of the times this TL unit was BMU or second BMU. The function requires
	not only the analysis file "anafile", but also the original "logfile" of the 
	simulation.
*/
TH2D* getDeltaVsBMUQ(const char* logfile, const char* anafile)
{
	TObjArray* maps = getAllRatemaps(anafile);

	Int_t nrOfNodes = maps->GetEntriesFast();

	TFile *inpFile = TFile::Open(logfile);

	TTree *feedTree = (TTree*)inpFile->Get("feedTree_0");

	feedTree->SetBranchStatus("*",0);

	Int_t bmuIdx;
	feedTree->SetBranchStatus("bmu",1);
	feedTree->SetBranchAddress("bmu",&bmuIdx);

	Int_t bmu2Idx;
	feedTree->SetBranchStatus("bmu2",1);
	feedTree->SetBranchAddress("bmu2",&bmu2Idx);

	Int_t treeSize = feedTree->GetEntries();

	Double_t *bmuCnt  = new Double_t[nrOfNodes];
	Double_t *bmu2Cnt = new Double_t[nrOfNodes];
	Double_t *inpCnt  = new Double_t[nrOfNodes];

	for (Int_t i = 0; i < nrOfNodes; ++i) {
		bmuCnt[i]  = 0.0;
		bmu2Cnt[i] = 0.0;
		inpCnt[i]  = 0.0;
	}

	for (Int_t i = 0; i < treeSize; ++i) {
		feedTree->GetEntry(i);
		bmuCnt[bmuIdx]   += 1.0;
		bmu2Cnt[bmu2Idx] += 1.0;
		inpCnt[bmuIdx]   += 1.0;
		inpCnt[bmu2Idx]  += 1.0;
	}

	inpFile->Close();
	delete inpFile;

	// generate histogram
	TH2D *result = new TH2D("deltavsbmuq","",50,0.0,475.0,20,-1.0,1.0);
	result->SetDirectory(0);

	for (Int_t nid = 0; nid < nrOfNodes; ++nid) {
		// get delta
		Double_t maxDelta = 0;
		//Double_t curDelta = 0;
		//Double_t curDeltaCnt = 0;

		TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
    	
    	Int_t mapsCnt = curArray->GetEntriesFast();
    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

    		
    		Double_t curDelta = getAbsDiff(m1,m2);
    		if (curDelta > maxDelta)
    			maxDelta = curDelta;
    		
    		
    		//curDelta += getAbsDiff(m1,m2);
			//curDeltaCnt += 1.0;
			
    	}
    	/*
    	if (curDeltaCnt > 0)
    		curDelta /= curDeltaCnt;
    	*/

    	// get bmuq
    	Double_t bmuq = bmuCnt[nid] - bmu2Cnt[nid];
    	if (inpCnt[nid] > 0)
    		bmuq /= inpCnt[nid];

    	// fill
    	//result->Fill(curDelta,bmuq);
    	result->Fill(maxDelta,bmuq);
	}

	delete maps;
	delete[] bmuCnt;
	delete[] bmu2Cnt;
	delete[] inpCnt;

	return result;
}

/*
	This function creates a histogram that correlates the maximum absolute 
	difference between consecutive rate maps of a TL unit with the normalized 
	number of times this TL unit received any input. The function requires
	not only the analysis file "anafile", but also the original "logfile" of the 
	simulation.
*/
TH2D* getDeltaVsBMUC(const char* logfile, const char* anafile)
{
	TObjArray* maps = getAllRatemaps(anafile);

	Int_t nrOfNodes = maps->GetEntriesFast();

	TFile *inpFile = TFile::Open(logfile);

	TTree *feedTree = (TTree*)inpFile->Get("feedTree_0");

	feedTree->SetBranchStatus("*",0);

	Int_t bmuIdx;
	feedTree->SetBranchStatus("bmu",1);
	feedTree->SetBranchAddress("bmu",&bmuIdx);

	Int_t bmu2Idx;
	feedTree->SetBranchStatus("bmu2",1);
	feedTree->SetBranchAddress("bmu2",&bmu2Idx);

	Int_t treeSize = feedTree->GetEntries();

	Double_t *bmuCnt  = new Double_t[nrOfNodes];
	Double_t *bmu2Cnt = new Double_t[nrOfNodes];
	Double_t *inpCnt  = new Double_t[nrOfNodes];

	for (Int_t i = 0; i < nrOfNodes; ++i) {
		bmuCnt[i]  = 0.0;
		bmu2Cnt[i] = 0.0;
		inpCnt[i]  = 0.0;
	}

	Double_t absInpCnt = 0;
	for (Int_t i = 0; i < treeSize; ++i) {
		feedTree->GetEntry(i);
		bmuCnt[bmuIdx]   += 1.0;
		bmu2Cnt[bmu2Idx] += 1.0;
		inpCnt[bmuIdx]   += 1.0;
		inpCnt[bmu2Idx]  += 1.0;
		absInpCnt += 1.0;
	}

	inpFile->Close();
	delete inpFile;

	// generate histogram
	TH2D *result = new TH2D("deltavsbmuq","",50,0.0,475.0,20,-1.0,1.0);
	result->SetDirectory(0);

	for (Int_t nid = 0; nid < nrOfNodes; ++nid) {
		// get delta
		//Double_t maxDelta = 0;
		Double_t curDelta = 0;
		Double_t curDeltaCnt = 0;

		TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
    	
    	Int_t mapsCnt = curArray->GetEntriesFast();
    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

    		/*
    		Double_t curDelta = getAbsDiff(m1,m2);
    		if (curDelta > maxDelta)
    			maxDelta = curDelta;
    		*/
    		
    		curDelta += getAbsDiff(m1,m2);
			curDeltaCnt += 1.0;
			
    	}
    	
    	if (curDeltaCnt > 0)
    		curDelta /= curDeltaCnt;
    	


    	// fill
    	result->Fill(curDelta,inpCnt[nid] / absInpCnt);

    	//result->Fill(maxDelta,inpCnt[nid]);
	}

	delete maps;
	delete[] bmuCnt;
	delete[] bmu2Cnt;
	delete[] inpCnt;

	return result;
}

/*
	This function creates a two-dimensional histogram that contains for each
	TL unit (x-dimension) the distribution of summed absolute differences 
	(y-dimension) between consecutive rate maps. The data is taken from the
	analysis file "anafile". If "nodeBins" is -1, the number of TL units is
	automatically determined and set as the number of x-bins in the resulting
	histogram.
*/
TH2D* getDeltaPerNode(const char* anafile, Int_t nodeBins = -1)
{
	TObjArray* maps = getAllRatemaps(anafile);

	Int_t nrOfNodes = maps->GetEntriesFast();	

	if (nodeBins == -1)
		nodeBins = nrOfNodes;

	// generate histogram
	TH2D *result = new TH2D("deltavpernode","",nodeBins,0.0,(Double_t)nodeBins,50,0.0,475.0);
	result->SetDirectory(0);

	for (Int_t nid = 0; nid < nrOfNodes; ++nid) {
		// get delta

		TObjArray *curArray = static_cast<TObjArray*>(maps->At(nid));
    	
    	Int_t mapsCnt = curArray->GetEntriesFast();
    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

    		
    		Double_t curDelta = getAbsDiff(m1,m2);
    		// fill
			result->Fill((Double_t)nid+0.5,curDelta);	
    	}
    	
	}

	delete maps;

	return result;
}

/*
	This function creates a two-dimensional histogram that describes the 
	distributions of summed absolute differences (SAD, y-dimension) between
	consecutive rate maps over time (x-dimension). The SAD distributions are
	taken from the TL unit with ID "nid" in the analysis file "anafile".
*/
TH2D* getDeltaOverTime(const char* anafile, Int_t nid)
{
	TObjArray* maps = getAllRatemaps(anafile,nid);

	Int_t mapsCnt = maps->GetEntriesFast();

	// generate histogram
	TH2D *result = new TH2D("deltavpernode","",mapsCnt,0.0,(Double_t)mapsCnt,50,0.0,475.0);
	result->SetDirectory(0);
   	
	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

		TMatrixD *m1 = static_cast<TMatrixD*>(maps->At(mi));
		TMatrixD *m2 = static_cast<TMatrixD*>(maps->At(mi+1));
		
		Double_t curDelta = getAbsDiff(m1,m2);
		// fill
		result->Fill((Double_t)mi+0.5,curDelta);	
	}
    	
	delete maps;

	return result;
}

/*
	This function creates a two-dimensional histogram that describes the 
	distributions of summed absolute differences (SAD, y-dimension) between
	consecutive rate maps over time (x-dimension). The SAD distributions cover
	all TL units. The data is taken from the analysis file "anafile".
*/
TH2D* getDeltaOverTime(const char* anafile)
{
	TObjArray* maps = getAllRatemaps(anafile);

	Int_t nrOfNodes = maps->GetEntriesFast();

	TObjArray *curArray = static_cast<TObjArray*>(maps->At(nrOfNodes/2));	
	Int_t mapsCnt = curArray->GetEntriesFast();


	// generate histogram
	TH2D *result = new TH2D("deltavpernode","",mapsCnt+1,0.0,(Double_t)(mapsCnt+1)*100000,50,0.0,475.0);
	result->SetDirectory(0);

	for (Int_t nid = 0; nid < nrOfNodes; ++nid) {
		// get delta

		curArray = static_cast<TObjArray*>(maps->At(nid));
    	
    	for (Int_t mi = 0; mi < mapsCnt-1; ++mi) {

    		TMatrixD *m1 = static_cast<TMatrixD*>(curArray->At(mi));
    		TMatrixD *m2 = static_cast<TMatrixD*>(curArray->At(mi+1));

    		
    		Double_t curDelta = getAbsDiff(m1,m2);
    		// fill
			result->Fill(((Double_t)mi+1.5)*100000,curDelta);	
    	}
    	
	}

	delete maps;

	return result;
}

/*
	This function plots a "DeltaPerNode"-Histogram (see getDeltaPerNode() ) 
	"hist" on a canvas "can" with option "option". It adjusts the layout
	accordingly.
*/
void plotDPN(TH2D *hist, TCanvas *can, Option_t* option = "")
{
	gStyle->SetOptStat(0);
	can->SetCanvasSize(300,300);
	can->SetLeftMargin(0.2466667);
	can->SetBottomMargin(0.1866667);

	hist->GetXaxis()->SetTitle("TL unit");
	hist->GetXaxis()->SetLabelFont(42);
	hist->GetXaxis()->SetLabelOffset(0.009);
	hist->GetXaxis()->SetLabelSize(0.06);
	hist->GetXaxis()->SetTitleSize(0.07);
	hist->GetXaxis()->SetTitleOffset(1.1);
	hist->GetXaxis()->SetTitleFont(42);
	hist->GetXaxis()->SetNdivisions(509);

	hist->GetYaxis()->SetTitle("abs. difference");
	hist->GetYaxis()->SetLabelFont(42);
	hist->GetYaxis()->SetLabelOffset(0.012);
	hist->GetYaxis()->SetLabelSize(0.06);
	hist->GetYaxis()->SetTitleSize(0.08);
	hist->GetYaxis()->SetTitleOffset(1.4);
	hist->GetYaxis()->SetTitleFont(42);

	hist->SetMarkerStyle(21);
	hist->SetMarkerSize(0.4);

	can->cd();
	hist->Draw(option);
}

/*
	This function plots a series of "DeltaPerNode"-histograms from analysis
	files defined by the string "base" and the vectors "sims" and "epis" as
	sprintf(logfile,base,sVal,eVal) where sVal and eVal are elements of the
	vectors. The histograms are plotted using plotDPN() with option "option"
	on canvas "can". The plots are stored in files defined by the string 
	"pdfbase" as sprintf(pdffile,pdfbase,sVal,eVal).
*/
void plotAllDPN(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &epis, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {

			Int_t sVal = sims[s];
			Int_t eVal = epis[e];


		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    TH2D *dpn = getDeltaPerNode(logfile);
		    plotDPN(dpn,can,option);

		    char pdffile[200];
		    sprintf(pdffile,pdfbase,sVal,eVal);

		    can->SaveAs(pdffile);

		    delete dpn;

		}		
	}

}

/*
	This function plots a series of "DeltaPerNode"-histograms from analysis
	files defined by the string "base" and the vectors "sims", "tlm", and "blm" 
	as sprintf(logfile,base,sVal,tlVal,blVal) where sVal, tlVal, and blVal are 
	elements of the	vectors. The histograms are plotted using plotDPN() with 
	option "option"	on canvas "can". The plots are stored in files defined by 
	the string "pdfbase" as sprintf(pdffile,pdfbase,sVal,tlVal,blVal).
*/
void plotAllDPN(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims = sims.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = sims[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH2D *dpn = getDeltaPerNode(logfile);
			    plotDPN(dpn,can,option);

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete dpn;
			}
		}		
	}

}

/*
	This function plots a "DeltaOverTime"-Histogram (see getDeltaOverTime() )
	"hist" on canvas "can" using option "option". It adjusts the layout
	accordingly.
*/
void plotDOT(TH2D *hist, TCanvas *can, Option_t* option = "")
{
	gStyle->SetOptStat(0);
	can->SetCanvasSize(300,300);
	can->SetLeftMargin(0.2466667);
	can->SetBottomMargin(0.1866667);

	hist->GetXaxis()->SetTitle("x1000 steps");
	hist->GetXaxis()->SetLabelFont(42);
	hist->GetXaxis()->SetLabelOffset(0.009);
	hist->GetXaxis()->SetLabelSize(0.06);
	hist->GetXaxis()->SetTitleSize(0.07);
	hist->GetXaxis()->SetTitleOffset(1.1);
	hist->GetXaxis()->SetTitleFont(42);
	//hist->GetXaxis()->SetNdivisions(509); // for 2m
	hist->GetXaxis()->SetNdivisions(505); // for 3m

	hist->GetYaxis()->SetTitle("abs. difference");
	hist->GetYaxis()->SetLabelFont(42);
	hist->GetYaxis()->SetLabelOffset(0.012);
	hist->GetYaxis()->SetLabelSize(0.06);
	hist->GetYaxis()->SetTitleSize(0.08);
	hist->GetYaxis()->SetTitleOffset(1.4);
	hist->GetYaxis()->SetTitleFont(42);

	hist->SetMarkerStyle(21);
	hist->SetMarkerSize(0.4);

	can->cd();
	hist->Draw(option);
}

/*
	This function plots a series of "DeltaOverTime"-histograms from analysis
	files defined by the string "base" and the vectors "sims" and "epis" as
	sprintf(logfile,base,sVal,eVal) where sVal and eVal are elements of the
	vectors. The histograms are plotted using plotDOT() with option "option"
	on canvas "can". The plots are stored in files defined by the string 
	"pdfbase" as sprintf(pdffile,pdfbase,sVal,eVal).
*/
void plotAllDOT(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &epis, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims     = sims.size();
	Int_t nrOfEpisodes = epis.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t e = 0; e < nrOfEpisodes; ++e) {

			Int_t sVal = sims[s];
			Int_t eVal = epis[e];


		    char logfile[200];
		    sprintf(logfile,base,sVal,eVal);

		    TH2D *dot = getDeltaOverTime(logfile);
		    plotDOT(dot,can,option);

		    char pdffile[200];
		    sprintf(pdffile,pdfbase,sVal,eVal);

		    can->SaveAs(pdffile);

		    delete dot;

		}		
	}

}

/*
	This function plots a series of "DeltaOverTime"-histograms from analysis
	files defined by the string "base" and the vectors "sims", "tlm", and "blm" 
	as sprintf(logfile,base,sVal,tlVal,blVal) where sVal, tlVal, and blVal are 
	elements of the	vectors. The histograms are plotted using plotDOT() with 
	option "option"	on canvas "can". The plots are stored in files defined by 
	the string "pdfbase" as sprintf(pdffile,pdfbase,sVal,tlVal,blVal).
*/
void plotAllDOT(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims = sims.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = sims[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH2D *dot = getDeltaOverTime(logfile);
			    plotDOT(dot,can,option);

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete dot;
			}
		}		
	}

}

/*
	This function plots a series of merged rate maps and stores the merged rate
	maps in either a root file, or in external files (e.g. as pdf), or both. The
	merged rate maps are based on the first rate maps found after "mintime" time
	steps in the corresponding input files.  The input files are defined by the
	string "filebase" and the vectors "trials", "tlm", and "blm" as
	sprintf(logfile,filebase,tVal,tlVal,blVal) where tVal, tlVal, and blVal are
	elements of the vectors. If the name for an output root file "store" is
	provided, the merged rate maps are stored in this root file as TMatrixD with
	a name based on "namebase" as sprintf(name,namebase,tlVal,blVal,tVal). If
	the string "pdfbase" is provided, the plots are saved as 
	sprintf(pdfname,pdfbase,tVal,tlVal,blVal). The plots are drawn using option
	"option" and the merged rate maps are created using merge mode "mergemode".
*/
void mergeRatemaps(ULong64_t mintime, const char *filebase, const char *namebase, const char *pdfbase, std::vector<int> &trials, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "ACOL", const char *store = 0, int mergemode = 0)
{
	TFile *f = 0;
	if (store) {
		f = TFile::Open(store,"new");
	}

	for (Int_t t = 0; t < trials.size(); ++t) {

		Int_t tVal = trials[t];

		for (Int_t tl = 0; tl < tlm.size(); ++tl) {

			Int_t tlVal = tlm[tl];

			for (Int_t bl = 0; bl < blm.size(); ++bl) {

				Int_t blVal = blm[bl];

			    char logfile[200];
			    sprintf(logfile,filebase,tVal,tlVal,blVal);

			    TMatrixD *mergedMap = mergeRatemaps(logfile,mintime,mergemode);

			    // temporary histogram to have better control over drawing
			    TH2D mmHist(*mergedMap);

			    if (mergemode == 0) {
			    	mmHist.GetZaxis()->SetRangeUser(0.0,mergedMap->Max());
			    } else if (mergemode == 1) {
			    	mmHist.GetZaxis()->SetRangeUser(0.0,1.0);
			    }

			    if (can) {
			    	can->cd();
			    	mmHist.Draw(option);
				}

				if ((pdfbase != 0) && (can != 0)) {
			    	char pdfname[200];
			    	sprintf(pdfname,pdfbase,tVal,tlVal,blVal);
			    	can->SaveAs(pdfname);
				}

			    if ((store != 0) && (namebase != 0)) {
			    	char name[200];
			    	sprintf(name,namebase,tlVal,blVal,tVal);
			    	f->cd();
			    	mergedMap->Write(name);
			    }

			}
		}
	}

	if (store) {
		f->Close();
		delete f;
	}
 	
}


/*
	This function samples single rate maps from a series of input files. The 
	input files are defined by the string "filebase" and the vectors "trials",
	"tlm", and "blm" as sprintf(logfile,filebase,tVal,tlVal,blVal) where tVal,
	tlVal, and blVal are elements of the vectors. The rate maps are stored in
	files defined by the string "pdfbase" as 
	sprintf(pdfname,pdfbase,tVal,tlVal,blVal). For each input file a rate map
	is choosen at random from the set of rate maps that occurs first after
	"mintime" time steps in the corresponding analysis file. The rate maps are
	drawn with option "option" on canvas "can".
*/
void sampleRatemaps(ULong64_t mintime, const char *filebase, const char *pdfbase, std::vector<int> &trials, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "ACOL")
{

	for (Int_t t = 0; t < trials.size(); ++t) {

		Int_t tVal = trials[t];

		for (Int_t tl = 0; tl < tlm.size(); ++tl) {

			Int_t tlVal = tlm[tl];

			for (Int_t bl = 0; bl < blm.size(); ++bl) {

				Int_t blVal = blm[bl];

			    char logfile[200];
			    sprintf(logfile,filebase,tVal,tlVal,blVal);

			    TObjArray* rmaps = getRatemaps(logfile,mintime);
			    rmaps->SetOwner(kTRUE);

			    // get random map
			    TMatrixD *onemap = (TMatrixD*)rmaps->At(gRandom->Integer(rmaps->GetEntriesFast()));

			    // temporary histogram to have better control over drawing
			    TH2D mmHist(*onemap);

			    if (can) {
			    	can->cd();
			    	mmHist.Draw(option);
				}

				if ((pdfbase != 0) && (can != 0)) {
			    	char pdfname[200];
			    	sprintf(pdfname,pdfbase,tVal,tlVal,blVal);
			    	can->SaveAs(pdfname);
				}

				delete rmaps;


			}
		}
	}

 	
}

/*
	This function plot "DeltaOverTime"-Histograms (see getDeltaOverTime() ) for
	a series of input files that are defined by the string "base" and the 
	vectors "sims", "tlm", and "blm" as sprintf(logfile,base,sVal,tlVal,blVal)
	where sVal, tlVal, and blVal are elements of the vectors. The histograms
	are plotted on canvas "can" using option "otpion" and are stored as external
	files defined by the string "pdfbase" as 
	sprintf(pdffile,pdfbase,sVal,tlVal,blVal). As this function is intended to
	produce small, iconic versions of the DOT-histograms, the marker style and
	marker size are adjusted accordingly.
*/
void plotAllDOTRaw(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims = sims.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = sims[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH2D *dot = getDeltaOverTime(logfile);
			    
				dot->SetMarkerStyle(21);
				dot->SetMarkerSize(0.4);

			    dot->Draw(option);

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete dot;
			}
		}		
	}

}

/*
	This function plot "DeltaPerNode"-Histograms (see getDeltaPerNode() ) for
	a series of input files that are defined by the string "base" and the 
	vectors "sims", "tlm", and "blm" as sprintf(logfile,base,sVal,tlVal,blVal)
	where sVal, tlVal, and blVal are elements of the vectors. The histograms
	are plotted on canvas "can" using option "otpion" and are stored as external
	files defined by the string "pdfbase" as 
	sprintf(pdffile,pdfbase,sVal,tlVal,blVal). As this function is intended to
	produce small, iconic versions of the DPN-histograms, the marker style and
	marker size are adjusted accordingly.
*/
void plotAllDPNRaw(const char* base, const char* pdfbase, std::vector<int> &sims, std::vector<int> &tlm, std::vector<int> &blm, TCanvas *can, Option_t* option = "")
{
	Int_t nrOfSims = sims.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = sims[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH2D *dpn = getDeltaPerNode(logfile);

				dpn->SetMarkerStyle(21);
				dpn->SetMarkerSize(0.4);

			    dpn->Draw(option);
			    

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete dpn;
			}
		}		
	}

}

/*
	This function plots the average gridness across trials with different
	numbers of TL units against the number of BL units. The input files are
	defined by the string "base" and the vectors "trials", "tlm", and "blm" as
	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, tlVal, and blVal are
	elements of the vectors. For each element in "trials" a graph is plotted
	and labeled with a corresponding entry in "titles". The gridness scores are 
	aquired from each input file by getAvgGridness("mintime",logfile) and 
	averaged over the entries of "tlm".
*/
void plotAvgGridnessVsBL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfBLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		x[bl] = bl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfBLs];
		Int_t tVal = trials[t];
		for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
			Int_t blVal = blm[bl];
			y[t][bl]   = 0;
			for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
				Int_t tlVal = tlm[tl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    y[t][bl] += getAvgGridness(mintime,logfile);		    
			}
			y[t][bl] /= (Double_t)nrOfTLs;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfBLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		Int_t blVal = blm[bl];
		xAxis->SetBinLabel(xAxis->FindBin(bl+1),Form("%d",blVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of BL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(-1.0,1.5);
	yAxis->SetTitle("gridness");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}

/*
	This function plots the average gridness across trials with different
	numbers of BL units against the number of TL units. The input files are
	defined by the string "base" and the vectors "trials", "tlm", and "blm" as
	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, tlVal, and blVal are
	elements of the vectors. For each element in "trials" a graph is plotted
	and labeled with a corresponding entry in "titles". The gridness scores are 
	aquired from each input file by getAvgGridness("mintime",logfile) and 
	averaged over the entries of "blm".
*/
void plotAvgGridnessVsTL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfTLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		x[tl] = tl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfTLs];
		Int_t tVal = trials[t];
		for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
			Int_t tlVal = tlm[tl];
			y[t][tl]   = 0;
			for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
				Int_t blVal = blm[bl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    y[t][tl] += getAvgGridness(mintime,logfile);		    
			}
			y[t][tl] /= (Double_t)nrOfBLs;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfTLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		Int_t tlVal = tlm[tl];
		xAxis->SetBinLabel(xAxis->FindBin(tl+1),Form("%d",tlVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(-1.0,1.5);
	yAxis->SetTitle("gridness");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}

/*
	This function plots the average coverage of the merged rate maps across 
	trials with different numbers of TL units against the number of BL units. 
	The input files are	defined by the string "base" and the vectors "trials", 
	"tlm", and "blm" as	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, 
	tlVal, and blVal are elements of the vectors. For each element in "trials" 
	a graph is plotted and labeled with a corresponding entry in "titles". The 
	merged rate map coverages are aquired from each input file by 
	getRmapsCoverage(logfile,"mintime","mergemode") and averaged over the 
	entries of "tlm".
*/
void plotEMDCoverVsBL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, int mergemode)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfBLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		x[bl] = bl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfBLs];
		Int_t tVal = trials[t];

		for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
			Int_t blVal = blm[bl];
			y[t][bl] = 0;

			for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
				Int_t tlVal = tlm[tl];

		    	char logfile[200];
		    	sprintf(logfile,base,tVal,tlVal,blVal);

		    	y[t][bl] += getRmapsCoverage(logfile,mintime,mergemode);
			}
			y[t][bl] /= (Double_t)nrOfTLs;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfBLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		Int_t blVal = blm[bl];
		xAxis->SetBinLabel(xAxis->FindBin(bl+1),Form("%d",blVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of BL units");
	xAxis->SetTitleOffset(1.65);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,0.05);
	yAxis->SetTitle("approx. EMD");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(1.1);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}


/*
	This function plots the average coverage of the merged rate maps across 
	trials with different numbers of BL units against the number of TL units. 
	The input files are	defined by the string "base" and the vectors "trials", 
	"tlm", and "blm" as	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, 
	tlVal, and blVal are elements of the vectors. For each element in "trials" 
	a graph is plotted and labeled with a corresponding entry in "titles". The 
	merged rate map coverages are aquired from each input file by 
	getRmapsCoverage(logfile,"mintime","mergemode") and averaged over the 
	entries of "blm".
*/
void plotEMDCoverVsTL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, int mergemode)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfTLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		x[tl] = tl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfTLs];
		Int_t tVal = trials[t];

		for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
			Int_t tlVal = tlm[tl];
			y[t][tl] = 0;

			for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
				Int_t blVal = blm[bl];

		    	char logfile[200];
		    	sprintf(logfile,base,tVal,tlVal,blVal);

		    	y[t][tl] += getRmapsCoverage(logfile,mintime,mergemode);
			}
			y[t][tl] /= (Double_t)nrOfBLs;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfTLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		Int_t tlVal = tlm[tl];
		xAxis->SetBinLabel(xAxis->FindBin(tl+1),Form("%d",tlVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.65);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,0.05);
	yAxis->SetTitle("approx. EMD");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(1.1);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}


/*
	This function creates a candle plot that compares the activity distributions
	of several simulation runs with varying number of TL and BL units. The
	activity distributions are averaged over the simulations with varying TL
	units. The input files are defined by the string "base" and the vectors
	"trials", "tlm", and "blm" as sprintf(logfile,base,tVal,tlVal,blVal) where
	tVal, tlVal, and blVal are elements of these vectors. The different candle
	plots are labeled with the strings in "titles". The MAX-based merged rate 
	maps are derived from the first set of rate maps after "mintime" time steps
	in each analysis file.
*/
TObjArray* getMergedCoverVsBL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, TCanvas *can = 0)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Int_t xBinCnt = (nrOfTrials+1)*nrOfBLs+1;
	Int_t yBinCnt = 20;

	// setting up histogram
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

	for (Int_t s = 0; s < nrOfTrials; ++s) {
		TH2D *curHist = new TH2D(Form("mergedCoverBLHist_%d",s),"",xBinCnt,0.0,(Double_t)xBinCnt,yBinCnt,0.0,1.000001);		
		curHist->SetDirectory(0);
		for (Int_t i = 1; i <= xBinCnt; ++i)
			curHist->GetXaxis()->SetBinLabel(i,"");

		// set all labels
		for (Int_t s2 = 0; s2 < nrOfTrials; ++s2) {
			for (Int_t e = 0; e < nrOfBLs; ++e) {
				curHist->GetXaxis()->SetBinLabel(s2 + (nrOfTrials+1) * e + 2,titles[s2]);
			}
		}

		// some formatting
		curHist->GetXaxis()->LabelsOption("v");

		curHist->GetXaxis()->SetTitle("# of BL units");
		curHist->GetXaxis()->SetTitleOffset(3.0);
		curHist->GetXaxis()->SetTitleSize(0.06);
		curHist->GetXaxis()->SetLabelSize(0.07);
		curHist->GetXaxis()->SetLabelOffset(0.01);

		curHist->GetXaxis()->SetTickLength(-0.21);

		curHist->GetYaxis()->SetTitle("activity");
		curHist->GetYaxis()->SetTitleOffset(0.80);
		curHist->GetYaxis()->SetTitleSize(0.06);
		curHist->GetYaxis()->SetLabelSize(0.05);

		curHist->GetYaxis()->SetRangeUser(-0.1,1.1);

		switch (s) {
			case 0 : {				
				curHist->SetLineColor(kBlue);				
				//curHist->GetXaxis()->SetLabelColor(kBlue);
			} break;

			case 1 : {				
				curHist->SetLineColor(kRed+1);				
				//curHist->GetXaxis()->SetLabelColor(kRed+1);
			} break;

			case 2 : {
				curHist->SetLineColor(kOrange+2);		
				//curHist->GetXaxis()->SetLabelColor(kOrange+2);		
			} break;

			case 3 : {
				curHist->SetLineColor(kGreen+3);		
				//curHist->GetXaxis()->SetLabelColor(kGreen+3);		
			} break;

			default : {
				curHist->SetLineColor(kGray);
				//curHist->GetXaxis()->SetLabelColor(kGray);
			}
		}		

		result->Add(curHist);
	}


	// getting data
	for (Int_t s = 0; s < nrOfTrials; ++s) {

		// get cur hist
		TH2D *curHist = static_cast<TH2D*>(result->At(s));

		Int_t tVal = trials[s];

		for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
			
			Int_t blVal = blm[bl];

			for (Int_t tl = 0; tl < nrOfTLs; ++tl) {

				Int_t tlVal = tlm[tl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    TMatrixD *mMap = mergeRatemaps(logfile,mintime,1);

			    Int_t mmRows = mMap->GetNrows();
			    Int_t mmCols = mMap->GetNcols();

				Double_t xVal = (Double_t)(s + (nrOfTrials+1) * bl) + 1.5;
			    for (Int_t r = 0; r < mmRows; ++r) {
			    	for (Int_t c = 0; c < mmCols; ++c) {
			    		if ((*mMap)(r,c) > 0)
			    			curHist->Fill(xVal,(*mMap)(r,c));
			    	}
			    }

			    delete mMap;
			}
		}		
	}

	if (can == 0) {
		can = new TCanvas();
	}

	// draw
	gStyle->SetOptStat(0);

	can->SetLeftMargin(0.1);
	can->SetRightMargin(0.02);
	can->SetTopMargin(0.035);
	can->SetBottomMargin(0.33);

	// draw first
	TH2D *curHist = static_cast<TH2D*>(result->At(0));
	curHist->Draw("CANDLE");

	// draw the rest
	for (Int_t s = 1; s < nrOfTrials; ++s) {
		curHist = static_cast<TH2D*>(result->At(s));
		curHist->Draw("CANDLE SAME");
	}

	return result;
}

/*
	This function creates a candle plot that compares the activity distributions
	of several simulation runs with varying number of TL and BL units. The
	activity distributions are averaged over the simulations with varying BL
	units. The input files are defined by the string "base" and the vectors
	"trials", "tlm", and "blm" as sprintf(logfile,base,tVal,tlVal,blVal) where
	tVal, tlVal, and blVal are elements of these vectors. The different candle
	plots are labeled with the strings in "titles". The MAX-based merged rate 
	maps are derived from the first set of rate maps after "mintime" time steps
	in each analysis file.
*/
TObjArray* getMergedCoverVsTL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, TCanvas *can = 0)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Int_t xBinCnt = (nrOfTrials+1)*nrOfTLs+1;
	Int_t yBinCnt = 20;

	// setting up histogram
	TObjArray *result = new TObjArray();
	result->SetOwner(kTRUE);

	for (Int_t s = 0; s < nrOfTrials; ++s) {
		TH2D *curHist = new TH2D(Form("mergedCoverBLHist_%d",s),"",xBinCnt,0.0,(Double_t)xBinCnt,yBinCnt,0.0,1.000001);		
		curHist->SetDirectory(0);
		for (Int_t i = 1; i <= xBinCnt; ++i)
			curHist->GetXaxis()->SetBinLabel(i,"");

		// set all labels
		for (Int_t s2 = 0; s2 < nrOfTrials; ++s2) {
			for (Int_t e = 0; e < nrOfTLs; ++e) {
				curHist->GetXaxis()->SetBinLabel(s2 + (nrOfTrials+1) * e + 2,titles[s2]);
			}
		}

		// some formatting
		curHist->GetXaxis()->LabelsOption("v");

		curHist->GetXaxis()->SetTitle("# of TL units");
		curHist->GetXaxis()->SetTitleOffset(3.0);
		curHist->GetXaxis()->SetTitleSize(0.06);
		curHist->GetXaxis()->SetLabelSize(0.07);
		curHist->GetXaxis()->SetLabelOffset(0.01);

		curHist->GetXaxis()->SetTickLength(-0.21);

		curHist->GetYaxis()->SetTitle("activity");
		curHist->GetYaxis()->SetTitleOffset(0.80);
		curHist->GetYaxis()->SetTitleSize(0.06);
		curHist->GetYaxis()->SetLabelSize(0.05);

		curHist->GetYaxis()->SetRangeUser(-0.1,1.1);

		switch (s) {
			case 0 : {				
				curHist->SetLineColor(kBlue);				
				//curHist->GetXaxis()->SetLabelColor(kBlue);
			} break;

			case 1 : {				
				curHist->SetLineColor(kRed+1);				
				//curHist->GetXaxis()->SetLabelColor(kRed+1);
			} break;

			case 2 : {
				curHist->SetLineColor(kOrange+2);		
				//curHist->GetXaxis()->SetLabelColor(kOrange+2);		
			} break;

			case 3 : {
				curHist->SetLineColor(kGreen+3);		
				//curHist->GetXaxis()->SetLabelColor(kGreen+3);		
			} break;

			default : {
				curHist->SetLineColor(kGray);
				//curHist->GetXaxis()->SetLabelColor(kGray);
			}
		}		

		result->Add(curHist);
	}


	// getting data
	for (Int_t s = 0; s < nrOfTrials; ++s) {

		// get cur hist
		TH2D *curHist = static_cast<TH2D*>(result->At(s));

		Int_t tVal = trials[s];

		for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
			
			Int_t blVal = blm[bl];

			for (Int_t tl = 0; tl < nrOfTLs; ++tl) {

				Int_t tlVal = tlm[tl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    TMatrixD *mMap = mergeRatemaps(logfile,mintime,1);

			    Int_t mmRows = mMap->GetNrows();
			    Int_t mmCols = mMap->GetNcols();

				Double_t xVal = (Double_t)(s + (nrOfTrials+1) * tl) + 1.5;
			    for (Int_t r = 0; r < mmRows; ++r) {
			    	for (Int_t c = 0; c < mmCols; ++c) {
			    		if ((*mMap)(r,c) > 0)
			    			curHist->Fill(xVal,(*mMap)(r,c));
			    	}
			    }

			    delete mMap;
			}
		}		
	}

	if (can == 0) {
		can = new TCanvas();
	}

	// draw
	gStyle->SetOptStat(0);

	can->SetLeftMargin(0.1);
	can->SetRightMargin(0.02);
	can->SetTopMargin(0.035);
	can->SetBottomMargin(0.33);

	// draw first
	TH2D *curHist = static_cast<TH2D*>(result->At(0));
	curHist->Draw("CANDLE");

	// draw the rest
	for (Int_t s = 1; s < nrOfTrials; ++s) {
		curHist = static_cast<TH2D*>(result->At(s));
		curHist->Draw("CANDLE SAME");
	}

	return result;
}

/*
	This function creates a gridness distribution from analysis file "file"
	using "nrOfBins" bins based on the first set of rate maps found after 
	"mintime" inputs.
*/
TH1D* getGridnessDist(ULong64_t mintime, const char *file, Int_t nrOfBins)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    int maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    if (maxNodeID == -1) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return 0;
    }

    std::vector<Double_t> nodeGridness(maxNodeID+1,-3.0);

    printf("nr of nodes: %d",maxNodeID+1);

    Double_t gridness = 0;
    rm->SetBranchStatus("gridness",1);
    rm->SetBranchAddress("gridness",&gridness);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);

    int j = 0;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

    	if ((nodeGridness[nodeID] != -3.0) || (time < mintime))
    		continue;
    	
		nodeGridness[nodeID] = gridness;

    	if (j++ >= maxNodeID)
    		break;
    }

    TH1D *result = new TH1D(Form("griddist_at_%lld",mintime),"",nrOfBins,-1.5,1.5);
    result->SetDirectory(0);

    for (Double_t gd : nodeGridness) {
    	if (gd != -3.0) {
    		result->Fill(gd);
    	}
    }

    f->Close();
    delete f;

    return result;
}


/*
	This function creates a distribution of grid orientations from analysis file
	"file" using "nrOfBins" bins based on the first set of rate maps found after 
	"mintime" inputs. It considers all rate maps that have a gridness score 
	greater or equal "gThres" and considers a maximum orientation angle of 
	"maxAngle".
*/
TH1D* getOrientationDist(ULong64_t mintime, const char *file, Double_t gThres = 0.4, Int_t nrOfBins = 60, Int_t maxAngle = 60)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    int maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    if (maxNodeID == -1) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return 0;
    }

    std::vector<Double_t> nodeOrientation(maxNodeID+1,-10.0);

    //printf("nr of nodes: %d",maxNodeID+1);

    Double_t orientation = 0;
    rm->SetBranchStatus("orientation",1);
    rm->SetBranchAddress("orientation",&orientation);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);

    Double_t gridness = 0;
    rm->SetBranchStatus("gridness",1);
    rm->SetBranchAddress("gridness",&gridness);

    int j = 0;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

    	if ((nodeOrientation[nodeID] != -10.0) || (time < mintime) || (gridness < gThres))
    		continue;
    	
		nodeOrientation[nodeID] = orientation;

    	if (j++ >= maxNodeID)
    		break;
    }

    TH1D *result = new TH1D(Form("orientdist_at_%lld",mintime),"",nrOfBins,0,maxAngle);
    result->SetDirectory(0);

    for (Double_t angle : nodeOrientation) {
    	if (angle == -10.0)
    		continue;

   		result->Fill(angle * 180. / TMath::Pi());
    }

    f->Close();
    delete f;

    return result;
}

/*
	This function creates a distribution of grid spacings from analysis file
	"file" using "maxSpace" bins based on the first set of rate maps found after 
	"mintime" inputs. It considers all rate maps that have a gridness score 
	greater or equal "gThres". The range of the histogram will be 0 to 
	(maxSpace*6) representing centimeters if a rate map resolution of 3cm per
	rate map bin is assumed.
*/
TH1D* getSpacingDist(ULong64_t mintime, const char *file, Double_t gThres = 0.4, Int_t maxSpace = 33)
{
    TFile *f = TFile::Open(file);

    TTree *rm = (TTree*)f->Get("ratemaps");

    gROOT->cd();

    rm->SetBranchStatus("*",0);
    rm->SetBranchStatus("nodeID",1);

    Int_t nodeID = 0;
    rm->SetBranchAddress("nodeID",&nodeID);

    int maxNodeID = -1;

    Int_t nrOfEntries = rm->GetEntries();
    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);
    	if (nodeID > maxNodeID) {
    		maxNodeID = nodeID;
    	}
    }

    if (maxNodeID == -1) {
    		printf("no entries found\n");
		    f->Close();
		    delete f;
		    return 0;
    }

    std::vector<Double_t> nodeSpacing(maxNodeID+1,-10.0);

    //printf("nr of nodes: %d",maxNodeID+1);

    Double_t spacing = 0;
    rm->SetBranchStatus("spacing",1);
    rm->SetBranchAddress("spacing",&spacing);

	ULong64_t  time;
    rm->SetBranchStatus("inputCnt",1);
    rm->SetBranchAddress("inputCnt",&time);

    Double_t gridness = 0;
    rm->SetBranchStatus("gridness",1);
    rm->SetBranchAddress("gridness",&gridness);

    int j = 0;

    for (int i = 0; i < nrOfEntries; ++i) {
    	rm->GetEntry(i);

    	if ((nodeSpacing[nodeID] != -10.0) || (time < mintime) || (gridness < gThres))
    		continue;
    	
		nodeSpacing[nodeID] = spacing;

    	if (j++ >= maxNodeID)
    		break;
    }

    TH1D *result = new TH1D(Form("spacingdist_at_%lld",mintime),"",maxSpace,0,maxSpace*6);
    result->SetDirectory(0);

    for (Double_t space : nodeSpacing) {
    	if (space == -10.0)
    		continue;

   		result->Fill(space * 3);
    }

    f->Close();
    delete f;

    return result;
}


/*
	This function returns the number of major grid orientations that occur in
	the first set of rate maps after "mintime" time steps in analysis file 
	"file". Only rate maps with a gridness score higher "gThres" are considered.
	An orientation is considered to be a major orientation if its value in the
	corresponding histogram is larger than the maximum histogram value times 
	"mThres".
*/
Double_t getOrientations(ULong64_t mintime, const char *file, Double_t gThres = 0.4, Double_t mThres = 0.8)
{
	TH1D *orientHist = getOrientationDist(mintime,file,gThres);

	Double_t maxVal = orientHist->GetMaximum() * mThres;

	Double_t result = 0;

	for (Int_t i = 1; i <= orientHist->GetNbinsX(); ++i)
		if (orientHist->GetBinContent(i) > maxVal)
			result += 1.0;

	delete orientHist;

	if (result < 1)
		printf("Warning: less than 1 orientations\n");

	return result;
}


/*
	This function plots the average number of major orientations across 
	trials with different numbers of TL units against the number of BL units. 
	The input files are	defined by the string "base" and the vectors "trials", 
	"tlm", and "blm" as	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, 
	tlVal, and blVal are elements of the vectors. For each element in "trials" 
	a graph is plotted and labeled with a corresponding entry in "titles". The 
	numbers of major orientations are aquired from each input file by 
	getOrientations("mintime",logfile,0.4,"mThres") and averaged over the 
	entries of "tlm".
*/
void plotAvgOrientationsVsBL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, Double_t mThres = 0.8)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfBLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		x[bl] = bl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfBLs];
		Int_t tVal = trials[t];
		for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
			Int_t blVal = blm[bl];
			y[t][bl]   = 0;
			Double_t oCnt = 0;
			for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
				Int_t tlVal = tlm[tl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    Double_t orientations = getOrientations(mintime,logfile,0.4,mThres);
			    if (orientations >= 1) {
			    	y[t][bl] += orientations;
			    	oCnt += 1.0;
			    }
			    
			}
			if (oCnt > 0)
				y[t][bl] /= oCnt;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfBLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
		Int_t blVal = blm[bl];
		xAxis->SetBinLabel(xAxis->FindBin(bl+1),Form("%d",blVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of BL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0.0,4.0);
	yAxis->SetTitle("# of orientations");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}


/*
	This function plots the average number of major orientations across 
	trials with different numbers of BL units against the number of TL units. 
	The input files are	defined by the string "base" and the vectors "trials", 
	"tlm", and "blm" as	sprintf(logfile,base,tVal,tlVal,blVal) where tVal, 
	tlVal, and blVal are elements of the vectors. For each element in "trials" 
	a graph is plotted and labeled with a corresponding entry in "titles". The 
	numbers of major orientations are aquired from each input file by 
	getOrientations("mintime",logfile,0.4,"mThres") and averaged over the 
	entries of "blm".
*/
void plotAvgOrientationsVsTL(const char *base, std::vector<int> &trials, std::vector<const char*> &titles, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime)
{
	Int_t nrOfTrials = trials.size();
	Int_t nrOfTLs    = tlm.size();
	Int_t nrOfBLs    = blm.size();

	Double_t *x  = new Double_t[nrOfTLs];
	Double_t **y = new Double_t*[nrOfTrials];

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		x[tl] = tl+1;
	}

	// getting data
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		y[t] = new Double_t[nrOfTLs];
		Int_t tVal = trials[t];
		for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
			Int_t tlVal = tlm[tl];
			y[t][tl]   = 0;
			Double_t oCnt = 0;
			for (Int_t bl = 0; bl < nrOfBLs; ++bl) {
				Int_t blVal = blm[bl];

			    char logfile[200];
			    sprintf(logfile,base,tVal,tlVal,blVal);

			    Double_t orientations = getOrientations(mintime,logfile,0.4);
			    if (orientations >= 1) {
			    	y[t][tl] += orientations;
			    	oCnt += 1.0;
			    }

			}
			if (oCnt > 0)
				y[t][tl] /= oCnt;
		}
	}

	// building graphs
	TObjArray graphs;

	for (Int_t t = 0; t < nrOfTrials; ++t) {
		TGraph *curGraph = new TGraph(nrOfTLs,x,y[t]);
		graphs.Add(curGraph);

		Double_t markerScale = 1.2;

		curGraph->SetTitle(titles[t]);

		// set styles
		switch (t) {
			case 0 : {				
				curGraph->SetMarkerStyle(20);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kBlue+2);
				curGraph->SetLineColor(kBlue);				
				curGraph->SetLineWidth(2);
			} break;

			case 1 : {				
				curGraph->SetMarkerStyle(21);
				curGraph->SetMarkerSize(1.0 * markerScale);
				curGraph->SetMarkerColor(kRed+2);
				curGraph->SetLineColor(kRed);				
				curGraph->SetLineWidth(2);
			} break;

			case 2 : {
				curGraph->SetMarkerStyle(22);
				curGraph->SetMarkerSize(1.2 * markerScale);
				curGraph->SetMarkerColor(kOrange+2);
				curGraph->SetLineColor(kOrange+1);				
				curGraph->SetLineWidth(2);
			} break;

			case 3 : {
				curGraph->SetMarkerStyle(33);
				curGraph->SetMarkerSize(1.8 * markerScale);
				curGraph->SetMarkerColor(kGreen+3);
				curGraph->SetLineColor(kGreen+2);				
				curGraph->SetLineWidth(2);
			} break;

			default : {
				curGraph->SetMarkerStyle(2);
				curGraph->SetMarkerColor(kGray+2);
				curGraph->SetLineColor(kGray);
				curGraph->SetLineWidth(1);
			}
		}

		curGraph->SetFillColor(kWhite);
	}

	// draw graphs
	TCanvas *can = new TCanvas();

	// get first graph
	TGraph *curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle("");
	curGraph->Draw("apl");

	// label and format axis
	TAxis *xAxis = curGraph->GetXaxis();
	TAxis *yAxis = curGraph->GetYaxis();

	for (Int_t tl = 0; tl < nrOfTLs; ++tl) {
		Int_t tlVal = tlm[tl];
		xAxis->SetBinLabel(xAxis->FindBin(tl+1),Form("%d",tlVal));
	}
	xAxis->SetLabelSize(0.08);

	xAxis->SetTitle("# of TL units");
	xAxis->SetTitleOffset(1.4);
	xAxis->SetTitleSize(0.06);

	yAxis->SetRangeUser(0,4);
	yAxis->SetTitle("# of orientations");
	yAxis->SetTitleSize(0.06);
	yAxis->SetTitleOffset(0.9);
	yAxis->SetLabelSize(0.05);

	// draw other graphs
	for (Int_t t = 1; t < nrOfTrials; ++t) {
		TGraph *curGraph = (TGraph*)graphs.At(t);
		curGraph->Draw("same lp");
	}

	can->SetLeftMargin(0.15);
	can->SetBottomMargin(0.2);

	can->Modified();
	can->Update();

	curGraph = (TGraph*)graphs.At(0);
	curGraph->SetTitle(titles[0]);

	can->BuildLegend();

	// cleaning up
	delete[] x;
	for (Int_t t = 0; t < nrOfTrials; ++t) {
		delete[] y[t];
	}
	delete[] y;
}

/*
	This function plots Orientation-Histograms (see getOrientationDist() ) for
	a series of input files that are defined by the string "base" and the 
	vectors "trials", "tlm", and "blm" as sprintf(logfile,base,sVal,tlVal,blVal)
	where sVal, tlVal, and blVal are elements of the vectors. The histograms
	are plotted on canvas "can" using option "option" and are stored as external
	files defined by the string "pdfbase" as 
	sprintf(pdffile,pdfbase,sVal,tlVal,blVal). The number of bins used in the
	orientation histograms is defined by the "nrOfBins" parameter. The 
	histograms are colored according to their number of major orientations, 
	i.e., 0 = gray, 1 = green, 2 = orange, 3 = blue, 4+ = white.
*/
void plotAllOrientRaw(const char* base, const char* pdfbase, std::vector<int> &trials, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, TCanvas *can, Option_t* option = "", Int_t nrOfBins = 60)
{
	Int_t nrOfSims = trials.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = trials[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH1D *od = getOrientationDist(mintime,logfile,0.4,nrOfBins);

			    od->SetFillColor(kBlack);

				// calculate orientations
				Double_t maxVal = od->GetMaximum() * 0.6;

				Int_t oCnt = 0;

				for (Int_t i = 1; i <= od->GetNbinsX(); ++i)
					if (od->GetBinContent(i) > maxVal)
						oCnt++;

				switch (oCnt) {
					case 0 : {				
						can->SetFrameFillColor(kGray);
					} break;

					case 1 : {				
						//can->SetFrameFillColor(kSpring+2);
						can->SetFrameFillColor(TColor::GetColor(191,217,191));
					} break;

					case 2 : {
						//can->SetFrameFillColor(kOrange-4);
						can->SetFrameFillColor(TColor::GetColor(242,217,191));
					} break;

					case 3 : {
						//can->SetFrameFillColor(kAzure-9);
						can->SetFrameFillColor(TColor::GetColor(191,206,255));
					} break;

					default : {
						can->SetFrameFillColor(kWhite);
					}
				}

			    od->Draw(option);		    

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete od;
			}
		}		
	}

}

/*
	This function plots Gridness-Histograms (see getGridnessDist() ) for
	a series of input files that are defined by the string "base" and the 
	vectors "trials", "tlm", and "blm" as sprintf(logfile,base,sVal,tlVal,blVal)
	where sVal, tlVal, and blVal are elements of the vectors. The histograms
	are plotted on canvas "can" using option "option" and are stored as external
	files defined by the string "pdfbase" as 
	sprintf(pdffile,pdfbase,sVal,tlVal,blVal). The number of bins used in the
	gridness histograms is defined by the "nrOfBins" parameter. If "skipColor"
	is false, the histograms are colored according to their number of major 
	orientations, i.e., 0 = gray, 1 = green, 2 = orange, 3 = blue, 4+ = white.
	The range of the gridness histograms is defined by the "minRange" and 
	"maxRange" parameters.
*/
void plotAllGridRaw(const char* base, const char* pdfbase, std::vector<int> &trials, 
	std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, TCanvas *can, 
	Option_t* option = "", Int_t nrOfBins = 30, Bool_t skipColor = kFALSE, 
	Double_t minRange = -0.5, Double_t maxRange = 1.5)
{
	Int_t nrOfSims = trials.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = trials[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH1D *od = getOrientationDist(mintime,logfile,0.4,24);

			    od->SetFillColor(kBlack);

				// calculate orientations
				Double_t maxVal = od->GetMaximum() * 0.6;

				Int_t oCnt = 0;

				for (Int_t i = 1; i <= od->GetNbinsX(); ++i)
					if (od->GetBinContent(i) > maxVal)
						oCnt++;

				delete od;

				if (skipColor == kTRUE) {
					oCnt = 4;
				}

				switch (oCnt) {
					case 0 : {				
						can->SetFrameFillColor(kGray);
					} break;

					case 1 : {				
						//can->SetFrameFillColor(kSpring+2);
						can->SetFrameFillColor(TColor::GetColor(191,217,191));
					} break;

					case 2 : {
						//can->SetFrameFillColor(kOrange-4);
						can->SetFrameFillColor(TColor::GetColor(242,217,191));
					} break;

					case 3 : {
						//can->SetFrameFillColor(kAzure-9);
						can->SetFrameFillColor(TColor::GetColor(191,206,255));
					} break;

					default : {
						can->SetFrameFillColor(kWhite);
					}
				}

				TH1D* gd = getGridnessDist(mintime,logfile,nrOfBins);

				gd->SetFillColor(kBlack);

				gd->GetXaxis()->SetRangeUser(minRange,maxRange);

			    gd->Draw(option);		    

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete gd;
			}
		}		
	}

}

/*
	This function plots Spacing-Histograms (see getSpacingDist() ) for
	a series of input files that are defined by the string "base" and the 
	vectors "trials", "tlm", and "blm" as sprintf(logfile,base,sVal,tlVal,blVal)
	where sVal, tlVal, and blVal are elements of the vectors. The histograms
	are plotted on canvas "can" using option "option" and are stored as external
	files defined by the string "pdfbase" as 
	sprintf(pdffile,pdfbase,sVal,tlVal,blVal). The number of bins used in the
	orientation histograms is defined by the "nrOfBins" parameter. 
*/
void plotAllSpacingRaw(const char* base, const char* pdfbase, std::vector<int> &trials, std::vector<int> &tlm, std::vector<int> &blm, ULong64_t mintime, TCanvas *can, Option_t* option = "", Int_t nrOfBins = 33)
{
	Int_t nrOfSims = trials.size();
	Int_t nrOfTL   = tlm.size();
	Int_t nrOfBL   = blm.size();

	for (Int_t s = 0; s < nrOfSims; ++s) {
		for (Int_t tl = 0; tl < nrOfTL; ++tl) {
			for (Int_t bl = 0; bl < nrOfBL; ++bl) {

				Int_t sVal  = trials[s];
				Int_t tlVal = tlm[tl];
				Int_t blVal = blm[bl];


			    char logfile[200];
			    sprintf(logfile,base,sVal,tlVal,blVal);

			    TH1D *sd = getSpacingDist(mintime,logfile,0.4,nrOfBins);		    

			    sd->SetFillColor(kBlack);

				can->SetFrameFillColor(kWhite);

			    sd->Draw(option);		    

			    char pdffile[200];
			    sprintf(pdffile,pdfbase,sVal,tlVal,blVal);

			    can->SaveAs(pdffile);

			    delete sd;
			}
		}		
	}

}


